# Some Useful Notes

## Execution related notes

Use the following command in the terminal to safely run [FEB3](https://bitbucket.org/vasvav/feb3-finite-element-bioengineering-in-3d/wiki/Home "Phoebe") remotely. Make sure that path and filename of the executable is correct, as well as the output-streams used by *nohup*.

    `nohup mpiexec -n [INTEGER] ./feb3-opt.exe -i [STRING] >feb3-opt.out  2>feb3-opt.err  </dev/null  &`

The *integer* corresponds to the number of processes FEB3 will use, while the *string* corresponds to the name of the input ASCII *dat*-file.

## Paraview related notes

Adopt the follow strings when using [Paraview](https://www.paraview.org/ "Paraview")'s calculator filter to post-process various variables. First, do confirm that the labels used for each solver (see input file) agree with the ones used below.

### Solid mechanics (3D) problems

- displacement vector:

    `NC3D__u0*iHat + NC3D__u1*jHat + NC3D__u2*kHat`

- mean solid stress (or hydrostatic pressure; positive values for compressive stress):

    `-(NC3D__S00+NC3D__S11+NC3D__S22)/3`

- deformation gradient determinant (`NC3D__J`):

    `NC3D__F00*(NC3D__F11*NC3D__F22-NC3D__F12*NC3D__F21)-NC3D__F01*(NC3D__F10*NC3D__F22-NC3D__F12*NC3D__F20)+NC3D__F02*(NC3D__F10*NC3D__F21-NC3D__F11*NC3D__F20)`

- Cauchy stress tensor components (`NC3D__Sc00`, `NC3D__Sc01`, `NC3D__Sc02`, `NC3D__Sc11`, `NC3D__Sc12` and `NC3D__Sc22`; applicable for forward analysis):

    `(NC3D__F00*(NC3D__F00*NC3D__S00+NC3D__F01*NC3D__S01+NC3D__F02*NC3D__S02)+NC3D__F01*(NC3D__F00*NC3D__S01+NC3D__F01*NC3D__S11+NC3D__F02*NC3D__S12)+NC3D__F02*(NC3D__F00*NC3D__S02+NC3D__F01*NC3D__S12+NC3D__F02*NC3D__S22))/NC3D__J`

    `(NC3D__F10*(NC3D__F00*NC3D__S00+NC3D__F01*NC3D__S01+NC3D__F02*NC3D__S02)+NC3D__F11*(NC3D__F00*NC3D__S01+NC3D__F01*NC3D__S11+NC3D__F02*NC3D__S12)+NC3D__F12*(NC3D__F00*NC3D__S02+NC3D__F01*NC3D__S12+NC3D__F02*NC3D__S22))/NC3D__J`

    `(NC3D__F20*(NC3D__F00*NC3D__S00+NC3D__F01*NC3D__S01+NC3D__F02*NC3D__S02)+NC3D__F21*(NC3D__F00*NC3D__S01+NC3D__F01*NC3D__S11+NC3D__F02*NC3D__S12)+NC3D__F22*(NC3D__F00*NC3D__S02+NC3D__F01*NC3D__S12+NC3D__F02*NC3D__S22))/NC3D__J`

    `(NC3D__F10*(NC3D__F10*NC3D__S00+NC3D__F11*NC3D__S01+NC3D__F12*NC3D__S02)+NC3D__F11*(NC3D__F10*NC3D__S01+NC3D__F11*NC3D__S11+NC3D__F12*NC3D__S12)+NC3D__F12*(NC3D__F10*NC3D__S02+NC3D__F11*NC3D__S12+NC3D__F12*NC3D__S22))/NC3D__J`

    `(NC3D__F20*(NC3D__F10*NC3D__S00+NC3D__F11*NC3D__S01+NC3D__F12*NC3D__S02)+NC3D__F21*(NC3D__F10*NC3D__S01+NC3D__F11*NC3D__S11+NC3D__F12*NC3D__S12)+NC3D__F22*(NC3D__F10*NC3D__S02+NC3D__F11*NC3D__S12+NC3D__F12*NC3D__S22))/NC3D__J`

    `(NC3D__F20*(NC3D__F20*NC3D__S00+NC3D__F21*NC3D__S01+NC3D__F22*NC3D__S02)+NC3D__F21*(NC3D__F20*NC3D__S01+NC3D__F21*NC3D__S11+NC3D__F22*NC3D__S12)+NC3D__F22*(NC3D__F20*NC3D__S02+NC3D__F21*NC3D__S12+NC3D__F22*NC3D__S22))/NC3D__J`

- Von Mises stress:

    `sqrt(NC3D__Sc00^2+NC3D__Sc11^2+NC3D__Sc22^2-NC3D__Sc00*NC3D__Sc11-NC3D__Sc00*NC3D__Sc22-NC3D__Sc11*NC3D__Sc22+3*NC3D__Sc01^2+3*NC3D__Sc02^2+3*NC3D__Sc12^2)`

### Tumour Angiogenesis coupled problem

- interstitial fluid velocity vector:

    `NC3D_TA3D__v0_INTE*iHat + NC3D_TA3D__v1_INTE*jHat + NC3D_TA3D__v2_INTE*kHat`

- transvascular fluid velocity vector:

    `NC3D_TA3D__v0_TRAN*iHat + NC3D_TA3D__v1_TRAN*jHat + NC3D_TA3D__v2_TRAN*kHat`

- displacement vector for the vascular network domain:

    `TA3D__u0*iHat + TA3D__u1*jHat + TA3D__u2*kHat`

- tumour volume increase (percentage) for the solid tissue domain (*XXX* the initial/reference value of the tumour volume):

    `100.*((Volume / XXX) - 1)`

### Multiscale solid mechanics solver (3D) problems

1. Use the *Calculator* filter in [Paraview](https://www.paraview.org/ "Paraview") and define the variable `MS__u`:

    `MS__u0*iHat + MS__u1*jHat + MS__u2*kHat`

2. Similarly define the variable `MS__C`:

    `MS__C0*iHat + MS__C1*jHat + MS__C2*kHat`

3. Define the variables `MS__iHat`, `MS__jHat` and `MS__kHat` respectively:

    `MS__iHat0*iHat + MS__iHat1*jHat + MS__iHat2*kHat`

    `MS__jHat0*iHat + MS__jHat1*jHat + MS__jHat2*kHat`

    `MS__kHat0*iHat + MS__kHat1*jHat + MS__kHat2*kHat`

4. Define the variable `MS__XYZ`:

    `(coordsX-MS__L/2)*MS__iHat+(coordsY-MS__L/2)*MS__jHat+(coordsZ-MS__L/2)*MS__kHat`

5. Define the variable `coordsN` and use it as *Coordinate Results* in [Paraview](https://www.paraview.org/ "Paraview"). For the parameter *S*, use some meaningful value to scale the RVE mesh accordingly.

    `MS__XYZ * S + MS__C`

6. Use the *WarpByVector* filter in [Paraview](https://www.paraview.org/ "Paraview") to deform the RVEs.
