## Intro

Welcome to [FEB3](https://bitbucket.org/vasvav/feb3-finite-element-bioengineering-in-3d/wiki/Home "Phoebe") (pronounced *Phoebe*), a C++ Finite Element and Isogeometric Analysis software for modelling three-dimensional multiphysics problems in bioengineering.
FEB3 is designed as a modular framework, permitting the user to plug in their own constitutive relations and boundary conditions with ease.
It incorporates both forward and inverse finite deformation analysis methods, and implements both implicit and explicit Total Lagrangian solvers.



## About

[FEB3](https://bitbucket.org/vasvav/feb3-finite-element-bioengineering-in-3d/wiki/Home "Phoebe") is capable simulating static and dynamic solid mechanics, parabolic equations, multiphase materials, and mechanics at multiple spatiotemporal scales.
Currently it features the following analyses:

* Nonlinear elasticity
  * Compressible solids
  * Nearly-incompressible solids
  * Isotropic, transversely isotropic and orthotropic elastic materials
  * Continuum scale fibre-reinforced materials
  * Growth mechanics
  * Multiphase mechanics
  * Biphasic and poroelastic deformation and fluid flow
* Coupled micro-macro nonlinear solid mechanics
  * Micro scale incompressible fibre networks
* Reaction-convection-diffusion dynamics

Developed biomedical applications are:

* Soft tissue biomechanics
* Tumour growth
* Angiogenesis
* Drug delivery
* Wound healing

These applications feature coupled multiphysics solvers, permitting a physiological description of the chosen process with user-defined constitutive relations, material properties and boundary conditions.

[FEB3](https://bitbucket.org/vasvav/feb3-finite-element-bioengineering-in-3d/wiki/Home "Phoebe") is an open-source software intended to be used for academic or research purposes. Its core is founded on various open-source libraries (e.g. libMesh, PETSc, GSL, MPICH).
Compilation and utilisation of FEB3 necessitates installation of the open-source libraries listed below, which are freely available online.

Currently, FEB3 supports [Gmsh](http://gmsh.info/ "Gmsh")-formatted input meshes for Finite Element Analysis simulations, while for Isogeometric Analysis simulations a specialised combination of Gmsh and native input data files are required (for details please consult corresponding examples). 
Also, all outputs results are generated in [VTK](https://www.vtk.org/ "The Visualization Toolkit") and [Exodus II](http://prod.sandia.gov/techlib/access-control.cgi/1992/922137.pdf "EXODUS II: A Finite Element Data Model") format files.
Both can be processed using the open-source visualisation platform [Paraview](https://www.paraview.org/ "Paraview").



## Setup

Before compiling FEB3 (*Phoebe*), one needs to configure/compile/install a set of numerical libraries and modelling platforms.
These include [MPICH](https://www.mpich.org/), [GSL](http://www.gnu.org/software/gsl/), [blitz++](http://sourceforge.net/projects/blitz/), [Biodynamo](https://biodynamo.web.cern.ch/), [PETSc](www.mcs.anl.gov/petsc/) and [libMesh](http://libmesh.github.io/).
Guidelines how to install all these libraries are provided below.

Say FEB3 has been `git clone`'d in a local directory, e.g. `/home/johndoe/bitbucket/FEB3`, then a generic *Makefile* can be found in `/home/johndoe/bitbucket/FEB3/Makefile.generic`.
The user can create a copy of this file and name it for example `Makefile` and edit it where necessary, as explained below.
This new *Makefile* can be used as a basis for a user to customise the compilation / installation, and is split in separate sections.
The first section is the only section that needs to be edited by the user - the information are self-explanatory.
The rest of the sections could be left unmodified by regular users, however, developers are free to edit accordingly.

*Phoebe* setup identifies two steps: installation of the dependencies (libraries and platforms) and compilation of FEB3.

### Installations

Firstly, make sure this has been set: `SETUP_LIBS_FIRST = true`.
Subsequently, execute the rules needed to configure / compile / install several libraries (i.e., `setup_gsl`, `configure_gsl`, `install_gsl`, `setup_blitz`, and so on) sequentially. These rules are listed in the third section of the *Makefile*.
Some rules are not however mandatory to execute (i.e., `test_libmesh`, `install_libmesh_doc`), while this rule `test_petsc` it is advisable to be execute and test whether PETSc has been successfully installed.
Also some libraries are not imperative to be installed (i.e., VTK, FLANN).

The proposed installation pipeline assumes that libMesh compilation and execution of FEB3 calls a PETSc-internally installed MPICH version.
This means that PETSc will install a copy of the up-to-date version of MPICH locally (e.g., `/home/johndoe/bitbucket/FEB3/libs/installed/petsc/bin/`).
As such, should one needs to use the *mpiexec* (or equivalently *mpirun*) scripts to run FEB3 in parallel mode then they need to invoke the corresponding locally-installed scripts of MPICH.
Also, installation of FLANN is not imperative for the compilation of the FEB3 code!

### Compilation

After successful compilation and installation of the dependencies, one needs to set this flag to: `SETUP_LIBS_FIRST = false`.
Then, executre the rule `feb3` so that FEB3 starts compiling.
If code compilation and binaries linking is completed, an executable is generated in the directory `/home/johndoe/bitbucket/FEB3/bin/`.

To quickly test that FEB3 works, the user could execute in the terminal `/home/johndoe/bitbucket/FEB3/bin/./feb3-*.exe -t`.
This will run a simple test problem - currently a simple print out of a message ;)

In general, setup of FEB3 has been tested on Unix-based machines only (Ubuntu and macOS); nonetheless, we plan to test also installing FEB3 on Windows in the near future.

### Testing

We are doing our best to keep up testing and maintaining [FEB3](https://bitbucket.org/vasvav/feb3-finite-element-bioengineering-in-3d/wiki/Home "Phoebe") up-to-speed! :)
However, the user can find a series of examples to simulate; these are located in the directory `/home/johndoe/bitbucket/FEB3/examples/`.



## Referencing

To reference [FEB3 (*Phoebe*)](https://bitbucket.org/vasvav/feb3-finite-element-bioengineering-in-3d/wiki/Home "Phoebe"), please cite any of our published works that present this modelling framework:

1. [Wijeratne et al. (2017) PLOS One, https://doi.org/10.1371/journal.pone.0184511](https://doi.org/10.1371/journal.pone.0184511)
2. [Vavourakis et al. (2017) PLOS Comp Bio, https://doi.org/10.1371/journal.pcbi.1005259](https://doi.org/10.1371/journal.pcbi.1005259)
3. [Vavourakis et al. (2016) PLOS One, https://doi.org/10.1371/journal.pone.0159766](https://doi.org/10.1371/journal.pone.0159766)
4. [Wijeratne et al. (2016) Biomech Model Mechanobiol, https://doi.org/10.1007/s10237-015-0745-2](https://doi.org/10.1007/s10237-015-0745-2)



## Credits

The logo of [FEB3 (Phoebe)](https://bitbucket.org/vasvav/feb3-finite-element-bioengineering-in-3d/wiki/Home "Phoebe") was inspired by the illustration of the bird Black Phoebe (*Sayornis nigricans*) found from [here](http://www.audubon.org/field-guide/bird/black-phoebe?site=ny&section=search_results&site=ny&section=search_results "Black Phoebe").
