/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#ifndef __BINOMIAL_BASIS_H__
#define __BINOMIAL_BASIS_H__
//=================================================================================================
class BinomialBasis {
public:
    BinomialBasis () { this->_set_table(); }
    ~BinomialBasis () {}
    //
    inline
    unsigned int operator() (int I, int N) const {
        return ( ( I < 0 || I > N ) ? 0 : this->_bb[N][I] );
    }
private:
    // fundamental function to evaluate the binomial basis
    unsigned int _calc (int I, int N);
    // initialize the table
    void _set_table ();
private:
    // maximum number of bionomials to evaluate
    static constexpr unsigned int max_binom = 500;
    // array containing the binomial basis components
    std::vector< std::vector<unsigned int> > _bb;
};
//=================================================================================================
//
//
//
//=================================================================================================
inline
unsigned int BinomialBasis::_calc (int I, int N) {
    if      ( I <  0 || I >  N ) return 0;
    else if ( I == 0 || I == N ) return 1;
    else {
        Real res = 1.0;
        for (int j=1; j<=I; j++)
            res *= (N-I+j)/(j+0.0);
        return static_cast<unsigned int>(res);
    }
}
inline
void BinomialBasis::_set_table () {
    this->_bb.resize(max_binom);
    for (unsigned int N=0; N<max_binom; N++) {
        this->_bb[N].assign(N+1, 0);
        for (unsigned int I=0; I<N+1; I++)
            this->_bb[N][I] = this->_calc(I, N);
    }
}
//=================================================================================================
#endif // __BINOMIAL_BASIS_H__
//=================================================================================================
