/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#ifndef __BLITZ_TENSORS_H__
#define __BLITZ_TENSORS_H__
//=================================================================================================
template<class X, int N>
class Vector {
public:
    Vector () { comp.resize(N); comp=X(0); }
    Vector (const blitz::Array<X, 1>& a) { comp.resize(N); comp = a; }
    Vector (const Vector<X, N>& v) { comp.resize(N); comp = v.comp; }
    ~Vector () {}
    //
    inline
    const X& operator() (int i) const { return this->comp(i); }
    inline
    Vector<X, N>& operator= (const Vector<X, N>& in) { this->comp = in.comp; return *this; }
public:
    blitz::Array<X, 1> comp;
};
//-------------------------------------------------------------------------------------------------
template<class X, int N>
class Dyadic {
public:
    Dyadic () { comp.resize(N, N); comp=X(0); }
    Dyadic (const blitz::Array<X, 2>& a) { comp.resize(N, N); comp = a; }
    Dyadic (const Dyadic<X, N>& d) { comp.resize(N, N); comp = d.comp; }
    ~Dyadic () {}
    //
    inline
    const X& operator() (int i, int j) const { return this->comp(i,j); }
    inline
    Dyadic<X, N>& operator= (const Dyadic<X, N>& in) { this->comp = in.comp; return *this; }
public:
    blitz::Array<X, 2> comp;
};
//-------------------------------------------------------------------------------------------------
template<class X, int N>
class Triadic {
public:
    Triadic () { comp.resize(N, N, N); comp=X(0); }
    Triadic (const blitz::Array<X, 3>& a) { comp.resize(N, N, N); comp = a; }
    Triadic (const Triadic<X, N>& d) { comp.resize(N, N, N); comp = d.comp; }
    ~Triadic () {}
    //
    inline
    const X& operator() (int i, int j, int k) const { return this->comp(i,j,k); }
    inline
    Triadic<X, N>& operator= (const Triadic<X, N>& in) { this->comp = in.comp; return *this; }
public:
    blitz::Array<X, 3> comp;
};
//-------------------------------------------------------------------------------------------------
template<class X, int N>
class Quindic {
public:
    Quindic () { comp.resize(N, N, N, N); comp=X(0); }
    Quindic (const blitz::Array<X, 4>& a) { comp.resize(N, N, N, N); comp = a; }
    Quindic (const Quindic<X, N>& d) { comp.resize(N, N, N, N); comp = d.comp; }
    ~Quindic () {}
    //
    inline
    const X& operator() (int i, int j, int k, int l) const { return this->comp(i,j,k,l); }
    inline
    Quindic<X, N>& operator= (const Quindic<X, N>& in) { this->comp = in.comp; return *this; }
public:
    blitz::Array<X, 4> comp;
};
//-------------------------------------------------------------------------------------------------
template<class X, int N>
class Sextic {
public:
    Sextic () { comp.resize(N, N, N, N, N, N); comp=X(0); }
    Sextic (const blitz::Array<X, 6>& a) { comp.resize(N, N, N, N, N, N); comp = a; }
    Sextic (const Sextic<X, N>& d) { comp.resize(N, N, N, N, N, N); comp = d.comp; }
    ~Sextic () {}
    //
    inline
    const X& operator() (int i, int j, int k, int l, int r, int s) const { return this->comp(i,j,k,l,r,s); }
    inline
    Sextic<X, N>& operator= (const Sextic<X, N>& in) { this->comp = in.comp; return *this; }
public:
    blitz::Array<X, 6> comp;
};
//=================================================================================================
#endif //  __BLITZ_TENSORS_H__
//=================================================================================================
