/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#ifndef __EXPLICIT_NAVIER_CAUCHY_SOLVER_3D_AUX_INLINE_H__
#define __EXPLICIT_NAVIER_CAUCHY_SOLVER_3D_AUX_INLINE_H__
//=================================================================================================
std::vector<Point> RVE_vertices (const Point& z, const Real& dz) {
    const Real dz_2 = 0.5*dz;
    //
    std::vector<Point> v(8);
    v[0] = z + Point(-dz_2, -dz_2, -dz_2);
    v[1] = z + Point( dz_2, -dz_2, -dz_2);
    v[2] = z + Point( dz_2,  dz_2, -dz_2);
    v[3] = z + Point(-dz_2,  dz_2, -dz_2);
    v[4] = z + Point(-dz_2, -dz_2,  dz_2);
    v[5] = z + Point( dz_2, -dz_2,  dz_2);
    v[6] = z + Point( dz_2,  dz_2,  dz_2);
    v[7] = z + Point(-dz_2,  dz_2,  dz_2);
    //
    return v;
}
//-------------------------------------------------------------------------------------------------
#define calc_fields_RVE_(_fe_, _fld_) \
    do { \
        if ( ri_ptr->mat->get_biomech_param(100) ) { \
            const std::vector<Point> rve_v                                      \
                = RVE_vertices(q_pnt[qp], ri_ptr->mat->get_biomech_param(101)); \
            \
            const std::vector< std::vector<Real> >&             Theta = (_fe_)->get_phi();  \
            const std::vector< std::vector<RealGradient> >& dTheta_dX = (_fe_)->get_dphi(); \
            const std::vector<Real>& dZ0_dX0 = (_fe_)->get_dxidx();   \
            const std::vector<Real>& dZ0_dX1 = (_fe_)->get_dxidy();   \
            const std::vector<Real>& dZ0_dX2 = (_fe_)->get_dxidz();   \
            const std::vector<Real>& dZ1_dX0 = (_fe_)->get_detadx();  \
            const std::vector<Real>& dZ1_dX1 = (_fe_)->get_detady();  \
            const std::vector<Real>& dZ1_dX2 = (_fe_)->get_detadz();  \
            const std::vector<Real>& dZ2_dX0 = (_fe_)->get_dzetadx(); \
            const std::vector<Real>& dZ2_dX1 = (_fe_)->get_dzetady(); \
            const std::vector<Real>& dZ2_dX2 = (_fe_)->get_dzetadz(); \
            (_fe_)->reinit(elem, &rve_v);                             \
            \
            for (unsigned int v=0; v<8; v++) {                                            \
                (_fld_)[v].current_time = current_time;                                   \
                \
                for (unsigned int ivar=0; ivar<SP3D; ivar++) {                            \
                    (_fld_)[v].auxiliary[ivar] = rve_v[v](ivar);                          \
                }                                                                         \
                \
                for (unsigned int ivar=0; ivar<SP3D; ivar++) {                            \
                    for (unsigned int l=0; l<var_dofs[ivar]; l++) {                       \
                        const Real ui = u_sys.current_solution(dof_indices_var[ivar][l]); \
                        (_fld_)[v].displacement(ivar) += Theta[l][v] * ui;                \
                        (_fld_)[v].grad_displacement(ivar) += dTheta_dX[l][v] * ui;       \
                    }                                                                     \
                }                                                                         \
                (_fld_)[v].deform_gradient =                                              \
                    deformation_gradient_3d((_fld_)[v].grad_displacement);                \
                \
                (_fld_)[v].jacobian.comp = dZ0_dX0[v], dZ0_dX1[v], dZ0_dX2[v],            \
                                           dZ1_dX0[v], dZ1_dX1[v], dZ1_dX2[v],            \
                                           dZ2_dX0[v], dZ2_dX1[v], dZ2_dX2[v];            \
            }                                                                             \
        } \
    } while (false)
//=================================================================================================
#endif // __EXPLICIT_NAVIER_CAUCHY_SOLVER_3D_AUX_INLINE_H__
//=================================================================================================
