/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#ifndef __PARAVIEW_IO_H__
#define __PARAVIEW_IO_H__
//=================================================================================================
#include "libmesh/libmesh_common.h"
#include "libmesh/mesh_input.h"
#include "libmesh/mesh_output.h"
#include "libmesh/mesh_base.h"
#include "libmesh/equation_systems.h"
#include "libmesh/elem.h"
#include "feb3/auxiliary.h"
//=================================================================================================
class ParaviewIO : public MeshInput<MeshBase>, public MeshOutput<MeshBase> {
// member functions:
public:
    ParaviewIO (MeshBase& mesh);
    //
    void write_nodal_data (const std::string& fname, EquationSystems& es);
    virtual
    void write_nodal_data ( const std::string& fname, const std::vector<Number>& soln,
                            const std::vector<std::string>& names );
    void write_boundary (const std::string& fname);
    // functions to control the outlook of the output produced
    void write_reduced_mesh (bool doit);
    void write_refined_mesh (bool doit);
    void hide_variable (const std::string& name);
    virtual
    void read  (const std::string& fname);
    virtual
    void write (const std::string& fname);
private:
    void _write_ascii ( const std::string& fname, const std::vector<Number>* soln,
                        const std::vector<std::string>* names );
    void _cell_connectivity (const Elem* elem, std::vector<unsigned int>& vtk_cell_conn);
    unsigned int _cell_type   (const Elem* elem);
    unsigned int _cell_offset (const Elem* elem);
    bool _is_hidden_variable (const std::string& name);
// member objects:
private:
    bool _write_reduced_mesh,
         _write_refined_mesh;
    std::set<std::string> _hide_vars;
};
//=================================================================================================
#include "feb3/paraview_io_inl.h"
//=================================================================================================
#endif // __PARAVIEW_IO_H__
//=================================================================================================
