/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#ifndef __UTILS__BOUNDARY_CONDITION_INFO_H__
#define __UTILS__BOUNDARY_CONDITION_INFO_H__
//=================================================================================================
class BoundaryConditionInfo {
public:
    enum Type {
        DIRICHLET_PRESCRIBED_X  = 100,
        DIRICHLET_PRESCRIBED_Y  = 101,
        DIRICHLET_PRESCRIBED_Z  = 102,
        DIRICHLET_PRESCRIBED    = 103,
        DIRICHLET_PRESCRIBED_RD = 104,
        DIRICHLET_PRESCRIBED_PL = 105,
        DIRICHLET_PRESCRIBED_CY = 106,
        DIRICHLET_PRESCRIBED_PR = 107,
        DIRICHLET_PRESCRIBED_XY = 110,
        DIRICHLET_PRESCRIBED_YZ = 111,
        DIRICHLET_PRESCRIBED_XZ = 112,
        NEUMANN_NORMAL          = 203,
        NEUMANN_VECTOR          = 204
    };
    static constexpr Real generic_coord = 1.0e+22;
public:
    BoundaryConditionInfo () : type(NEUMANN_NORMAL) {}
    ~BoundaryConditionInfo () {}
    //
    inline
    void set_active_function  (Real (*f_ptr)(const std::vector<Real>&, const Real)) { this->_active_function = f_ptr; }
    inline
    void set_time_function    (Real (*f_ptr)(const std::vector<Real>&, const Real)) { this->_time_function = f_ptr; }
    inline
    bool calc_active_function (const Real t) const { return static_cast<bool>(this->_active_function(this->time_params, t)); }
    inline
    Real calc_time_function   (const Real t) const { return this->_time_function(this->time_params, t); }
private:
    Real (*_active_function) (const std::vector<Real>&, const Real);
    Real (*_time_function)   (const std::vector<Real>&, const Real);
public:
    Type type;
    std::vector<Real> active_params;
    std::vector<Real> time_params;
    Vector3D value;
    Point center, normal;
};
//=================================================================================================
#endif // __UTILS__BOUNDARY_CONDITION_INFO_H__
//=================================================================================================
