/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#ifndef __UTILS_FIELD_INFO_H__
#define __UTILS_FIELD_INFO_H__
//=================================================================================================
class FieldInfo {
public:
    FieldInfo (const unsigned int n_bc, const unsigned int n_bm, const unsigned int n_aux);
    ~FieldInfo () {}
public:
    int a_flag;
    Real current_time, time_step, refence_volume;
    std::vector<Real> biochem;
    std::vector<RealGradient> grad_biochem;
    std::vector<Real> biochem_rate;
    Displacement         displacement;
    DisplacementGradient grad_displacement;
    Velocity             velocity;
    VelocityGradient     grad_velocity;
    Real                 div_velocity;
    std::vector<Real> biomech;
    Dyadic3D         deform_gradient;
    mutable Dyadic3D deform_grad_elas, deform_grad_inel;
    Dyadic3D strain,      stress,
             strain_incr, stress_incr;
    Real pressure;
    Vector3D fluid_velocity;
    Vector3DGradient grad_space_reference,
                     grad_space_current;
    Vector3D fibre_ref[SP3D], fibre_curr[SP3D];
    std::vector<Real> auxiliary;
    std::vector<RealGradient> grad_auxiliary;
    std::vector<Real> auxiliary_rate;
    std::vector<Real> cells;
    std::vector<Real> cells_rate;
    mutable Dyadic3D jacobian;
    #ifdef BIODYNAMO_ENABLED
    const BDM_Domain* bdm_ptr;
    #endif
};
//=================================================================================================
//
//
//
//=================================================================================================
inline
FieldInfo::FieldInfo (const unsigned int n_bc, const unsigned int n_bm, const unsigned int n_aux) {
    Vector3D fibre_default;
    fibre_default.comp = sqrt(frac_1o3);
    //
    this->a_flag = 0;
    //
    this->current_time = 0.0;
    this->time_step = 1.0;
    this->refence_volume = 0.0;
    if ( n_bc ) {
        this->biochem.assign(n_bc, 0.0);
        this->grad_biochem.assign(n_bc, RealGradient(0.0, 0.0, 0.0));
        this->biochem_rate.assign(n_bc, 0.0);
    }
    this->div_velocity = 0.0;
    if ( n_bm ) {
        this->biomech.assign(n_bm, 0.0);
    }
    this->deform_gradient = identity_matrix;
    this->deform_grad_elas = identity_matrix;
    this->deform_grad_inel = identity_matrix;
    this->pressure = 0.0;
    for (int ispdm=0; ispdm<SP3D; ispdm++) {
        this->fibre_ref[ispdm] = fibre_default;
        this->fibre_curr[ispdm] = fibre_default;
    }
    if ( n_aux ) {
        this->auxiliary.assign(n_aux, 0.0);
        this->grad_auxiliary.assign(n_aux, RealGradient(0.0, 0.0, 0.0));
        this->auxiliary_rate.assign(n_aux, 0.0);
    }
    this->jacobian = identity_matrix;
    //
    #ifdef BIODYNAMO_ENABLED
    this->bdm_ptr = 0;
    #endif
}
//=================================================================================================
#endif // __UTILS_FIELD_INFO_H__
//=================================================================================================
