/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#ifndef __UTILS__LOADING_INFO_H__
#define __UTILS__LOADING_INFO_H__
//=================================================================================================
class BodyForceInfo {
public:
    BodyForceInfo () {}
    ~BodyForceInfo () {}
    //
    inline
    void set_time_function  (Real (*f_ptr)(const std::vector<Real>&, const Real)) { this->_time_function = f_ptr; }
    inline
    Real calc_time_function (const Real t) const { return this->_time_function(this->time_params, t); }
private:
    Real (*_time_function) (const std::vector<Real>&, const Real);
public:
    Vector3D value;
    std::vector<Real> time_params;
};
//-------------------------------------------------------------------------------------------------
class NodalForceInfo {
public:
    NodalForceInfo () : load_vec(0) {}
    ~NodalForceInfo () {}
    //
    inline
    void set_time_function__employ  (Real (*f_ptr)(const std::vector<Real>&, const Real)) { this->_time_function__employ = f_ptr; }
    inline
    Real calc_time_function__employ (const Real t) const { return this->_time_function__employ(this->time_params__employ, t); }
    inline
    void set_time_function__evaluate  (Real (*f_ptr)(const std::vector<Real>&, const Real)) { this->_time_function__evaluate = f_ptr; }
    inline
    bool calc_time_function__evaluate (const Real t) const { return (bool)this->_time_function__evaluate(this->time_params__evaluate, t); }
private:
    Real (*_time_function__employ)   (const std::vector<Real>&, const Real);
    Real (*_time_function__evaluate) (const std::vector<Real>&, const Real);
public:
    std::set<unsigned int> node_local;
    NumericVector<Number>* load_vec;
    std::vector<Real> time_params__employ,
                      time_params__evaluate;
};
//-------------------------------------------------------------------------------------------------
class PressureLoadInfo {
public:
    PressureLoadInfo () {}
    ~PressureLoadInfo () {}
    //
    inline
    void set_time_function__employ  (Real (*f_ptr)(const std::vector<Real>&, const Real)) { this->_time_function__employ = f_ptr; }
    inline
    Real calc_time_function__employ (const Real t) const { return this->_time_function__employ(this->time_params__employ, t); }
    inline
    void set_time_function__evaluate  (Real (*f_ptr)(const std::vector<Real>&, const Real)) { this->_time_function__evaluate = f_ptr; }
    inline
    bool calc_time_function__evaluate (const Real t) const { return (bool)this->_time_function__evaluate(this->time_params__evaluate, t); }
private:
    Real (*_time_function__employ)   (const std::vector<Real>&, const Real);
    Real (*_time_function__evaluate) (const std::vector<Real>&, const Real);
public:
    blitz::Array<Real, 1> load_vec;
    std::vector<Real> time_params__employ,
                      time_params__evaluate;
};
//=================================================================================================
#endif // __UTILS__LOADING_INFO_H__
//=================================================================================================
