/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#ifndef __UTILS__SOME_PAIRS_H__
#define __UTILS__SOME_PAIRS_H__
//=================================================================================================
template<class X>
class Dual {
public:
    Dual () : _first(X()), _second(X()) {}
    Dual (const X& a, const X& b) : _first(a), _second(b) {}
    Dual (const Dual<X>& in) { _first=in._first; _second=in._second; }
    ~Dual () {}
    //
    inline
    Dual& operator= (const Dual<X>& in) { this->_first=in.get_first(); this->_second=in.get_second(); return *this; }
    inline
    const X& operator() (int i) const { if (i==0) return this->_first; else if (i==1) return this->_second; else libmesh_error(); }
    inline
    X&       operator() (int i)       { if (i==0) return this->_first; else if (i==1) return this->_second; else libmesh_error(); }
    inline
    const X& get_first () const { return this->_first; }
    inline
    const X& get_second () const { return this->_second; }
private:
    X _first, _second;
};
//-------------------------------------------------------------------------------------------------
template<class X>
class Triple {
public:
    Triple () : _first(X()), _second(X()), _third(X()) {}
    Triple (const X& a, const X& b, const X& c) : _first(a), _second(b), _third(c) {}
    Triple (const Triple<X>& in) { _first=in._first; _second=in._second; _third=in._third; }
    ~Triple () {}
    //
    inline
    Triple& operator= (const Triple<X>& in) { this->_first=in.get_first(); this->_second=in.get_second(); this->_third=in.get_third(); return *this; }
    inline
    const X& operator() (int i) const { if (i==0) return this->_first; else if (i==1) return this->_second; else if (i==2) return this->_third; else libmesh_error(); }
    inline
    X&       operator() (int i)       { if (i==0) return this->_first; else if (i==1) return this->_second; else if (i==2) return this->_third; else libmesh_error(); }
    inline
    const X& get_first () const { return this->_first; }
    inline
    const X& get_second () const { return this->_second; }
    inline
    const X& get_third () const { return this->_third; }
private:
    X _first, _second, _third;
};
//-------------------------------------------------------------------------------------------------
template<class X, class Y>
class my_pair {
public:
    my_pair () : _first(X()), _second(Y()) {}
    my_pair (const X& a, const Y& b) : _first(a), _second(b) {}
    my_pair (const my_pair<X,Y>& in) { _first=in._first; _second=in._second; }
    ~my_pair () {}
    //
    inline
    my_pair& operator= (const my_pair<X,Y>& in) { this->_first=in.get_first(); this->_second=in.get_second(); return *this; }
    inline
    const X& get_first () const { return this->_first; }
    inline
    const Y& get_second () const { return this->_second; }
private:
    X _first;
    Y _second;
};
//-------------------------------------------------------------------------------------------------
template<class X, class Y, class Z>
class my_triple {
public:
    my_triple () : _first(X()), _second(Y()), _third(Z()) {}
    my_triple (const X& a, const Y& b, const Z& c) : _first(a), _second(b), _third(c) {}
    my_triple (const my_triple<X,Y,Z>& in) { _first=in._first; _second=in._second; _third=in._third; }
    ~my_triple () {}
    //
    inline
    my_triple& operator= (const my_triple<X,Y,Z>& in) { this->_first=in.get_first(); this->_second=in.get_second(); this->_third=in.get_third(); return *this; }
    inline
    const X& get_first () const { return this->_first; }
    inline
    const Y& get_second () const { return this->_second; }
    inline
    const Z& get_third () const { return this->_third; }
private:
    X _first;
    Y _second;
    Z _third;
};
//-------------------------------------------------------------------------------------------------
class IndexPair {
public:
    IndexPair () {}
    ~IndexPair () {}
    //
    void clear () { _index.clear(); _id_map.clear(); }
    inline
    void insert (unsigned int id) { _index.push_back(id); _id_map.insert( std::make_pair(id, _index.size()-1) ); }
    inline
    unsigned int get_id (unsigned int k) const { libmesh_assert( k<_index.size() ); return _index[k]; }
    inline
    unsigned int get_index (unsigned int id) const { libmesh_assert( _id_map.end()!=_id_map.find(id) ); return _id_map.find(id)->second; }
    inline
    bool confirm_id (unsigned int id) const { return ( _id_map.end() != _id_map.find(id) ) ? true : false; }
private:
    std::vector<unsigned int> _index;
    std::map<unsigned int, unsigned int> _id_map;
};
//=================================================================================================
#endif // __UTILS__SOME_PAIRS_H__
//=================================================================================================
