/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#ifndef __UTILS__STRINGS_H__
#define __UTILS__STRINGS_H__
//=================================================================================================
#include <vector>
#include <random>
#include <functional>
#include <algorithm>
#include <string>
//=================================================================================================
inline
char_array charset () {
    return char_array( {'0','1','2','3','4',
                        '5','6','7','8','9',
                        '_',
                        'A','B','C','D','E',
                        'F','G','H','I','J',
                        'K','L','M','N','O',
                        'P','Q','R','S','T',
                        'U','V','W','X','Y',
                        'Z',
                        'a','b','c','d','e',
                        'f','g','h','i','j',
                        'k','l','m','n','o',
                        'p','q','r','s','t',
                        'u','v','w','x','y',
                        'z' } );
}
inline
std::string random_string (size_t length, std::function<char(void)> rand_char) {
    std::string s(length, 0);
    std::generate_n(s.begin(), length, rand_char);
    return s;
}
//-------------------------------------------------------------------------------------------------
inline
void find_n_replace (std::string& source, const char* find, const char* replace) {
   size_t findLen = strlen(find);
   size_t replaceLen = strlen(replace);
   size_t pos = 0;
   while ((pos = source.find(find, pos)) != std::string::npos) {
      source.replace( pos, findLen, replace );
      pos += replaceLen;
   }
}
template <typename T>
inline
std::string number2string (const T& Number) {
    std::ostringstream ss;
    ss << Number;
    return ss.str();
}
template <typename T>
inline
T string2number (const std::string& Text) {
    std::istringstream ss(Text);
    T result;
    return ( ss >> result ? result : 0 );
}
//-------------------------------------------------------------------------------------------------
inline
std::string stringfy_number (const Real& Number) {
    char cs[buffer_len];
    sprintf(cs, "%.10f", Number);
    std::string s = cs;
    find_n_replace(s, ".", "P");
    return s;
}
//-------------------------------------------------------------------------------------------------
inline
bool is_double (std::string const& s) {
    std::istringstream ss(s);
    double d;
    return (ss >> d) && (ss >> std::ws).eof();
}
//=================================================================================================
#endif // __UTILS__STRINGS_H__
//=================================================================================================
