/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#ifndef __UTILS__VASCULAR_INFO_H__
#define __UTILS__VASCULAR_INFO_H__
//=================================================================================================
class VascularInfo {
public:
    enum Flow {
        INVALID_FLOW = 1000,
        BLOOD = 1001,
        LYMPH = 1002
    };
    enum Type {
        INVALID_TYPE = 100,
        INLET   = 101,
        OUTLET  = 102,
        SPROUT  = 103,
        MATURE  = 104,
        BRANCH  = 105,
        TIP     = 106
    };
    enum Status {
        NORMAL     = 0,
        COMPRESSED = 1,
        COLLAPSED  = 2
    };
public:
    VascularInfo ();
    ~VascularInfo () {}
public:
    int level;
    // basic features of the vascular network node
    Flow flow;
    Type type;
    Status status;
    // original time of generation and branching, time to become mature
    Real time_generate, time_branch, time_mature;
    // ID of the three-dimensional finite element (containing this vascular node)
    dof_id_type elem3d_id;
    // length of vascular segment (beginning from foregoing branch/bifurcation) and from parent vessel
    Real length, total_length;
    // blood vessel radius and thickness at the node, mean pore radius and pores fraction of vessel
    Real radius, thickness, pore_radius, pores_fraction;
    // ID of the corresponding network branch and ID of parent branch
    unsigned int branch_id, parent_branch_id;
    // full list of vascular elements (segments) associated to this node
    std::set<dof_id_type> vsc_elem;
    // pressure loading ratio, critical load for collapse, and scale of compression (stretch ratio)
    Real press_ratio, press_ratio_cr, compression_scale;
};
//=================================================================================================
//
//
//
//=================================================================================================
inline
VascularInfo::VascularInfo () {
    this->level = 0;
    // set default values for type of flow, vessel type and condition
    this->flow = VascularInfo::INVALID_FLOW;
    this->type = VascularInfo::INVALID_TYPE;
    this->status = VascularInfo::NORMAL;
    //
    this->time_generate =
    this->time_branch   =
    this->time_mature   = 0.0;
    //
    this->elem3d_id = DofObject::invalid_id;
    //
    this->length       =
    this->total_length = 0.0;
    //
    this->radius         =
    this->thickness      =
    this->pore_radius    =
    this->pores_fraction = 0.0;
    //
    this->branch_id        =
    this->parent_branch_id = 0;
    //
    this->press_ratio = 0.0;
    this->press_ratio_cr = 1.0;
    this->compression_scale = 1.0;
}
//=================================================================================================
#endif // __UTILS__VASCULAR_INFO_H__
//=================================================================================================
