/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#ifndef __UTILS_H__
#define __UTILS_H__
//=================================================================================================
#include <iostream>
#include <vector>
#include <array>
#include <map>
#include <set>
#include <algorithm>
#include <functional>
#include <random>
#include <unistd.h>
#include <stdio.h>
#include <sys/stat.h>
#include "feb3/auxiliary.h"
#include "libmesh/point.h"
#include "libmesh/getpot.h"
#include "libmesh/parallel.h"
#include "libmesh/vector_value.h"
#include "libmesh/numeric_vector.h"
#include "libmesh/elem.h"
#include "libmesh/point.h"
#include "libmesh/tensor_value.h"
#include "libmesh/system.h"
#ifdef BIODYNAMO_ENABLED
#include "feb3/bdm_interface.h"
#include "feb3/bdm_migrating_cell.h"
#endif
#include "gsl/gsl_math.h"
#include "gsl/gsl_errno.h"
#include "gsl/gsl_poly.h"
#include "gsl/gsl_eigen.h"
//=================================================================================================
// forward class declarations
template<class X> class Dual;
template<class X> class Triple;
template<class X, int N> class Vector;
template<class X, int N> class Dyadic;
template<class X, int N> class Triadic;
template<class X, int N> class Quindic;
template<class X, int N> class Sextic;
class Vector3DGradient;
//-------------------------------------------------------------------------------------------------
// compact type definitions
typedef std::vector<char>   char_array;
typedef std::array<double, 3>  RealVector3d;
typedef Vector<Real, VOIGT_SP3D>  VectorVoigt3D;
typedef Dyadic<Real, VOIGT_SP3D>  DyadicVoigt3D;
typedef Vector<Real, SP3D>        Vector3D;
typedef Dyadic<Real, SP3D>        Dyadic3D;
typedef Triadic<Real, SP3D>       Triadic3D;
typedef Quindic<Real, SP3D>       Quindic3D;
typedef Sextic<Real, SP3D>        Sextic3D;
typedef Vector<Real, VOIGT_SP2D>  VectorVoigt2D;
typedef Dyadic<Real, VOIGT_SP2D>  DyadicVoigt2D;
typedef Vector<Real, SP2D>        Vector2D;
typedef Dyadic<Real, SP2D>        Dyadic2D;
typedef Triadic<Real, SP2D>       Triadic2D;
typedef Quindic<Real, SP2D>       Quindic2D;
typedef Sextic<Real, SP2D>        Sextic2D;
typedef blitz::Array<Point, 1>          Blitz_VectorPoint;
typedef blitz::Array<Real, 1>           Blitz_VectorReal;
typedef blitz::Array<Real, 2>           Blitz_DyadicReal;
typedef blitz::Array<RealGradient, 1>   Blitz_VectorGradient;
typedef blitz::Array<RealGradient, 2>   Blitz_DyadicGradient;
typedef blitz::Array<int, 1>            Blitz_VectorInt;
typedef blitz::Array<int, 2>            Blitz_DyadicInt;
typedef Dual<unsigned int>      DualUInt;
typedef Triple<unsigned int>    TripleUInt;
typedef std::set<unsigned int>  SetUInt;
typedef TensorValue<Real>       Tensor2o;
typedef VectorValue<Real>       Displacement;
typedef VectorValue<Real>       Velocity;
typedef Vector3DGradient        DisplacementGradient;
typedef Vector3DGradient        VelocityGradient;
typedef dof_id_type             material_id_type;
typedef subdomain_id_type       region_id_type;
typedef std::pair<dof_id_type, dof_id_type>  pair_dof_id_type;
//-------------------------------------------------------------------------------------------------
// forward function declarations
void solve_polynom (const std::vector<Real>&, std::vector<Real>&, const Real);
void inverse (const Blitz_DyadicReal&, Blitz_DyadicReal&, const Real);
template<int N>
void inverse (const Dyadic<Real, N>&,  Dyadic<Real, N>&,  const Real);
template<int N>
void eigen_solve (const Dyadic<Real, N>&, std::vector<Real>&, std::vector< Vector<Real, N> >&);
//=================================================================================================
enum ProblemCase {
    NO_SOLVER                =    0,
    FEM_BIOMECHANICS_SOLVER  = 1001, IGA_BIOMECHANICS_SOLVER  = 2001,
    FEM_BIOCHEMICS_SOLVER    = 1002, FEM_INVITRO_TUMOROID_SOLVER = 2002, FEM_RT_TUMOUR_FIBROSIS_SOLVER = 3002,
    FEM_WOUND_HEALING_SOLVER       = 1003,
    FEM_TUMOUR_ANGIOGENESIS_SOLVER = 1004
};
//=================================================================================================
#include "feb3/blitz_tensors.h"
#include "feb3/binomial_basis.h"
#include "feb3/utils-strings.h"
#include "feb3/utils-some_pairs.h"
#include "feb3/utils-miscela.h"
#include "feb3/utils-algebra_2d.h"
#include "feb3/utils-algebra_3d.h"
#include "feb3/utils-algebra_more.h"
#include "feb3/utils-stress_strain.h"
#include "feb3/utils-boundary_condition_info.h"
#include "feb3/utils-loading_info.h"
#include "feb3/utils-field_info.h"
#include "feb3/utils-material_info.h"
#include "feb3/utils-region_info.h"
#include "feb3/utils-vascular_info.h"
#include "feb3/utils-invariants.h"
#include "feb3/time_functions.h"
//=================================================================================================
#endif // __UTILS_H__
//=================================================================================================
