/*************************************************************************************************

    Copyright (C) 2009-2019    Jean de Montigny (jean.demontigny@hotmail.fr)
                               Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#ifndef __PARAMS_H__
#define __PARAMS_H__
//=================================================================================================
#include "feb3/utils.h"
#include "libmesh/parameters.h"
#include "libmesh/getpot.h"
//=================================================================================================
template <class X>
inline
X get_param (GetPot& in, std::set<std::string>& in_p, const std::string& name) {
  X par;
  if ( in.have_variable(name) ) {
    in_p.insert(name);
    par = in(name, X());
  }
  else {
    std::cout << " ### could not retrieve variable \"" << name << "\" ###" << std::endl;
    libmesh_error();
  }
  return par;
}
template <class X>
inline
X get_param (GetPot& in, std::set<std::string>& in_p, const char* name) {
  return get_param<X>(in, in_p, std::string(name));
}
//-------------------------------------------------------------------------------------------------
template <class X>
inline
X get_param (GetPot& in, std::set<std::string>& in_p, const std::string& name, const X def) {
  X par;
  if ( in.have_variable(name) ) {
    in_p.insert(name);
    par = in(name, def);
  }
  else {
    std::cout << " *** could not retrieve variable \"" << name << "\", using the default value \"" << def << "\" ***" << std::endl;
    par = def;
  }
  return par;
}
template <class X>
inline
X get_param (GetPot& in, std::set<std::string>& in_p, const char* name, const X def) {
  return get_param<X>(in, in_p, std::string(name), def);
}
//-------------------------------------------------------------------------------------------------
template <class X>
inline
unsigned int get_param_size (GetPot& in, std::set<std::string>& in_p, const std::string& name) {
  unsigned int n;
  if ( in.have_variable(name) ) {
    in_p.insert(name);
    n = in.vector_variable_size(name);
  }
  else {
    std::cout << " ### could not retrieve variable \"" << name << "\" ###" << std::endl;
    libmesh_error();
  }
  return n;
}
template <class X>
inline
unsigned int get_param_size (GetPot& in, std::set<std::string>& in_p, const char* name) {
  return get_param_size<X>(in, in_p, std::string(name));
}
//-------------------------------------------------------------------------------------------------
template <class X>
inline
X get_param_comp (GetPot& in, std::set<std::string>& in_p, const std::string& name, unsigned int index) {
  X par;
  if ( in.have_variable(name) ) {
    in_p.insert(name);
    par = in(name, X(), index);
  } else {
    std::cout << " ### could not retrieve variable \"" << name << "\" ###" << std::endl;
    libmesh_error();
  }
  return par;
}
template <class X>
inline
X get_param_comp (GetPot& in, std::set<std::string>& in_p, const char* name, unsigned int index) {
  return get_param_comp<X>(in, in_p, std::string(name), index);
}
//=================================================================================================
#endif
//=================================================================================================
