/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/auxiliary.h"
#include "feb3/fem-context.h"
#include "libmesh/elem.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
//=================================================================================================
void FEM_Context::get_var_list (std::vector<unsigned short int>& vars) const {
    const unsigned int nvars = this->_system->n_vars();
    //
    vars.resize(nvars);
    for (unsigned int ivar=0; ivar<nvars; ivar++) {
        const std::string& name = this->_system->variable_name(ivar);
        unsigned short int vrbl = this->_system->variable_number(name);
        vars[ivar] = vrbl;
    }
}
//=================================================================================================
void FEM_Context::get_dof_indices ( const Elem* elem,
                                    std::vector<dof_id_type>& dofs,
                                    std::vector<dof_id_type>* dofs_var,
                                    std::vector<unsigned int>* vn ) const {
    libmesh_assert( elem );
    libmesh_assert( dofs_var );

    if ( vn ) {
        const unsigned int nvars = vn->size();
        //const unsigned int nnode = elem->n_nodes();
        //
        const DofMap& dof_map = this->_system->get_dof_map();
        //
        dofs.clear();
        for (unsigned int jvar=0; jvar<nvars; jvar++) {
            const unsigned int ivar = vn->at(jvar);
            dof_map.dof_indices(elem, dofs_var[jvar], ivar);
            for (unsigned int i=0; i<dofs_var[jvar].size(); i++) {
                dofs.push_back( dofs_var[jvar][i] );
            }
        }
    } else {
        const unsigned int nvars = this->_system->n_vars();
        //const unsigned int nnode = elem->n_nodes();
        //
        std::vector<unsigned short int> vars;
        this->get_var_list(vars);
        //
        const DofMap& dof_map = this->_system->get_dof_map();
        //
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            dof_map.dof_indices(elem, dofs_var[ivar], vars[ivar]);
        }
        dof_map.dof_indices(elem, dofs);
    }
}
//=================================================================================================
