/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/ida.h"
#include <numeric>
//=================================================================================================
void InverseDistanceAlgorithm::init (const std::vector<Point>& pnt) {
//     libmesh_assert( pnt.size() < 3 );
    // all are initialized now
    this->_is_init = true;
    this->_pnt = &pnt;
}
//=================================================================================================
void InverseDistanceAlgorithm::calc (const Point& apnt, std::vector<Real>& phi) const {
    libmesh_assert( this->_is_init );
    const int npnt = this->_pnt->size();
    // allocate memory
    phi.resize(npnt);
    // evaluate the weights
    std::vector<Real> weights(npnt);
    for (int i=0; i<npnt; i++) {
        const Real ds = Point((*this->_pnt)[i]-apnt).norm();
        weights[i] = pow(ds, -this->_coeff);
    }
    const Real weights_tot = accumulate(weights.begin(), weights.end(), 0.0);
    // normalize the weights
    for (int i=0; i<npnt; i++) {
        phi[i] = weights[i] / weights_tot;
    }
}
//=================================================================================================
void InverseDistanceAlgorithm::calc (const std::vector<Point>& apnt, std::vector< std::vector<Real> >& phi) const {
    libmesh_assert( this->_is_init );
    const int npnt = this->_pnt->size();
    const int nappnts = apnt.size();
    // allocate memory
    phi.resize(npnt);
    for (int i=0; i<npnt; i++) {
        phi[i].resize(nappnts);
    }
    // do the calculations
    for (int ip=0; ip<nappnts; ip++) {
        // evaluate the weights
        std::vector<Real> weights(npnt);
        for (int i=0; i<npnt; i++) {
            const Real ds = Point((*this->_pnt)[i]-apnt[ip]).norm();
            weights[i] = pow(ds, -this->_coeff);
        }
        const Real weights_tot = accumulate(weights.begin(), weights.end(), 0.0);
        // normalize the weights
        for (int i=0; i<npnt; i++) {
            phi[i][ip] = weights[i] / weights_tot;
        }
    }
}
//=================================================================================================
