/*************************************************************************************************

    Copyright (C) 2009-2019    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/utils.h"
//=================================================================================================
#ifdef _LOCAL_ARGUMENTS_
    #error "an unexpected error occurred!"
#else
    #define _LOCAL_ARGUMENTS_ \
        const MaterialInfo& mat,        \
        FieldInfo& fld,                 \
        Blitz_VectorReal& Crd,          \
        Blitz_VectorGradient& Grad_Crd, \
        Blitz_DyadicReal& Ksm,          \
        Blitz_VectorPoint& Velc,        \
        Blitz_DyadicReal& Diff,         \
        Blitz_DyadicGradient& Taxs
#endif
#ifdef _VARIABILITY_1pct_
    #error "an unexpected error occurred!"
#else
    #define _VARIABILITY_1pct_  normal_distro(1., 0.01)
#endif
#ifdef _VARIABILITY_2pct_
    #error "an unexpected error occurred!"
#else
    #define _VARIABILITY_2pct_  normal_distro(1., 0.02)
#endif
#ifdef _VARIABILITY_5pct_
    #error "an unexpected error occurred!"
#else
    #define _VARIABILITY_5pct_  normal_distro(1., 0.05)
#endif
//=================================================================================================
//
//
//
//=================================================================================================
namespace biochemics_reaction_diffusion_solver {
//=================================================================================================
void damaged_region_1312 ( _LOCAL_ARGUMENTS_ ) {
    const unsigned int nvars = 13;
    Crd.resize(nvars);
    Crd = 0.0;
    Grad_Crd.resize(nvars);
    Grad_Crd = RealGradient(0.0, 0.0, 0.0);
    //
    #ifdef BIODYNAMO_ENABLED
    if ( ! fld.bdm_ptr ) libmesh_error();
    #endif
    const Real oxyg = fld.biochem[0],
               cellNc = fld.biochem[1],
               cell1N = fld.biochem[2],
               cell1H = fld.biochem[3],
               cell2N = fld.biochem[4],
               cell2H = fld.biochem[5],
               cell3N = fld.biochem[6],
               cell3H = fld.biochem[7],
               gf1 = fld.biochem[ 8],
               gf2 = fld.biochem[ 9],
               gf3 = fld.biochem[10],
               mmp = fld.biochem[11],
               ecm = fld.biochem[12];
    const RealGradient Grad_oxyg = fld.grad_biochem[0],
                       Grad_cellNc = fld.grad_biochem[1],
                       Grad_cell1N = fld.grad_biochem[2],
                       Grad_cell1H = fld.grad_biochem[3],
                       Grad_cell2N = fld.grad_biochem[4],
                       Grad_cell2H = fld.grad_biochem[5],
                       Grad_cell3N = fld.grad_biochem[6],
                       Grad_cell3H = fld.grad_biochem[7],
                       Grad_gf1 = fld.grad_biochem[ 8],
                       Grad_gf2 = fld.grad_biochem[ 9],
                       Grad_gf3 = fld.grad_biochem[10],
                       Grad_mmp = fld.grad_biochem[11],
                       Grad_ecm = fld.grad_biochem[12];
    #ifdef BIODYNAMO_ENABLED
    const Real Rate_cellNc = fld.bdm_ptr->cells_population_increment[0]/fld.time_step,
               Rate_cell1N = fld.bdm_ptr->cells_population_increment[1]/fld.time_step,
               Rate_cell1H = fld.bdm_ptr->cells_population_increment[2]/fld.time_step,
               Rate_cell2N = fld.bdm_ptr->cells_population_increment[3]/fld.time_step,
               Rate_cell2H = fld.bdm_ptr->cells_population_increment[4]/fld.time_step,
               Rate_cell3N = fld.bdm_ptr->cells_population_increment[5]/fld.time_step,
               Rate_cell3H = fld.bdm_ptr->cells_population_increment[6]/fld.time_step;
    const RealGradient Len_Rate_cellNc = RealGradient( fld.bdm_ptr->cell_dynamics[0][0]/fld.time_step ,
                                                       fld.bdm_ptr->cell_dynamics[0][1]/fld.time_step ,
                                                       fld.bdm_ptr->cell_dynamics[0][2]/fld.time_step ),
                       Len_Rate_cell1N = RealGradient( fld.bdm_ptr->cell_dynamics[1][0]/fld.time_step ,
                                                       fld.bdm_ptr->cell_dynamics[1][1]/fld.time_step ,
                                                       fld.bdm_ptr->cell_dynamics[1][2]/fld.time_step ),
                       Len_Rate_cell1H = RealGradient( fld.bdm_ptr->cell_dynamics[2][0]/fld.time_step ,
                                                       fld.bdm_ptr->cell_dynamics[2][1]/fld.time_step ,
                                                       fld.bdm_ptr->cell_dynamics[2][2]/fld.time_step ),
                       Len_Rate_cell2N = RealGradient( fld.bdm_ptr->cell_dynamics[3][0]/fld.time_step ,
                                                       fld.bdm_ptr->cell_dynamics[3][1]/fld.time_step ,
                                                       fld.bdm_ptr->cell_dynamics[3][2]/fld.time_step ),
                       Len_Rate_cell2H = RealGradient( fld.bdm_ptr->cell_dynamics[4][0]/fld.time_step ,
                                                       fld.bdm_ptr->cell_dynamics[4][1]/fld.time_step ,
                                                       fld.bdm_ptr->cell_dynamics[4][2]/fld.time_step ),
                       Len_Rate_cell3N = RealGradient( fld.bdm_ptr->cell_dynamics[5][0]/fld.time_step ,
                                                       fld.bdm_ptr->cell_dynamics[5][1]/fld.time_step ,
                                                       fld.bdm_ptr->cell_dynamics[5][2]/fld.time_step ),
                       Len_Rate_cell3H = RealGradient( fld.bdm_ptr->cell_dynamics[6][0]/fld.time_step ,
                                                       fld.bdm_ptr->cell_dynamics[6][1]/fld.time_step ,
                                                       fld.bdm_ptr->cell_dynamics[6][2]/fld.time_step );
    const Real VF_ECM = ecm,
               VF_IF  = mat.get_biochem_param(1),
               VF_Cell= fld.bdm_ptr->cells_volume_fraction,
               // function that modulates oxygenation dynamics with respect to the "vacant" volume fraction
               Ovf = pow(apply_lbound(0.0, 1.0-VF_ECM-VF_IF-VF_Cell), mat.get_biochem_param(7)),
               // function that modulates ECM dynamics with respect to the "vacant" volume fraction
               Hvf = pow(apply_lbound(0.0, 1.0-VF_ECM-VF_IF-VF_Cell), mat.get_biochem_param(8));
    #else
    const Real VF_ECM = ecm,
               VF_IF  = mat.get_biochem_param(1),
               // function that modulates oxygenation dynamics with respect to the "vacant" volume fraction
               Ovf = pow(apply_lbound(0.0, 1.0-VF_ECM-VF_IF), mat.get_biochem_param(7)),
               // function that modulates ECM dynamics with respect to the "vacant" volume fraction
               Hvf = pow(apply_lbound(0.0, 1.0-VF_ECM-VF_IF), mat.get_biochem_param(8));
    #endif
    //
    const Real oxyg_supply = fld.biochem[nvars+0]; // discontinuous field variable
    //
    const Real Drug_dose = fld.auxiliary[0],
               RT_dose = fld.auxiliary[1];
    //
    if ( -1 == fld.a_flag ) return;
    //
    const Point velocity = coordinates(fld.fluid_velocity);
    //
    const Real alpha_oxyg        = mat.get_biochem_param( 9)*_VARIABILITY_1pct_*Ovf*oxyg_supply,
               delta_oxyg_cell1N = mat.get_biochem_param(10)*_VARIABILITY_2pct_*cell1N,
               delta_oxyg_cell1H = mat.get_biochem_param(11)*_VARIABILITY_2pct_*cell1H,
               delta_oxyg_cell2N = mat.get_biochem_param(12)*_VARIABILITY_2pct_*cell2N,
               delta_oxyg_cell2H = mat.get_biochem_param(13)*_VARIABILITY_2pct_*cell2H,
               delta_oxyg_cell3N = mat.get_biochem_param(14)*_VARIABILITY_2pct_*cell3N,
               delta_oxyg_cell3H = mat.get_biochem_param(15)*_VARIABILITY_2pct_*cell3H;
    const Real diff_oxyg         = mat.get_biochem_param(20)*_VARIABILITY_2pct_;
    //
    // oxygen
    Crd(0) = alpha_oxyg
           - delta_oxyg_cell1N - delta_oxyg_cell1H
           - delta_oxyg_cell2N - delta_oxyg_cell2H
           - delta_oxyg_cell3N - delta_oxyg_cell3H
           - velocity * Grad_oxyg;
    Grad_Crd(0) = diff_oxyg * Grad_oxyg;
    //
    // cell-necrotic
    #ifdef BIODYNAMO_ENABLED
    Crd(1) = Rate_cellNc;
    Grad_Crd(1) = Len_Rate_cellNc;
    #endif
    //
    // cell-1-normoxic
    #ifdef BIODYNAMO_ENABLED
    Crd(2) = Rate_cell1N;
    Grad_Crd(2) = Len_Rate_cell1N;
    #endif
    //
    // cell-1-hypoxic
    #ifdef BIODYNAMO_ENABLED
    Crd(3) = Rate_cell1H;
    Grad_Crd(3) = Len_Rate_cell1H;
    #endif
    //
    // cell-2-normoxic
    #ifdef BIODYNAMO_ENABLED
    Crd(4) = Rate_cell2N;
    Grad_Crd(4) = Len_Rate_cell2N;
    #endif
    //
    // cell-2-hypoxic
    #ifdef BIODYNAMO_ENABLED
    Crd(5) = Rate_cell2H;
    Grad_Crd(5) = Len_Rate_cell2H;
    #endif
    //
    // cell-2-normoxic
    #ifdef BIODYNAMO_ENABLED
    Crd(6) = Rate_cell3N;
    Grad_Crd(6) = Len_Rate_cell3N;
    #endif
    //
    // cell-2-hypoxic
    #ifdef BIODYNAMO_ENABLED
    Crd(7) = Rate_cell3H;
    Grad_Crd(7) = Len_Rate_cell3H;
    #endif
    //
    //
    const Real alpha_gf1_cell1N = mat.get_biochem_param(110)*_VARIABILITY_2pct_*cell1N,
               alpha_gf1_cell1H = mat.get_biochem_param(111)*_VARIABILITY_2pct_*cell1H,
               alpha_gf1_cell2N = mat.get_biochem_param(112)*_VARIABILITY_2pct_*cell2N,
               alpha_gf1_cell2H = mat.get_biochem_param(113)*_VARIABILITY_2pct_*cell2H,
               alpha_gf1_cell3N = mat.get_biochem_param(114)*_VARIABILITY_2pct_*cell3N,
               alpha_gf1_cell3H = mat.get_biochem_param(115)*_VARIABILITY_2pct_*cell3H,
               delta_gf1        = mat.get_biochem_param(120)*_VARIABILITY_2pct_*gf1;
    const Real diff_gf1         = mat.get_biochem_param(121)*_VARIABILITY_2pct_;
    //
    // growth factor-1
    Crd(8) = alpha_gf1_cell1N + alpha_gf1_cell1H
           + alpha_gf1_cell2N + alpha_gf1_cell2H
           + alpha_gf1_cell3N + alpha_gf1_cell3H
           - delta_gf1
           - velocity * Grad_gf1;
    Grad_Crd(8) = diff_gf1 * Grad_gf1;
    //
    const Real alpha_gf2_cell1N = mat.get_biochem_param(130)*_VARIABILITY_2pct_*cell1N,
               alpha_gf2_cell1H = mat.get_biochem_param(131)*_VARIABILITY_2pct_*cell1H,
               alpha_gf2_cell2N = mat.get_biochem_param(132)*_VARIABILITY_2pct_*cell2N,
               alpha_gf2_cell2H = mat.get_biochem_param(133)*_VARIABILITY_2pct_*cell2H,
               alpha_gf2_cell3N = mat.get_biochem_param(134)*_VARIABILITY_2pct_*cell3N,
               alpha_gf2_cell3H = mat.get_biochem_param(135)*_VARIABILITY_2pct_*cell3H,
               delta_gf2        = mat.get_biochem_param(140)*_VARIABILITY_2pct_*gf2;
    const Real diff_gf2         = mat.get_biochem_param(141)*_VARIABILITY_2pct_;
    //
    // growth factor-2
    Crd(9) = alpha_gf2_cell1N + alpha_gf2_cell1H
           + alpha_gf2_cell2N + alpha_gf2_cell2H
           + alpha_gf2_cell3N + alpha_gf2_cell3H
           - delta_gf2
           - velocity * Grad_gf2;
    Grad_Crd(9) = diff_gf2 * Grad_gf2;
    //
    const Real alpha_gf3_cell1N = mat.get_biochem_param(150)*_VARIABILITY_2pct_*cell1N,
               alpha_gf3_cell1H = mat.get_biochem_param(151)*_VARIABILITY_2pct_*cell1H,
               alpha_gf3_cell2N = mat.get_biochem_param(152)*_VARIABILITY_2pct_*cell2N,
               alpha_gf3_cell2H = mat.get_biochem_param(153)*_VARIABILITY_2pct_*cell2H,
               alpha_gf3_cell3N = mat.get_biochem_param(154)*_VARIABILITY_2pct_*cell3N,
               alpha_gf3_cell3H = mat.get_biochem_param(155)*_VARIABILITY_2pct_*cell3H,
               delta_gf3        = mat.get_biochem_param(160)*_VARIABILITY_2pct_*gf3;
    const Real diff_gf3         = mat.get_biochem_param(161)*_VARIABILITY_2pct_;
    //
    // growth factor-3
    Crd(10) = alpha_gf3_cell1N + alpha_gf3_cell1H
            + alpha_gf3_cell2N + alpha_gf3_cell2H
            + alpha_gf3_cell3N + alpha_gf3_cell3H
            - delta_gf3
            - velocity * Grad_gf3;
    Grad_Crd(10) = diff_gf3 * Grad_gf3;
    //
    const Real alpha_mmp_cell1N = mat.get_biochem_param(170)*_VARIABILITY_2pct_*cell1N,
               alpha_mmp_cell1H = mat.get_biochem_param(171)*_VARIABILITY_2pct_*cell1H,
               alpha_mmp_cell2N = mat.get_biochem_param(172)*_VARIABILITY_2pct_*cell2N,
               alpha_mmp_cell2H = mat.get_biochem_param(173)*_VARIABILITY_2pct_*cell2H,
               alpha_mmp_cell3N = mat.get_biochem_param(174)*_VARIABILITY_2pct_*cell3N,
               alpha_mmp_cell3H = mat.get_biochem_param(175)*_VARIABILITY_2pct_*cell3H,
               delta_mmp        = mat.get_biochem_param(180)*_VARIABILITY_2pct_*mmp;
    const Real diff_mmp         = mat.get_biochem_param(181)*_VARIABILITY_2pct_;
    //
    // matrix metalloproteinase
    Crd(11) = alpha_mmp_cell1N + alpha_mmp_cell1H
            + alpha_mmp_cell2N + alpha_mmp_cell2H
            + alpha_mmp_cell3N + alpha_mmp_cell3H
            - delta_mmp
            - velocity * Grad_mmp;
    Grad_Crd(11) = diff_mmp * Grad_mmp;
    //
    const Real alpha_ecm_cell1N = mat.get_biochem_param(190)*_VARIABILITY_2pct_*Hvf*cell1N,
               alpha_ecm_cell1H = mat.get_biochem_param(191)*_VARIABILITY_2pct_*Hvf*cell1H,
               alpha_ecm_cell2N = mat.get_biochem_param(192)*_VARIABILITY_2pct_*Hvf*cell2N,
               alpha_ecm_cell2H = mat.get_biochem_param(193)*_VARIABILITY_2pct_*Hvf*cell2H,
               alpha_ecm_cell3N = mat.get_biochem_param(194)*_VARIABILITY_2pct_*Hvf*cell3N,
               alpha_ecm_cell3H = mat.get_biochem_param(195)*_VARIABILITY_2pct_*Hvf*cell3H,
               delta_ecm_mmp    = mat.get_biochem_param(200)*_VARIABILITY_2pct_*ecm*mmp,
               alpha_ecm        = mat.get_biochem_param(201)*_VARIABILITY_1pct_*Hvf;
    //
    // extracellular matrix
    Crd(12) = alpha_ecm_cell1N + alpha_ecm_cell1H
            + alpha_ecm_cell2N + alpha_ecm_cell2H
            + alpha_ecm_cell3N + alpha_ecm_cell3H
            + alpha_ecm
            - delta_ecm_mmp;
    //
}
//=================================================================================================
void healthy_region_1312 ( _LOCAL_ARGUMENTS_ ) {
    biochemics_reaction_diffusion_solver::damaged_region_1312(mat, fld, Crd, Grad_Crd, Ksm, Velc, Diff, Taxs);
}
//=================================================================================================
} // end of namespace biochemics_reaction_diffusion_solver...
//=================================================================================================
//
//
//
//=================================================================================================
#undef _LOCAL_ARGUMENTS_
#undef _VARIABILITY_1pct_
#undef _VARIABILITY_2pct_
#undef _VARIABILITY_5pct_
//=================================================================================================
