/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/utils.h"
//=================================================================================================
#ifdef _LOCAL_ARGUMENTS_
    #error "an unexpected error occurred!"
#else
    #define _LOCAL_ARGUMENTS_ \
        const MaterialInfo& mat,        \
        FieldInfo& fld,                 \
        Blitz_VectorReal& Crd,          \
        Blitz_VectorGradient& Grad_Crd, \
        Blitz_DyadicReal& Ksm,          \
        Blitz_VectorPoint& Velc,        \
        Blitz_DyadicReal& Diff,         \
        Blitz_DyadicGradient& Taxs
#endif
#ifdef _VARIABILITY_1pct_
    #error "an unexpected error occurred!"
#else
    #define _VARIABILITY_1pct_  normal_distro(1., 0.01)
#endif
#ifdef _VARIABILITY_2pct_
    #error "an unexpected error occurred!"
#else
    #define _VARIABILITY_2pct_  normal_distro(1., 0.02)
#endif
#ifdef _VARIABILITY_5pct_
    #error "an unexpected error occurred!"
#else
    #define _VARIABILITY_5pct_  normal_distro(1., 0.05)
#endif
//=================================================================================================
//
//
//
//=================================================================================================
namespace invitro_tumoroid_growth {
//=================================================================================================
void damaged_region_711 ( _LOCAL_ARGUMENTS_ ) {
    const unsigned int nvars = 7;
    Crd.resize(nvars);
    Crd = 0.0;
    Grad_Crd.resize(nvars);
    Grad_Crd = RealGradient(0.0, 0.0, 0.0);
    Ksm.resize(nvars, nvars);
    Ksm = 0.0;
    Velc.resize(nvars);
    Velc = Point(0.0, 0.0, 0.0);
    Diff.resize(nvars, nvars);
    Diff = 0.0;
    Taxs.resize(nvars, nvars);
    Taxs = RealGradient(0.0, 0.0, 0.0);
    //
    const Real nrmx_[] = { apply_bounds(0.0, fld.biochem[0], 1.0) ,
                           apply_bounds(0.0, fld.biochem[1], 1.0) },
               hypx_[] = { apply_bounds(0.0, fld.biochem[2], 1.0) ,
                           apply_bounds(0.0, fld.biochem[3], 1.0) },
               dead    = apply_bounds(0.0, fld.biochem[4], 1.0),
               viable  = apply_bounds(0.0, fld.biochem[0]+fld.biochem[1]+fld.biochem[2]+fld.biochem[3], 1.0),
               host    = apply_bounds(0.0, 1.0-viable-dead, 1.0),
               oxyg    = apply_bounds(0.0, fld.biochem[5], 1.0),
               ecm     = fld.biochem[6];
    const RealGradient Grad_nrmx_[] = { fld.grad_biochem[0] ,
                                        fld.grad_biochem[1] },
                       Grad_hypx_[] = { fld.grad_biochem[2] ,
                                        fld.grad_biochem[3] },
                       Grad_dead    = fld.grad_biochem[4],
                       Grad_oxyg    = fld.grad_biochem[5],
                       Grad_ecm     = fld.grad_biochem[6];
    //
    const Point velocity = coordinates(fld.fluid_velocity);
    //
    // const Real RT_dose = fld.biochem[nvars+0]; // discontinuous field variable
    //
    const Real perf_vasc_dens = fld.auxiliary[0];
    //
    if ( -1 == fld.a_flag ) return;
    //
    const Real vasc = perf_vasc_dens,
               //
               oxyg_h_thres = mat.get_biochem_param(19), // O2 threshold for hypoxia
               oxyg_d_thres = mat.get_biochem_param(20), // O2 threshold for apoptosis
               oxyg_thres   = mat.get_biochem_param(21), // O2 threshold
               //
               tumr_thres = mat.get_biochem_param(22),
               nrmx_thres = mat.get_biochem_param(23),
               hypx_thres = mat.get_biochem_param(24),
               //
               a_exp = mat.get_biochem_param(25),
               H_cell  = pow(host, a_exp),
               dH_cell = -a_exp*pow(host, a_exp-1.0),
               //
               b_exp = mat.get_biochem_param(26),
               F_oxyg  = 1.0-pow(1.0-oxyg, b_exp),
               dF_oxyg = b_exp*pow(1.0-oxyg, b_exp-1.0),
               G_oxyg  = pow(1.0-oxyg, b_exp),
               dG_oxyg = -b_exp*pow(1.0-oxyg, b_exp-1.0),
               H_oxyg  = (1.0-oxyg),
               dH_oxyg = -1.0,
               //
               Diff_nrmx_[] = { mat.get_biochem_param(0) ,
                                mat.get_biochem_param(2) },
                chi_nrmx_[] = { mat.get_biochem_param(1) ,
                                mat.get_biochem_param(3) },
               Diff_hypx_[] = { mat.get_biochem_param(4) ,
                                mat.get_biochem_param(6) },
                chi_hypx_[] = { mat.get_biochem_param(5) ,
                                mat.get_biochem_param(7) },
                     rho_[] = { oxyg > oxyg_h_thres ? mat.get_biochem_param(8) : 0.0 ,
                                oxyg > oxyg_h_thres ? mat.get_biochem_param(9) : 0.0 },
                     alpha_nrmx_[] = { ( oxyg <= oxyg_d_thres && nrmx_[0] > nrmx_thres && dead <= tumr_thres ) ? mat.get_biochem_param(10) : 0.0 ,
                                       ( oxyg <= oxyg_d_thres && nrmx_[1] > nrmx_thres && dead <= tumr_thres ) ? mat.get_biochem_param(10) : 0.0 },
                     alpha_hypx_[] = { ( oxyg <= oxyg_d_thres && hypx_[0] > hypx_thres && dead <= tumr_thres ) ? mat.get_biochem_param(11) : 0.0 ,
                                       ( oxyg <= oxyg_d_thres && hypx_[1] > hypx_thres && dead <= tumr_thres ) ? mat.get_biochem_param(11) : 0.0 },
                     beta_[]       = { ( oxyg <= oxyg_h_thres && nrmx_[0] > nrmx_thres && hypx_[0] <= tumr_thres ) ? mat.get_biochem_param(12) : 0.0 ,
                                       ( oxyg <= oxyg_h_thres && nrmx_[1] > nrmx_thres && hypx_[1] <= tumr_thres ) ? mat.get_biochem_param(12) : 0.0 },
                     gamma_[]      = { (                         hypx_[0] > hypx_thres && nrmx_[0] <= tumr_thres ) ? mat.get_biochem_param(13) : 0.0 ,
                                       (                         hypx_[1] > hypx_thres && nrmx_[1] <= tumr_thres ) ? mat.get_biochem_param(13) : 0.0 },
               //
               Diff_oxyg   = mat.get_biochem_param(14),
               kappa_vasc  = mat.get_biochem_param(15),
               delta_host  = mat.get_biochem_param(16),
               delta_nrmx  = oxyg >= oxyg_thres ? mat.get_biochem_param(17) : 0.0,
               delta_hypx  = oxyg >= oxyg_thres ? mat.get_biochem_param(18) : 0.0;
    //
    Crd(0) = rho_[0]*H_cell*nrmx_[0]*F_oxyg + gamma_[0]*hypx_[0]*F_oxyg - beta_[0]*nrmx_[0]*G_oxyg - alpha_nrmx_[0]*nrmx_[0]*H_oxyg;
    Crd(1) = rho_[1]*H_cell*nrmx_[1]*F_oxyg + gamma_[1]*hypx_[1]*F_oxyg - beta_[1]*nrmx_[1]*G_oxyg - alpha_nrmx_[1]*nrmx_[1]*H_oxyg;
    Crd(2) =                                - gamma_[0]*hypx_[0]*F_oxyg + beta_[0]*nrmx_[0]*G_oxyg - alpha_hypx_[0]*hypx_[0]*H_oxyg;
    Crd(3) =                                - gamma_[1]*hypx_[1]*F_oxyg + beta_[1]*nrmx_[1]*G_oxyg - alpha_hypx_[1]*hypx_[1]*H_oxyg;
    Crd(4) = alpha_nrmx_[0]*nrmx_[0]*H_oxyg + alpha_nrmx_[1]*nrmx_[1]*H_oxyg + alpha_hypx_[0]*hypx_[0]*H_oxyg + alpha_hypx_[1]*hypx_[1]*H_oxyg;
    Crd(5) = kappa_vasc*vasc - delta_host*host - delta_nrmx*(nrmx_[0]+nrmx_[1]) - delta_hypx*(hypx_[0]+hypx_[1]) - velocity*Grad_oxyg;
    //
    Grad_Crd(0) = Diff_nrmx_[0]*H_cell*Grad_nrmx_[0] - chi_nrmx_[0]*H_cell*nrmx_[0]*Grad_ecm;
    Grad_Crd(1) = Diff_nrmx_[1]*H_cell*Grad_nrmx_[1] - chi_nrmx_[1]*H_cell*nrmx_[1]*Grad_ecm;
    Grad_Crd(2) = Diff_hypx_[0]*H_cell*Grad_hypx_[0] - chi_hypx_[0]*H_cell*hypx_[0]*Grad_ecm;
    Grad_Crd(3) = Diff_hypx_[1]*H_cell*Grad_hypx_[1] - chi_hypx_[1]*H_cell*hypx_[1]*Grad_ecm;
    Grad_Crd(5) = Diff_oxyg*Grad_oxyg;
    //
    Ksm(0,0) = rho_[0]*dH_cell*nrmx_[0]*F_oxyg + rho_[0]*H_cell*F_oxyg - beta_[0]*G_oxyg - alpha_nrmx_[0]*H_oxyg;
    Ksm(0,1) = rho_[0]*dH_cell*nrmx_[0]*F_oxyg;
    Ksm(0,2) = rho_[0]*dH_cell*nrmx_[0]*F_oxyg + gamma_[0]*F_oxyg;
    Ksm(0,3) = rho_[0]*dH_cell*nrmx_[0]*F_oxyg;
    Ksm(0,4) = rho_[0]*dH_cell*nrmx_[0]*F_oxyg;
    Ksm(0,5) = rho_[0]*H_cell*nrmx_[0]*dF_oxyg + gamma_[0]*hypx_[0]*dF_oxyg - beta_[0]*nrmx_[0]*dG_oxyg - alpha_nrmx_[0]*nrmx_[0]*dH_oxyg;
    //
    Ksm(1,0) = rho_[1]*dH_cell*nrmx_[1]*F_oxyg;
    Ksm(1,1) = rho_[1]*dH_cell*nrmx_[1]*F_oxyg + rho_[1]*H_cell*F_oxyg - beta_[1]*G_oxyg - alpha_nrmx_[1]*H_oxyg;
    Ksm(1,2) = rho_[1]*dH_cell*nrmx_[1]*F_oxyg;
    Ksm(1,3) = rho_[1]*dH_cell*nrmx_[1]*F_oxyg + gamma_[1]*F_oxyg;
    Ksm(1,4) = rho_[1]*dH_cell*nrmx_[1]*F_oxyg;
    Ksm(1,5) = rho_[1]*H_cell*nrmx_[1]*dF_oxyg + gamma_[1]*hypx_[1]*dF_oxyg - beta_[1]*nrmx_[1]*dG_oxyg - alpha_nrmx_[1]*nrmx_[1]*dH_oxyg;
    //
    Ksm(2,0) = beta_[0]*G_oxyg;
    Ksm(2,2) = - gamma_[0]*F_oxyg - alpha_hypx_[0]*H_oxyg;
    Ksm(2,5) = - gamma_[0]*hypx_[0]*dF_oxyg + beta_[0]*nrmx_[0]*dG_oxyg - alpha_hypx_[0]*hypx_[0]*dH_oxyg;
    //
    Ksm(3,1) = beta_[1]*G_oxyg;
    Ksm(3,3) = - gamma_[1]*F_oxyg - alpha_hypx_[1]*H_oxyg;
    Ksm(3,5) = - gamma_[1]*hypx_[1]*dF_oxyg + beta_[1]*nrmx_[1]*dG_oxyg - alpha_hypx_[1]*hypx_[1]*dH_oxyg;
    //
    Ksm(4,0) = alpha_nrmx_[0]*H_oxyg;
    Ksm(4,1) = alpha_nrmx_[1]*H_oxyg;
    Ksm(4,2) = alpha_hypx_[0]*H_oxyg;
    Ksm(4,3) = alpha_hypx_[1]*H_oxyg;
    Ksm(4,5) = alpha_nrmx_[0]*nrmx_[0]*dH_oxyg + alpha_nrmx_[1]*nrmx_[1]*dH_oxyg + alpha_hypx_[0]*hypx_[0]*dH_oxyg + alpha_hypx_[1]*hypx_[1]*dH_oxyg;
    //
    Ksm(5,0) = delta_host - delta_nrmx;
    Ksm(5,1) = delta_host - delta_nrmx;
    Ksm(5,2) = delta_host - delta_hypx;
    Ksm(5,3) = delta_host - delta_hypx;
    Ksm(5,4) = delta_host;
    //
    Velc(5) = velocity;
    //
    Diff(0,0) = Diff_nrmx_[0]*H_cell;
    Diff(1,1) = Diff_nrmx_[1]*H_cell;
    Diff(2,2) = Diff_hypx_[0]*H_cell;
    Diff(3,3) = Diff_hypx_[1]*H_cell;
    Diff(5,5) = Diff_oxyg;
    //
    Taxs(0,0) = Diff_nrmx_[0]*dH_cell*Grad_nrmx_[0] - chi_nrmx_[0]*dH_cell*nrmx_[0]*Grad_ecm - chi_nrmx_[0]*H_cell*Grad_ecm;
    Taxs(0,1) = Diff_nrmx_[0]*dH_cell*Grad_nrmx_[0];
    Taxs(0,2) = Diff_nrmx_[0]*dH_cell*Grad_nrmx_[0];
    Taxs(0,3) = Diff_nrmx_[0]*dH_cell*Grad_nrmx_[0];
    Taxs(0,4) = Diff_nrmx_[0]*dH_cell*Grad_nrmx_[0];
    //
    Taxs(1,0) = Diff_nrmx_[1]*dH_cell*Grad_nrmx_[1];
    Taxs(1,1) = Diff_nrmx_[1]*dH_cell*Grad_nrmx_[1] - chi_nrmx_[1]*dH_cell*nrmx_[1]*Grad_ecm - chi_nrmx_[1]*H_cell*Grad_ecm;
    Taxs(1,2) = Diff_nrmx_[1]*dH_cell*Grad_nrmx_[1];
    Taxs(1,3) = Diff_nrmx_[1]*dH_cell*Grad_nrmx_[1];
    Taxs(1,4) = Diff_nrmx_[1]*dH_cell*Grad_nrmx_[1];
    //
    Taxs(2,0) = Diff_hypx_[0]*dH_cell*Grad_hypx_[0];
    Taxs(2,1) = Diff_hypx_[0]*dH_cell*Grad_hypx_[0];
    Taxs(2,2) = Diff_hypx_[0]*dH_cell*Grad_hypx_[0] - chi_hypx_[0]*dH_cell*hypx_[0]*Grad_ecm - chi_hypx_[0]*H_cell*Grad_ecm;
    Taxs(2,3) = Diff_hypx_[0]*dH_cell*Grad_hypx_[0];
    Taxs(2,4) = Diff_hypx_[0]*dH_cell*Grad_hypx_[0];
    //
    Taxs(3,0) = Diff_hypx_[1]*dH_cell*Grad_hypx_[1];
    Taxs(3,1) = Diff_hypx_[1]*dH_cell*Grad_hypx_[1];
    Taxs(3,2) = Diff_hypx_[1]*dH_cell*Grad_hypx_[1];
    Taxs(3,3) = Diff_hypx_[1]*dH_cell*Grad_hypx_[1] - chi_hypx_[1]*dH_cell*hypx_[1]*Grad_ecm - chi_hypx_[1]*H_cell*Grad_ecm;
    Taxs(3,4) = Diff_hypx_[1]*dH_cell*Grad_hypx_[1];
}
//=================================================================================================
void healthy_region_711 ( _LOCAL_ARGUMENTS_ ) {
    invitro_tumoroid_growth::damaged_region_711(mat, fld, Crd, Grad_Crd, Ksm, Velc, Diff, Taxs);
}
//=================================================================================================
void damaged_region_813 ( _LOCAL_ARGUMENTS_ ) {
    const unsigned int nvars = 8;
    Crd.resize(nvars);
    Crd = 0.0;
    Grad_Crd.resize(nvars);
    Grad_Crd = RealGradient(0.0, 0.0, 0.0);
    Ksm.resize(nvars, nvars);
    Ksm = 0.0;
    Velc.resize(nvars);
    Velc = Point(0.0, 0.0, 0.0);
    Diff.resize(nvars, nvars);
    Diff = 0.0;
    Taxs.resize(nvars, nvars);
    Taxs = RealGradient(0.0, 0.0, 0.0);
    //
    const Real coll  = fld.biochem[0],
               xlink = fld.biochem[1],
               fibrn = fld.biochem[2],
               oxyg  = apply_lbound(0.0, fld.biochem[3]),
               mde   = fld.biochem[4],
               fblas = fld.biochem[5],
               normt = fld.biochem[6],
               hypxt = fld.biochem[7];
    const RealGradient //Grad_coll  = fld.grad_biochem[0],
                       //Grad_xlink = fld.grad_biochem[1],
                       //Grad_fibrn = fld.grad_biochem[2],
                       Grad_oxyg  = fld.grad_biochem[3],
                       Grad_mde   = fld.grad_biochem[4],
                       Grad_fblas = fld.grad_biochem[5],
                       Grad_normt = fld.grad_biochem[6],
                       Grad_hypxt = fld.grad_biochem[7];
    //
    const Real RT_dose = fld.biochem[nvars+0]; // discontinuous field variable
    //
    RealGradient Fibre_coll = RealGradient( fld.auxiliary[0] ,
                                            fld.auxiliary[1] ,
                                            fld.auxiliary[2] ).unit();
    if ( Grad_oxyg*Fibre_coll < 0.0 ) Fibre_coll *= -1.0;
    //
    if ( -1 == fld.a_flag ) return;
    //
    const Real RVE_size   = mat.get_biochem_param(0),
               fblas_size = mat.get_biochem_param(1),
               normt_size = mat.get_biochem_param(2),
               hypxt_size = mat.get_biochem_param(3);
    //
    const Real size = fblas_size * (fblas>0.0 ? fblas : 0.0)
                    + normt_size * (normt>0.0 ? normt : 0.0)
                    + hypxt_size * (hypxt>0.0 ? hypxt : 0.0),
               total = 1.0 - size / RVE_size;
    //
    const Real delta_xlink = mde>0.0 ? mat.get_biochem_param(20) : 0.0;
    //
    const Real delta_fibrn        = mde>0.0 ? mat.get_biochem_param(30) : 0.0;
    const Real lambda_fibrn_fblas = mat.get_biochem_param(31),
               lambda_fibrn_normt = mat.get_biochem_param(32),
               lambda_fibrn_hypxt = mat.get_biochem_param(33);
    //
    const Real oxyg_supply = mat.get_biochem_param(50);
    const Real delta_oxyg_fblas = mat.get_biochem_param(51),
               delta_oxyg_normt = mat.get_biochem_param(52),
               delta_oxyg_hypxt = mat.get_biochem_param(53);
    const Real diff_oxyg = mat.get_biochem_param(60);
    //
    const Real delta_mde = mat.get_biochem_param(70);
    const Real lambda_mde_fblas = mat.get_biochem_param(71),
               lambda_mde_normt = mat.get_biochem_param(72),
               lambda_mde_hypxt = mat.get_biochem_param(73);
    const Real diff_mde = mat.get_biochem_param(80);
    //
    const Real diff_fblas = mat.get_biochem_param(90)
                          * total,
               alpha_fblas = mat.get_biochem_param(91)
                           * exp(-mat.get_biochem_param(92)*oxyg)
                           * (RT_dose>0.0 ? mat.get_biochem_param(93) : 1.0);
    const Real rho_fblas   = mat.get_biochem_param(100)
                           * (0.0<oxyg && oxyg<1.0 ? (1.0-pow(1.0-oxyg, mat.get_biochem_param(101))) : 1.0)
                           * total;
    //
    const Real diff_normt  = mat.get_biochem_param(110)
                           * total,
               alpha_normt = mat.get_biochem_param(111)
                           * exp(-mat.get_biochem_param(112)*oxyg)
                           * (RT_dose>0.0 ? mat.get_biochem_param(113) : 1.0);
    const Real diff_hypxt  = mat.get_biochem_param(120)
                           * total,
               alpha_hypxt = mat.get_biochem_param(121)
                           * exp(-mat.get_biochem_param(122)*oxyg)
                           * (RT_dose>0.0 ? mat.get_biochem_param(123) : 1.0);
    const Real rho_normt   = mat.get_biochem_param(130)
                           * (0.0<oxyg && oxyg<1.0 ? (1.0-pow(1.0-oxyg, mat.get_biochem_param(131))) : 1.0)
                           * total,
               beta_normt  = mat.get_biochem_param(132)
                           * pow(1.0-oxyg, mat.get_biochem_param(133)),
               gamma_hypxt = mat.get_biochem_param(134)
                           * (0.0<oxyg && oxyg<1.0 ? (1.0-pow(1.0-oxyg, mat.get_biochem_param(135))) : 1.0)
                           * total,
               hapto_normt = mat.get_biochem_param(136)
                           * (fibrn>0.0 ? pow(fibrn, mat.get_biochem_param(137)) : 0.0)
                           * (coll>0.0 ? (1.0 - exp(-mat.get_biochem_param(138)*coll)) : 0.0);
    //
    // collagen
    Crd(0) = 0.0;
    Grad_Crd(0) = RealGradient(0.0, 0.0, 0.0);
    //
    // collagen cross-links
    Crd(1) = - delta_xlink * mde * xlink;
    //
    // fibronectin
    Crd(2) = - delta_fibrn * mde * fibrn
             + lambda_fibrn_fblas * fblas
             + lambda_fibrn_normt * normt
             + lambda_fibrn_hypxt * hypxt;
    //
    // oxygen
    Crd(3) =   oxyg_supply
             - delta_oxyg_fblas * fblas
             - delta_oxyg_normt * normt
             - delta_oxyg_hypxt * hypxt;
    Grad_Crd(3) = diff_oxyg * Grad_oxyg;
    //
    // matrix degrading enzymes
    Crd(4) = - delta_mde * mde
             + lambda_mde_fblas * fblas
             + lambda_mde_normt * normt
             + lambda_mde_hypxt * hypxt;
    Grad_Crd(4) = diff_mde * Grad_mde;
    //
    // fibroblasts
    Crd(5) = - alpha_fblas * fblas
             + rho_fblas * fblas;
    Grad_Crd(5) = diff_fblas * Grad_fblas;
    //
    // normoxic tumor cells
    Crd(6) = - alpha_normt * normt
             - beta_normt  * normt
             + gamma_hypxt * hypxt
             + rho_normt * normt;
    Grad_Crd(6) = diff_normt * Grad_normt
                + hapto_normt * Fibre_coll;
    //
    // hypoxic tumor cells
    Crd(7) = - alpha_hypxt * hypxt
             + beta_normt  * normt
             - gamma_hypxt * hypxt;
    Grad_Crd(7) = diff_hypxt * Grad_hypxt;
}
//=================================================================================================
void healthy_region_813 ( _LOCAL_ARGUMENTS_ ) {
    invitro_tumoroid_growth::damaged_region_813(mat, fld, Crd, Grad_Crd, Ksm, Velc, Diff, Taxs);
}
//=================================================================================================
void damaged_region_1115 ( _LOCAL_ARGUMENTS_ ) {
    const unsigned int nvars = 11;
    Crd.resize(nvars);
    Crd = 0.0;
    Grad_Crd.resize(nvars);
    Grad_Crd = RealGradient(0.0, 0.0, 0.0);
    Ksm.resize(nvars, nvars);
    Ksm = 0.0;
    Velc.resize(nvars);
    Velc = Point(0.0, 0.0, 0.0);
    Diff.resize(nvars, nvars);
    Diff = 0.0;
    Taxs.resize(nvars, nvars);
    Taxs = RealGradient(0.0, 0.0, 0.0);
    //
    const Real RVE = mat.get_biochem_param(0),
               fblas_V = mat.get_biochem_param(1),
               normt_V = mat.get_biochem_param(2),
               hypxt_V = mat.get_biochem_param(3),
               endth_V = mat.get_biochem_param(4),
               fblas_Th = mat.get_biochem_param(11),
               normt_Th = mat.get_biochem_param(12),
               hypxt_Th = mat.get_biochem_param(13),
               endth_Th = mat.get_biochem_param(14),
               collg_Th = mat.get_biochem_param(300),
               //lamin_Th = mat.get_biochem_param(301),
               //fibrn_Th = mat.get_biochem_param(302),
               //hyalu_Th = mat.get_biochem_param(303),
               ecm_V_Th = mat.get_biochem_param(310);
    //
    const Real collg = fld.biochem[0],
               lamin = fld.biochem[1],
               fibrn = fld.biochem[2],
               hyalu = fld.biochem[3],
               oxyg  = apply_lbound(0.0, fld.biochem[4]),
               fblas = fld.biochem[5],
               normt = fld.biochem[6],
               hypxt = fld.biochem[7],
               endth = fld.biochem[8],
               angf  = fld.biochem[ 9],
               mde   = fld.biochem[10];
    // xlink = fld.biochem[]*(apply_lbound(0.0, coll)/RVE) // collagen X-links density for future extension of the model...
    const RealGradient Grad_collg = fld.grad_biochem[0],
                       Grad_lamin = fld.grad_biochem[1],
                       Grad_fibrn = fld.grad_biochem[2],
                       Grad_hyalu = fld.grad_biochem[3],
                       Grad_oxyg  = fld.grad_biochem[4],
                       Grad_fblas = fld.grad_biochem[5],
                       Grad_normt = fld.grad_biochem[6],
                       Grad_hypxt = fld.grad_biochem[7],
                       Grad_endth = fld.grad_biochem[8],
                       Grad_angf  = fld.grad_biochem[ 9],
                       Grad_mde   = fld.grad_biochem[10];
    //
    const Point velocity = coordinates(fld.fluid_velocity);
    //
    const Real RT_dose = fld.biochem[nvars+0]; // discontinuous field variable
    //
    const Point fibre = (fld.deform_gradient
                        *Point(fld.auxiliary[0], fld.auxiliary[1], fld.auxiliary[2])).unit();
    //
    const int alpha = mat.get_biochem_param(311) > 0.0
                    ? static_cast<int>(mat.get_biochem_param(311)) : 2;
    const Real beta = mat.get_biochem_param(312);
    if ( beta < 1.0e-12 ) libmesh_error();
    const Real feedback_pressure = fld.pressure > -beta
                                 ? (1.0+fld.pressure/beta) : 0.0;
    //
    const Real effective_V = fld.pressure > 0.0
                           ? RVE : RVE*feedback_pressure;
    //
    const Real cell_V = (fblas>0 ? fblas : 0.0)*fblas_V
                      + (normt>0 ? normt : 0.0)*normt_V
                      + (hypxt>0 ? hypxt : 0.0)*hypxt_V
                      + (endth>0 ? endth : 0.0)*endth_V,
               cell_V_norm = apply_bounds(0.0, cell_V/effective_V, 1.0),
               ecm_V_norm = pow(1.0-cell_V_norm, alpha);
    //
    fld.auxiliary_rate[3] = 0.0;
    fld.auxiliary_rate[4] = fld.current_time>0.0
                          ? (cell_V-fld.auxiliary[4])/fld.time_step : 0.0;
    fld.auxiliary[3] = cell_V;
    fld.auxiliary[4] = cell_V;
    //
    if ( -1 == fld.a_flag ) return;
    //
    // const Point sigma = ( fibre*Grad_collg<0.0 ? -1.0 : +1.0 ) * fibre;
    const Point eta   = ( fibre*Grad_oxyg<0.0  ? -1.0 : +1.0 ) * fibre;
    const Point zeta  = ( fibre*Grad_fblas<0.0 ? -1.0 : +1.0 ) * fibre;
    const Point theta = ( fibre*Grad_endth<0.0 ? -1.0 : +1.0 ) * fibre;
    const Point epsilon = coordinates(unit(max_eigen(fld.strain)));
    //
    // collagen
    {
      const Real delta_collg_mde = (collg>0 && mde>0 ? mat.get_biochem_param(20) : 0.0),
                 lambda_collg_fblas = (fblas>fblas_Th ? mat.get_biochem_param(21) : 0.0)
                                    * (collg>=0 && collg<=mat.get_biochem_param(22) ? (1.0-collg/mat.get_biochem_param(22)) : 0.0),
                 lambda_collg_normt = (normt>normt_Th ? mat.get_biochem_param(23) : 0.0)
                                    * (collg>=0 && collg<=mat.get_biochem_param(24) ? (1.0-collg/mat.get_biochem_param(24)) : 0.0),
                 lambda_collg_hypxt = (hypxt>hypxt_Th ? mat.get_biochem_param(25) : 0.0)
                                    * (collg>=0 && collg<=mat.get_biochem_param(26) ? (1.0-collg/mat.get_biochem_param(26)) : 0.0);
      //
      Crd(0) = - delta_collg_mde * collg * mde
               + lambda_collg_fblas * fblas
               + lambda_collg_normt * normt
               + lambda_collg_hypxt * hypxt;
      Grad_Crd(0) = RealGradient(0.0, 0.0, 0.0);
    }
    //
    // laminin
    {
      const Real delta_lamin_mde = (lamin>0 && mde>0 ? mat.get_biochem_param(30) : 0.0),
                 lambda_lamin_fblas = (fblas>fblas_Th ? mat.get_biochem_param(31) : 0.0)
                                    * (lamin>=0 && lamin<=mat.get_biochem_param(32) ? (1.0-lamin/mat.get_biochem_param(32)) : 0.0),
                 lambda_lamin_normt = (normt>normt_Th ? mat.get_biochem_param(33) : 0.0)
                                    * (lamin>=0 && lamin<=mat.get_biochem_param(34) ? (1.0-lamin/mat.get_biochem_param(34)) : 0.0),
                 lambda_lamin_hypxt = (hypxt>hypxt_Th ? mat.get_biochem_param(35) : 0.0)
                                    * (lamin>=0 && lamin<=mat.get_biochem_param(36) ? (1.0-lamin/mat.get_biochem_param(36)) : 0.0);
      //
      Crd(1) = - delta_lamin_mde * lamin * mde
               + lambda_lamin_fblas * fblas
               + lambda_lamin_normt * normt
               + lambda_lamin_hypxt * hypxt;
      Grad_Crd(1) = RealGradient(0.0, 0.0, 0.0);
    }
    //
    // fibronectin
    {
      const Real delta_fibrn_mde    = (fibrn>0 && mde>0 ? mat.get_biochem_param(40) : 0.0),
                 lambda_fibrn_fblas = (fblas>fblas_Th ? mat.get_biochem_param(41) : 0.0)
                                    * (fibrn>=0 && fibrn<=mat.get_biochem_param(42) ? (1.0-fibrn/mat.get_biochem_param(42)) : 0.0),
                 lambda_fibrn_normt = (normt>normt_Th ? mat.get_biochem_param(43) : 0.0)
                                    * (fibrn>=0 && fibrn<=mat.get_biochem_param(44) ? (1.0-fibrn/mat.get_biochem_param(44)) : 0.0),
                 lambda_fibrn_hypxt = (hypxt>hypxt_Th ? mat.get_biochem_param(45) : 0.0)
                                    * (fibrn>=0 && fibrn<=mat.get_biochem_param(46) ? (1.0-fibrn/mat.get_biochem_param(46)) : 0.0);
      //
      Crd(2) = - delta_fibrn_mde * fibrn * mde
               + lambda_fibrn_fblas * fblas
               + lambda_fibrn_normt * normt
               + lambda_fibrn_hypxt * hypxt;
      Grad_Crd(2) = RealGradient(0.0, 0.0, 0.0);
    }
    //
    // hyaluronan
    {
      const Real delta_hyalu_mde    = (hyalu>0 && mde>0 ? mat.get_biochem_param(50) : 0.0),
                 lambda_hyalu_fblas = (fblas>fblas_Th ? mat.get_biochem_param(51) : 0.0)
                                    * (hyalu>=0 && hyalu<=mat.get_biochem_param(52) ? (1.0-hyalu/mat.get_biochem_param(52)) : 0.0),
                 lambda_hyalu_normt = (normt>normt_Th ? mat.get_biochem_param(53) : 0.0)
                                    * (hyalu>=0 && hyalu<=mat.get_biochem_param(54) ? (1.0-hyalu/mat.get_biochem_param(54)) : 0.0),
                 lambda_hyalu_hypxt = (hypxt>hypxt_Th ? mat.get_biochem_param(55) : 0.0)
                                    * (hyalu>=0 && hyalu<=mat.get_biochem_param(56) ? (1.0-hyalu/mat.get_biochem_param(56)) : 0.0);
      Crd(3) = - delta_hyalu_mde * hyalu * mde
               + lambda_hyalu_fblas * fblas
               + lambda_hyalu_normt * normt
               + lambda_hyalu_hypxt * hypxt;
      Grad_Crd(3) = RealGradient(0.0, 0.0, 0.0);
    }
    //
    // oxygen
    {
      const Real delta_oxyg = (oxyg>0 ? mat.get_biochem_param(60) : 0.0),
                 delta_oxyg_fblas = (oxyg>0 && fblas>fblas_Th ? mat.get_biochem_param(61) : 0.0),
                 delta_oxyg_normt = (oxyg>0 && normt>normt_Th ? mat.get_biochem_param(62) : 0.0),
                 delta_oxyg_hypxt = (oxyg>0 && hypxt>hypxt_Th ? mat.get_biochem_param(63) : 0.0),
                 delta_oxyg_endth = (oxyg>0 && endth>endth_Th ? mat.get_biochem_param(64) : 0.0);
      const Real diff_oxyg = (oxyg>0 ? mat.get_biochem_param(70) : 0.0),
                 oxyg_supply = (oxyg>1 ? 0.0 : mat.get_biochem_param(71));
      //
      Crd(4) = - delta_oxyg * oxyg
               - delta_oxyg_fblas * fblas * oxyg
               - delta_oxyg_normt * normt * oxyg
               - delta_oxyg_hypxt * hypxt * oxyg
               - delta_oxyg_endth * endth * oxyg
               + oxyg_supply;
      Grad_Crd(4) = diff_oxyg * Grad_oxyg;
    }
    //
    // fibroblasts
    {
      const Real necr_fblas = mat.get_biochem_param(90)
                            * positive_only(&exp, -mat.get_biochem_param(91)*oxyg, oxyg)
                            * (RT_dose>0 ? mat.get_biochem_param(92) : 1.0),
                 alpha_fblas = (fblas>fblas_Th ? necr_fblas : 0.0),
                 rho_fblas   = (fblas>fblas_Th && ecm_V_norm>ecm_V_Th ? mat.get_biochem_param(100) : 0.0)
                             * (oxyg>0 ? (oxyg<1 ? (1.0-pow(1.0-oxyg, mat.get_biochem_param(101))) : 1.0) : 0.0);
      const Real diff_fblas  = (fblas>fblas_Th ? mat.get_biochem_param(110) : 0.0),
                 hapto_fblas = (fblas>fblas_Th ? mat.get_biochem_param(111) : 0.0);
      //
      Crd(5) = - alpha_fblas * fblas
               + rho_fblas * ecm_V_norm * fblas;
      Grad_Crd(5) = diff_fblas * ecm_V_norm * Grad_fblas
                  + hapto_fblas * ecm_V_norm * zeta * fblas;
    }
    //
    // normoxic & hypoxic tumor cells
    {
      const Real necr_normt = mat.get_biochem_param(130)
                            * positive_only(&exp, -mat.get_biochem_param(131)*oxyg, oxyg)
                            * (RT_dose>0 ? mat.get_biochem_param(132) : 1.0),
                 alpha_normt = (normt>normt_Th ? necr_normt : 0.0),
                 rho_normt   = (normt>normt_Th && ecm_V_norm>ecm_V_Th ? mat.get_biochem_param(140) : 0.0)
                             * (oxyg>0 ? (oxyg<1 ? (1.0-pow(1.0-oxyg, mat.get_biochem_param(141))) : 1.0) : 0.0);
      const Real diff_normt  = (normt>normt_Th                   ? mat.get_biochem_param(150) : 0.0),
                 hapto_normt = (normt>normt_Th && collg>collg_Th ? mat.get_biochem_param(151) : 0.0),
                 chemo_normt = (normt>normt_Th                   ? mat.get_biochem_param(152) : 0.0);
      //
      const Real beta_normt_hypxt = (normt>normt_Th ? mat.get_biochem_param(170) : 0.0)
                                  * (oxyg>0 ? (oxyg<1 ? (0.0+pow(1.0-oxyg, mat.get_biochem_param(171))) : 0.0) : 0.0),
                 beta_hypxt_normt = (hypxt>hypxt_Th ? mat.get_biochem_param(172) : 0.0)
                                  * (oxyg>0 ? (oxyg<1 ? (1.0-pow(1.0-oxyg, mat.get_biochem_param(173))) : 1.0) : 0.0);
      //
      Crd(6) = - alpha_normt * normt
               + rho_normt * ecm_V_norm * normt
               - beta_normt_hypxt * normt
               + beta_hypxt_normt * hypxt;
      Grad_Crd(6) = diff_normt * ecm_V_norm * Grad_normt
                  - hapto_normt * ecm_V_norm * Grad_collg * normt;
                  - chemo_normt * ecm_V_norm * Grad_oxyg  * normt;
      //
      const Real necr_hypxt = mat.get_biochem_param(180)
                            * positive_only(&exp, -mat.get_biochem_param(181)*oxyg, oxyg)
                            * (RT_dose>0 ? mat.get_biochem_param(182) : 1.0),
                 alpha_hypxt = (hypxt>hypxt_Th ? necr_hypxt : 0.0);
      const Real diff_hypxt  = (hypxt>hypxt_Th                   ? mat.get_biochem_param(190) : 0.0),
                 hapto_hypxt = (hypxt>hypxt_Th && collg>collg_Th ? mat.get_biochem_param(191) : 0.0),
                 chemo_hypxt = (hypxt>hypxt_Th                   ? mat.get_biochem_param(192) : 0.0);
      //
      Crd(7) = - alpha_hypxt * hypxt
               + beta_normt_hypxt * normt
               - beta_hypxt_normt * hypxt;
      Grad_Crd(7) = diff_hypxt * ecm_V_norm * Grad_hypxt
                  - hapto_hypxt * ecm_V_norm * Grad_collg * hypxt;
                  - chemo_hypxt * ecm_V_norm * Grad_oxyg  * hypxt;
    }
    //
    // endothelial cells
    {
      const Real necr_endth = mat.get_biochem_param(210)
                            * positive_only(&exp, -mat.get_biochem_param(211)*oxyg, oxyg)
                            * (RT_dose>0 ? mat.get_biochem_param(212) : 1.0),
                 alpha_endth = (endth>endth_Th ? necr_endth : 0.0),
                 rho_endth   = (endth>endth_Th && ecm_V_norm>ecm_V_Th ? mat.get_biochem_param(220) : 0.0)
                             * (oxyg>0 ? (oxyg<1 ? (1.0-pow(1.0-oxyg, mat.get_biochem_param(221))) : 1.0) : 0.0)
                             * (angf>0 ? pow(angf, mat.get_biochem_param(222)) : 0.0);
      const Real angf_Th = mat.get_biochem_param(240);
      const Real diff_endth  = (endth>endth_Th                   && angf>angf_Th ? mat.get_biochem_param(230) : 0.0),
                 hapto_endth = (endth>endth_Th && collg>collg_Th && angf>angf_Th ? mat.get_biochem_param(231) : 0.0),
                 chemo_endth = (endth>endth_Th                   && angf>angf_Th ? mat.get_biochem_param(232) : 0.0);
      //
      Crd(8) = - alpha_endth * endth
               + rho_endth * ecm_V_norm * endth;
      Grad_Crd(8) = diff_endth * ecm_V_norm * Grad_endth
                  - hapto_endth * ecm_V_norm * theta * collg * endth
                  - chemo_endth * ecm_V_norm * Grad_angf * endth;
    }
    //
    // pro-angiogenic factors
    {
      const Real delta_angf = (angf>0 ? mat.get_biochem_param(250) : 0.0),
                 lambda_angf_fblas = (fblas>fblas_Th ? mat.get_biochem_param(251) : 0.0),
                 lambda_angf_normt = (normt>normt_Th ? mat.get_biochem_param(252) : 0.0),
                 lambda_angf_hypxt = (hypxt>hypxt_Th ? mat.get_biochem_param(253) : 0.0),
                 lambda_angf_endth = (endth>endth_Th ? mat.get_biochem_param(254) : 0.0),
                 gamma_angf_endth = (angf>0 && endth>endth_Th ? mat.get_biochem_param(261) : 0.0) ;
      const Real angf_Th = mat.get_biochem_param(260);
      const Real diff_angf = (angf>angf_Th ? mat.get_biochem_param(270) : 0.0);
      //
      Crd(9) = - delta_angf * angf
               + lambda_angf_fblas * fblas
               + lambda_angf_normt * normt
               + lambda_angf_hypxt * hypxt
               + lambda_angf_endth * endth
               - gamma_angf_endth * endth;
      Grad_Crd(9) = diff_angf * Grad_angf;
    }
    //
    // matrix degrading enzymes
    {
      const Real delta_mde = (mde>0 ? mat.get_biochem_param(280) : 0.0),
                 lambda_mde_fblas = (fblas>fblas_Th ? mat.get_biochem_param(281) : 0.0),
                 lambda_mde_normt = (normt>normt_Th ? mat.get_biochem_param(282) : 0.0),
                 lambda_mde_hypxt = (hypxt>hypxt_Th ? mat.get_biochem_param(283) : 0.0),
                 lambda_mde_endth = (endth>endth_Th ? mat.get_biochem_param(284) : 0.0);
      const Real diff_mde = (mde>0 ? mat.get_biochem_param(290) : 0.0);
      //
      Crd(10) = - delta_mde * mde
               + lambda_mde_fblas * fblas
               + lambda_mde_normt * normt
               + lambda_mde_hypxt * hypxt
               + lambda_mde_endth * endth;
      Grad_Crd(10) = diff_mde * Grad_mde;
    }
}
//=================================================================================================
void healthy_region_1115 ( _LOCAL_ARGUMENTS_ ) {
    invitro_tumoroid_growth::damaged_region_1115(mat, fld, Crd, Grad_Crd, Ksm, Velc, Diff, Taxs);
}
//=================================================================================================
void multi_region_4596 ( _LOCAL_ARGUMENTS_ ) {
    const unsigned int nvars = 45;
    Crd.resize(nvars);
    Crd = 0.0;
    Grad_Crd.resize(nvars);
    Grad_Crd = RealGradient(0.0, 0.0, 0.0);
    Ksm.resize(nvars, nvars);
    Ksm = 0.0;
    Velc.resize(nvars);
    Velc = Point(0.0, 0.0, 0.0);
    Diff.resize(nvars, nvars);
    Diff = 0.0;
    Taxs.resize(nvars, nvars);
    Taxs = RealGradient(0.0, 0.0, 0.0);
    //
    if ( -1 == fld.a_flag ) return;
    //
    // const Real blv__dens = apply_lbound(0.0, fld.biochem[nvars+0]), // discontinuous field variables
    //            lyv__dens = apply_lbound(0.0, fld.biochem[nvars+1]),
    //            blv__tip  = apply_lbound(0, fld.biochem[nvars+2]),
    //            lyv__tip  = apply_lbound(0, fld.biochem[nvars+3]),
    //            blv__func = apply_lbound(0, fld.biochem[nvars+4]),
    //            lyv__func = apply_lbound(0, fld.biochem[nvars+5]),
    //            blv__perf = apply_lbound(0, fld.biochem[nvars+6]),
    //            lyv__perf = apply_lbound(0, fld.biochem[nvars+7]),
    //            RT_magn   = apply_lbound(0, fld.biochem[nvars+8]);
    const int blv__perf = fld.biochem[nvars+6];
    //
    const Real drugV_[] = { apply_lbound(0.0, fld.auxiliary[0]) , // auxiliary field variables
                            apply_lbound(0.0, fld.auxiliary[1]) ,
                            apply_lbound(0.0, fld.auxiliary[2]) ,
                            apply_lbound(0.0, fld.auxiliary[3]) ,
                            apply_lbound(0.0, fld.auxiliary[4]) ,
                            apply_lbound(0.0, fld.auxiliary[5]) };
    //
    const Point velocity = coordinates(fld.fluid_velocity);
    const Point epsilon = coordinates(unit(min_eigen(fld.strain)));
    //
    const Real O2G = apply_bounds(0., fld.biochem[0], 1.); // oxygen, glucose & other nutrients saturation level
    const Real ECM_cl = fld.biochem[1], // ECM collagen
               ECM_fn = fld.biochem[2], // ECM fibronectin
               ECM_lm = fld.biochem[3], // ECM laminin
               ECM_ha = fld.biochem[4]; // ECM hyaluronic acid
    const Real MSC  = fld.biochem[ 5], // mesenchymal stem cell
               FB   = fld.biochem[ 6], // fibroblast
               EC   = fld.biochem[ 7], // epithelial cell
               ICTL = fld.biochem[ 8], // immune cell: T lymphocyte
               ICMP = fld.biochem[ 9], // immune cell: macrophage
               TCN  = fld.biochem[10], // tumour: normoxic cancer cell
               TCH  = fld.biochem[11], // tumour: hypoxic cancer cell
               TASC = fld.biochem[12], // tumour-associated stem cell
               TAFB = fld.biochem[13], // tumour-associated fibroblast
               TAMP = fld.biochem[14], // tumour-associated macrophage
               NTC  = fld.biochem[15]; // necrotic tumour cell
    const Real VEGF = fld.biochem[16], // vascular endothelial growth factor (VEGF)
               ANG1 = fld.biochem[17], // angiopoietin 1
               ANG2 = fld.biochem[18], // angiopoietin 2
               EGF  = fld.biochem[19], // epidermal growth factor (EGF)
               TGFa = fld.biochem[20], // transforming growth factor 1
               TGFb = fld.biochem[21], // transforming growth factor 2
               MDGF = fld.biochem[22], // macrophage-derived growth factor (MDGF)
               MMP  = fld.biochem[23], // matrix metalloproteinase (MMP)
               ILX  = fld.biochem[24]; // interleukins (ILs)
    const Real VEGFR= fld.biochem[25], // receptor: vascular endothelial growth factor
               TIE1 = fld.biochem[26], // receptor: tyrosine kinase with immunoglobulin-like & EGF-like domains 1
               TIE2 = fld.biochem[27], // receptor: tyrosine kinase with immunoglobulin-like & EGF-like domains 2
               EGFR = fld.biochem[28], // receptor: epidermal growth factor
               TGFR = fld.biochem[29], // receptor: transforming growth factor
               WNT  = fld.biochem[30], // receptor: Wnt
               HDGH = fld.biochem[31], // receptor: hedgehog
               NTCH = fld.biochem[32]; // receptor: notch
    const Real drug0_[] = { fld.biochem[33], fld.biochem[34], fld.biochem[35] },
               drug1_[] = { fld.biochem[36], fld.biochem[37], fld.biochem[38] },
               drug2_[] = { fld.biochem[39], fld.biochem[40], fld.biochem[41] },
               drug3_[] = { fld.biochem[42], fld.biochem[43], fld.biochem[44] };
    //
    const RealGradient Grad_O2G = fld.grad_biochem[0];
    const RealGradient Grad_ECM_cl = fld.grad_biochem[1], // ECM collagen
                       Grad_ECM_fn = fld.grad_biochem[2], // ECM fibronectin
                       Grad_ECM_lm = fld.grad_biochem[3], // ECM laminin
                       Grad_ECM_ha = fld.grad_biochem[4], // ECM hyaluronic acid
                       Grad_ECM = Grad_ECM_cl + Grad_ECM_fn + Grad_ECM_lm + Grad_ECM_ha; // ECM (net)
    const RealGradient Grad_MSC  = fld.grad_biochem[ 5], // mesenchymal stem cell
                       Grad_FB   = fld.grad_biochem[ 6], // fibroblast
                       Grad_EC   = fld.grad_biochem[ 7], // epithelial cell
                       Grad_ICTL = fld.grad_biochem[ 8], // immune: T lymphocyte
                       Grad_ICMP = fld.grad_biochem[ 9], // immune: macrophage
                       Grad_TCN  = fld.grad_biochem[10], // tumour: normoxic cancer cell
                       Grad_TCH  = fld.grad_biochem[11], // tumour: hypoxic cancer cell
                       Grad_TASC = fld.grad_biochem[12], // tumour-associated stem cell
                       Grad_TAFB = fld.grad_biochem[13], // tumour-associated fibroblast
                       Grad_TAMP = fld.grad_biochem[14], // tumour-associated macrophage
                       Grad_NTC  = fld.grad_biochem[15]; // necrotic tumour cell
    const RealGradient Grad_VEGF = fld.grad_biochem[16], // vascular endothelial growth factor (VEGF)
                       Grad_ANG1 = fld.grad_biochem[17], // angiopoietin 1
                       Grad_ANG2 = fld.grad_biochem[18], // angiopoietin 2
                       Grad_EGF  = fld.grad_biochem[19], // epidermal growth factor (EGF)
                       Grad_TGFa = fld.grad_biochem[20], // transforming growth factor 1
                       Grad_TGFb = fld.grad_biochem[21], // transforming growth factor 2
                       Grad_MDGF = fld.grad_biochem[22], // macrophage-derived growth factor (MDGF)
                       Grad_MMP  = fld.grad_biochem[23], // matrix metalloproteinase (MMP)
                       Grad_ILX  = fld.grad_biochem[24]; // interleukins (ILs)
    //
    const Real Rate_MSC  = fld.biochem_rate[ 5], // mesenchymal stem cell
               Rate_FB   = fld.biochem_rate[ 6], // fibroblast
               Rate_EC   = fld.biochem_rate[ 7], // epithelial cell
               Rate_ICTL = fld.biochem_rate[ 8], // immune cell: lymphocyte
               Rate_ICMP = fld.biochem_rate[ 9], // immune cell: macrophage
               Rate_TCN  = fld.biochem_rate[10], // tumour: normoxic cancer cell
               Rate_TCH  = fld.biochem_rate[11], // tumour: hypoxic cancer cell
               Rate_TASC = fld.biochem_rate[12], // tumour-associated stem cell
               Rate_TAFB = fld.biochem_rate[13], // tumour-associated fibroblast
               Rate_TAMP = fld.biochem_rate[14], // tumour-associated macrophage
               Rate_NTC  = fld.biochem_rate[15]; // necrotic tumour cell
    //
    const Point fibre_[] = { coordinates(fld.fibre_curr[0]) ,
                             coordinates(fld.fibre_curr[1]) ,
                             coordinates(fld.fibre_curr[2]) };
    const Point fibre__O2G[] = { ( fibre_[0]*Grad_O2G<0.0 ? -1.0 : +1.0 ) * fibre_[0] ,
                                 ( fibre_[1]*Grad_O2G<0.0 ? -1.0 : +1.0 ) * fibre_[1] ,
                                 ( fibre_[2]*Grad_O2G<0.0 ? -1.0 : +1.0 ) * fibre_[2] };
    //
    const Real average_temp = mat.get_biochem_param(0); // K (tissue average temperature)
    //
    const Real blood_visc        = mat.get_biochem_param(1)*normal_distro(1., 0.025),
               lymph_visc        = mat.get_biochem_param(2)*normal_distro(1., 0.025),
               plasma_visc       = mat.get_biochem_param(3)*normal_distro(1., 0.025),
               interstitial_visc = mat.get_biochem_param(4)*normal_distro(1., 0.025);
    //
    Real PHI = 0.0;
    {
        const Real RVE = mat.get_biochem_param(10),
                   // ECM volume fraction: parenchyma
                   VF__Parenchyma = mat.get_biochem_param(9)*normal_distro(1., 0.05),
                   // ECM volume fraction: stroma
                   VF__Stroma = ( ECM_cl/(mat.get_biochem_param(31)*normal_distro(1., 0.05))
                                + ECM_fn/(mat.get_biochem_param(32)*normal_distro(1., 0.05))
                                + ECM_lm/(mat.get_biochem_param(33)*normal_distro(1., 0.05))
                                + ECM_ha/(mat.get_biochem_param(34)*normal_distro(1., 0.05)) )
                              / RVE,
                   // Cells volume fraction
                   VF__Cell = ( MSC *(mat.get_biochem_param(11)*normal_distro(1., 0.05))
                              + FB  *(mat.get_biochem_param(12)*normal_distro(1., 0.05))
                              + EC  *(mat.get_biochem_param(13)*normal_distro(1., 0.05))
                              + ICTL*(mat.get_biochem_param(14)*normal_distro(1., 0.05))
                              + ICMP*(mat.get_biochem_param(15)*normal_distro(1., 0.05))
                              + TCN *(mat.get_biochem_param(16)*normal_distro(1., 0.05))
                              + TCH *(mat.get_biochem_param(17)*normal_distro(1., 0.05))
                              + TASC*(mat.get_biochem_param(18)*normal_distro(1., 0.05))
                              + TAFB*(mat.get_biochem_param(19)*normal_distro(1., 0.05))
                              + TAMP*(mat.get_biochem_param(20)*normal_distro(1., 0.05))
                              + NTC *(mat.get_biochem_param(21)*normal_distro(1., 0.05)) )
                            / RVE,
                   // total volume fraction; must be <=1
                   VF = VF__Parenchyma + VF__Stroma + VF__Cell;
        // scalar exponential factor
        const Real a = mat.get_biochem_param(98)>1.0 ? mat.get_biochem_param(98) : 0.0;
        //
        if ( VF <= 0.999 ) {
            PHI = a
                ? pow(1.0-VF, a) : (1.0-VF/(1.0-VF__Parenchyma));
            PHI = apply_bounds(0.0, PHI, 1.0);
        } else {
            PHI = 0.0;
        }
    }
    //
    const Real Cell__LB = mat.get_biochem_param(99);
    //
    // ##########################################
    // ### state-variable 0: Oxygen & Glucose ###
    // ##########################################
    {
        const Real O2G__LB = mat.get_biochem_param(100);
        //
        const Real alpha = blv__perf>=1 ? mat.get_biochem_param(101)*normal_distro(1., 0.02) : 0.0 ;
        Crd(0) += alpha;
        //
        const Real beta_MSC  = ( O2G>O2G__LB ? mat.get_biochem_param(111)*normal_distro(1., 0.05) : 0.0 ),
                   beta_FB   = ( O2G>O2G__LB ? mat.get_biochem_param(112)*normal_distro(1., 0.05) : 0.0 ),
                   beta_EC   = ( O2G>O2G__LB ? mat.get_biochem_param(113)*normal_distro(1., 0.05) : 0.0 ),
                   beta_ICTL = ( O2G>O2G__LB ? mat.get_biochem_param(114)*normal_distro(1., 0.05) : 0.0 ),
                   beta_ICMP = ( O2G>O2G__LB ? mat.get_biochem_param(115)*normal_distro(1., 0.05) : 0.0 ),
                   beta_TCN  = ( O2G>O2G__LB ? mat.get_biochem_param(116)*normal_distro(1., 0.05) : 0.0 ),
                   beta_TCH  = ( O2G>O2G__LB ? mat.get_biochem_param(117)*normal_distro(1., 0.05) : 0.0 ),
                   beta_TASC = ( O2G>O2G__LB ? mat.get_biochem_param(118)*normal_distro(1., 0.05) : 0.0 ),
                   beta_TAFB = ( O2G>O2G__LB ? mat.get_biochem_param(119)*normal_distro(1., 0.05) : 0.0 ),
                   beta_TAMP = ( O2G>O2G__LB ? mat.get_biochem_param(120)*normal_distro(1., 0.05) : 0.0 ),
                   beta_NTC  = ( O2G>O2G__LB ? mat.get_biochem_param(121)*normal_distro(1., 0.05) : 0.0 );
        Crd(0) -= beta_MSC*MSC + beta_FB*FB + beta_EC*EC + beta_ICTL*ICTL + beta_ICMP*ICMP
                + beta_TCN*TCN + beta_TCH*TCH + beta_TASC*TASC + beta_TAFB*TAFB + beta_TAMP*TAMP + beta_NTC*NTC;
        //
        Crd(0) -= velocity*Grad_O2G;
        //
        // Stokes-Einstein equation [https://en.wikipedia.org/wiki/Einstein_relation_(kinetic_theory)]
        const Real D = 0.100e-9; // m (molecule size)
        const Real fickian = !(mat.get_biochem_param(199)>0)
                           ? Boltzmann_constant*average_temp/(3.0*mPI*interstitial_visc*D) : mat.get_biochem_param(199);
        const Real diff = O2G>=O2G__LB
                        ? fickian : 0.0;
        //
        Grad_Crd(0) += diff*Grad_O2G;
    }
    // ##################################
    // ### state-variable 1: Collagen ###
    // ##################################
    { // 200-...
        ;
    }
    // #####################################
    // ### state-variable 2: Fibronectin ###
    // #####################################
    { // 400-...
        ;
    }
    // #################################
    // ### state-variable 3: Laminin ###
    // #################################
    { // 600-...
        ;
    }
    // #########################################
    // ### state-variable 4: Hyaluronic acid ###
    // #########################################
    { // 800-...
        ;
    }
    // ###############################################
    // ### state-variable 5: Mesenchymal Stem Cell ###
    // ###############################################
    {
        const Real O2G__Ap = mat.get_biochem_param(1000);
        //
        const Real alpha1 = mat.get_biochem_param(1001)*normal_distro(1., 0.02),
                   alpha2 = mat.get_biochem_param(1002)*normal_distro(1., 0.02),
                    P_O2G = O2G>O2G__Ap && (alpha2+pow2(O2G))
                          ? alpha1*pow2(O2G)/(alpha2+pow2(O2G))*(1.0+alpha2) : 0.0;
        //
        Crd(5) += P_O2G*MSC*PHI;
        //
        const Real delta1 = mat.get_biochem_param(1003)*normal_distro(1., 0.02),
                   delta2 = mat.get_biochem_param(1004)*normal_distro(1., 0.02),
                    A_O2G = MSC>Cell__LB
                          ? (O2G>O2G__Ap ? delta1 : delta2) : 0.0;
        //
        Crd(5) -= A_O2G*MSC;
        //
        const Real beta1 = mat.get_biochem_param(2421)*normal_distro(1., 0.02), // @elsewhere
                  T_TGFa = MSC>Cell__LB
                         ? beta1*TGFa : 0.0;
        //
        Crd(5) -= T_TGFa*MSC;
        //
        const Real diff = MSC>Cell__LB && O2G>O2G__Ap
                        ? mat.get_biochem_param(1199)*normal_distro(1., 0.02) : 0.0;
        //
        Grad_Crd(5) += diff*Grad_MSC*PHI;
        //
        const Real duro = MSC>Cell__LB && O2G>O2G__Ap
                        ? mat.get_biochem_param(1198)*normal_distro(1., 0.02) : 0.0;
        //
        Grad_Crd(5) += duro*MSC*Grad_ECM*PHI;
        //
        const Real chemtx1 = mat.get_biochem_param(1197)*normal_distro(1., 0.02),
                   chemtx2 = mat.get_biochem_param(1196)*normal_distro(1., 0.02),
                     C_O2G = MSC>Cell__LB && O2G>O2G__Ap && (chemtx2+pow4(O2G))
                           ? chemtx1*(chemtx2+1.0)*pow4(O2G)/(chemtx2+pow4(O2G)) : 0.0;
        //
        Grad_Crd(5) += C_O2G*MSC*Grad_O2G*PHI;
        //
    }
    // ####################################
    // ### state-variable 6: Fibroblast ###
    // ####################################
    {
        const Real O2G__Ap = mat.get_biochem_param(1200);
        //
        const Real alpha1 = mat.get_biochem_param(1201)*normal_distro(1., 0.02),
                   alpha2 = mat.get_biochem_param(1202)*normal_distro(1., 0.02),
                    P_O2G = O2G>O2G__Ap && (alpha2+pow2(O2G))
                          ? alpha1*pow2(O2G)/(alpha2+pow2(O2G))*(1.0+alpha2) : 0.0;
        //
        Crd(6) += P_O2G*FB*PHI;
        //
        const Real delta1 = mat.get_biochem_param(1203)*normal_distro(1., 0.02),
                   delta2 = mat.get_biochem_param(1204)*normal_distro(1., 0.02),
                    A_O2G = FB>Cell__LB
                          ? (O2G>O2G__Ap ? delta1 : delta2) : 0.0;
        //
        Crd(6) -= A_O2G*FB;
        //
        const Real beta1 = mat.get_biochem_param(2621)*normal_distro(1., 0.02), // @elsewhere
                  T_TGFb = FB>Cell__LB
                         ? beta1*TGFb : 0.0;
        //
        Crd(6) -= T_TGFb*FB;
        //
        const Real diff = FB>Cell__LB && O2G>O2G__Ap
                        ? mat.get_biochem_param(1399)*normal_distro(1., 0.02) : 0.0;
        //
        Grad_Crd(6) += diff*Grad_FB*PHI;
        //
        const Real duro = FB>Cell__LB && O2G>O2G__Ap
                        ? mat.get_biochem_param(1398)*normal_distro(1., 0.02) : 0.0;
        //
        Grad_Crd(6) += duro*FB*Grad_ECM*PHI;
        //
        const Real chemtx1 = mat.get_biochem_param(1397)*normal_distro(1., 0.02),
                   chemtx2 = mat.get_biochem_param(1396)*normal_distro(1., 0.02),
                     C_O2G = FB>Cell__LB && O2G>O2G__Ap && (chemtx2+pow4(O2G))
                           ? chemtx1*(chemtx2+1.0)*pow4(O2G)/(chemtx2+pow4(O2G)) : 0.0;
        //
        Grad_Crd(6) += C_O2G*FB*Grad_O2G*PHI;
        //
    }
    // #########################################
    // ### state-variable 7: Epithelial Cell ###
    // #########################################
    {
        const Real O2G__Ap = mat.get_biochem_param(1400);
        //
        const Real alpha1 = mat.get_biochem_param(1401)*normal_distro(1., 0.02),
                   alpha2 = mat.get_biochem_param(1402)*normal_distro(1., 0.02),
                   P_VEGF = O2G>O2G__Ap && (alpha2+VEGF)
                          ? alpha1*VEGF/(alpha2+VEGF) : 0.0;
        //
        Crd(7) += P_VEGF*EC*PHI;
        //
        const Real delta1 = mat.get_biochem_param(1403)*normal_distro(1., 0.02),
                   delta2 = mat.get_biochem_param(1404)*normal_distro(1., 0.02),
                    A_O2G = EC>Cell__LB
                          ? (O2G>O2G__Ap ? delta1 : delta2) : 0.0;
        //
        Crd(7) -= A_O2G*EC;
        //
        const Real diff = EC>Cell__LB && O2G>O2G__Ap
                        ? mat.get_biochem_param(1599)*normal_distro(1., 0.02) : 0.0;
        //
        Grad_Crd(7) += diff*Grad_EC*PHI;
        //
        const Real duro = EC>Cell__LB && O2G>O2G__Ap
                        ? mat.get_biochem_param(1598)*normal_distro(1., 0.02) : 0.0;
        //
        Grad_Crd(7) += duro*EC*Grad_ECM*PHI;
        //
        const Real chemtx1 = mat.get_biochem_param(1597)*normal_distro(1., 0.02),
                   chemtx2 = mat.get_biochem_param(1596)*normal_distro(1., 0.02),
                     C_O2G = EC>Cell__LB && O2G>O2G__Ap && (chemtx2+pow4(O2G))
                           ? chemtx1*(chemtx2+1.0)*pow4(O2G)/(chemtx2+pow4(O2G)) : 0.0;
        //
        Grad_Crd(7) += C_O2G*EC*Grad_O2G*PHI;
        //
    }
    // ####################################################
    // ### state-variable 8: Immune Cell - T Lymphocyte ###
    // ####################################################
    {
        const Real O2G__Ap = mat.get_biochem_param(1600);
        //
        const Real alpha1 = mat.get_biochem_param(1601)*normal_distro(1., 0.02),
                   alpha2 = mat.get_biochem_param(1602)*normal_distro(1., 0.02),
                    P_O2G = O2G>O2G__Ap && (alpha2+pow2(O2G))
                          ? alpha1*pow2(O2G)/(alpha2+pow2(O2G))*(1.0+alpha2) : 0.0;
        //
        Crd(8) += P_O2G*ICTL*PHI;
        //
        const Real delta1 = mat.get_biochem_param(1603)*normal_distro(1., 0.02),
                   delta2 = mat.get_biochem_param(1604)*normal_distro(1., 0.02),
                    A_O2G = ICTL>Cell__LB
                          ? (O2G>O2G__Ap ? delta1 : delta2) : 0.0;
        //
        Crd(8) -= A_O2G*ICTL;
        //
        const Real diff = ICTL>Cell__LB && O2G>O2G__Ap
                        ? mat.get_biochem_param(1799)*normal_distro(1., 0.02) : 0.0;
        //
        Grad_Crd(8) += diff*Grad_ICTL*PHI;
        //
        const Real duro = ICTL>Cell__LB && O2G>O2G__Ap
                        ? mat.get_biochem_param(1798)*normal_distro(1., 0.02) : 0.0;
        //
        Grad_Crd(8) += duro*ICTL*Grad_ECM*PHI;
        //
        const Real chemtx1 = mat.get_biochem_param(1797)*normal_distro(1., 0.02),
                   chemtx2 = mat.get_biochem_param(1796)*normal_distro(1., 0.02),
                     C_O2G = ICTL>Cell__LB && O2G>O2G__Ap && (chemtx2+pow4(O2G))
                           ? chemtx1*(chemtx2+1.0)*pow4(O2G)/(chemtx2+pow4(O2G)) : 0.0;
        //
        Grad_Crd(8) += C_O2G*ICTL*Grad_O2G*PHI;
        //
    }
    // ##################################################
    // ### state-variable 9: Immune Cell - Macrophage ###
    // ##################################################
    {
        const Real O2G__Ap = mat.get_biochem_param(1800);
        //
        const Real alpha1 = mat.get_biochem_param(1801)*normal_distro(1., 0.02),
                   alpha2 = mat.get_biochem_param(1802)*normal_distro(1., 0.02),
                    P_O2G = O2G>O2G__Ap && (alpha2+pow2(O2G))
                          ? alpha1*pow2(O2G)/(alpha2+pow2(O2G))*(1.0+alpha2) : 0.0;
        //
        Crd(9) += P_O2G*ICMP*PHI;
        //
        const Real delta1 = mat.get_biochem_param(1803)*normal_distro(1., 0.02),
                   delta2 = mat.get_biochem_param(1804)*normal_distro(1., 0.02),
                    A_O2G = ICMP>Cell__LB
                          ? (O2G>O2G__Ap ? delta1 : delta2) : 0.0;
        //
        Crd(9) -= A_O2G*ICMP;
        //
        const Real beta1 = mat.get_biochem_param(2821)*normal_distro(1., 0.02), // @elsewhere
                   T_ILX = ICMP>Cell__LB
                         ? beta1*ILX : 0.0;
        //
        Crd(9) -= T_ILX*ICMP;
        //
        const Real diff = ICMP>Cell__LB && O2G>O2G__Ap
                        ? mat.get_biochem_param(1999)*normal_distro(1., 0.02) : 0.0;
        //
        Grad_Crd(9) += diff*Grad_ICMP*PHI;
        //
        const Real duro = ICMP>Cell__LB && O2G>O2G__Ap
                        ? mat.get_biochem_param(1998)*normal_distro(1., 0.02) : 0.0;
        //
        Grad_Crd(9) += duro*ICMP*Grad_ECM*PHI;
        //
        const Real chemtx1 = mat.get_biochem_param(1997)*normal_distro(1., 0.02),
                   chemtx2 = mat.get_biochem_param(1996)*normal_distro(1., 0.02),
                     C_O2G = ICMP>Cell__LB && O2G>O2G__Ap && (chemtx2+pow4(O2G))
                           ? chemtx1*(chemtx2+1.0)*pow4(O2G)/(chemtx2+pow4(O2G)) : 0.0;
        //
        Grad_Crd(9) += C_O2G*ICMP*Grad_O2G*PHI;
        //
    }
    // ##################################################
    // ### state-variables 10: Tumour Cell - Normoxic ###
    // ##################################################
    {
        const Real O2G__Hy = mat.get_biochem_param(2000);
        const Real O2G__Nc = mat.get_biochem_param(2200); // @elsewhere
        const Real O2G__Ap = mat.get_biochem_param(3000); // @elsewhere
        //
        const Real alpha1 = mat.get_biochem_param(2001)*normal_distro(1., 0.02),
                   alpha2 = mat.get_biochem_param(2002)*normal_distro(1., 0.02),
                    P_O2G = O2G>O2G__Ap && (alpha2+pow2(O2G))
                          ? alpha1*pow2(O2G)/(alpha2+pow2(O2G))*(1.0+alpha2) : 0.0;
        //
        Crd(10) += P_O2G*TCN*PHI;
        //
        const Real delta1 = mat.get_biochem_param(2003)*normal_distro(1., 0.02),
                   delta2 = mat.get_biochem_param(2004)*normal_distro(1., 0.02),
                    A_O2G = TCN>Cell__LB
                          ? (O2G>O2G__Ap ? delta1 : delta2) : 0.0;
        //
        Crd(10) -= A_O2G*TCN;
        //
        const Real beta1 = mat.get_biochem_param(2021)*normal_distro(1., 0.02),
                   T_O2G = TCH>Cell__LB
                         ? (O2G>=O2G__Hy ? beta1*O2G : 0.0) : 0.0;
        const Real beta2 = mat.get_biochem_param(2221)*normal_distro(1., 0.02), // @elsewhere
                   H_O2G = TCN>Cell__LB
                         ? (O2G<O2G__Hy ? beta2*(1.0-O2G) : 0.0) : 0.0;
        //
        Crd(10) += T_O2G*TCH;
        Crd(10) -= H_O2G*TCN;
        //
        const Real omega1 = mat.get_biochem_param(3011)*normal_distro(1., 0.02), // @elsewhere
                   W_TCN  = TCN>Cell__LB && O2G<O2G__Nc
                          ? omega1*TCN : 0.0;
        //
        Crd(10) -= W_TCN;
        //
        const Real diff = TCN>Cell__LB && O2G>O2G__Ap
                        ? mat.get_biochem_param(2199)*normal_distro(1., 0.02) : 0.0;
        //
        Grad_Crd(10) += diff*Grad_TCN*PHI;
        //
        const Real duro = TCN>Cell__LB && O2G>O2G__Ap
                        ? mat.get_biochem_param(2198)*normal_distro(1., 0.02) : 0.0;
        //
        Grad_Crd(10) += duro*TCN*Grad_ECM*PHI;
        //
        const Real chemtx1 = mat.get_biochem_param(2197)*normal_distro(1., 0.02),
                   chemtx2 = mat.get_biochem_param(2196)*normal_distro(1., 0.02),
                     C_O2G = TCN>Cell__LB && O2G>O2G__Ap && (chemtx2+pow4(O2G))
                           ? chemtx1*(chemtx2+1.0)*pow4(O2G)/(chemtx2+pow4(O2G)) : 0.0;
        //
        Grad_Crd(10) += C_O2G*TCN*Grad_O2G*PHI;
        //
        const Real metast1_EGF = mat.get_biochem_param(2195)*normal_distro(1., 0.02),
                   metast2_EGF = mat.get_biochem_param(2194)*normal_distro(1., 0.02),
                     C_EGF = TCN>Cell__LB && O2G>O2G__Ap && (metast2_EGF+EGF)
                           ? metast1_EGF*pow2(metast2_EGF)/pow2(metast2_EGF+EGF) : 0.0;
        //
        Grad_Crd(10) -= C_EGF*TCN*Grad_EGF*PHI;
        //
    }
    // #################################################
    // ### state-variables 11: Tumour Cell - Hypoxic ###
    // #################################################
    {
        const Real O2G__Hy = mat.get_biochem_param(2000); // @elsewhere
        const Real O2G__Nc = mat.get_biochem_param(2200);
        const Real O2G__Ap = mat.get_biochem_param(3000); // @elsewhere
        //
        const Real delta1 = mat.get_biochem_param(2203)*normal_distro(1., 0.02),
                   delta2 = mat.get_biochem_param(2204)*normal_distro(1., 0.02),
                    A_O2G = TCH>Cell__LB
                          ? (O2G>O2G__Ap ? delta1 : delta2) : 0.0;
        //
        Crd(11) -= A_O2G*TCH;
        //
        const Real beta1 = mat.get_biochem_param(2021)*normal_distro(1., 0.02), // @elsewhere
                   T_O2G = TCH>Cell__LB
                         ? (O2G>=O2G__Hy ? beta1*O2G : 0.0) : 0.0;
        const Real beta2 = mat.get_biochem_param(2221)*normal_distro(1., 0.02),
                   H_O2G = TCN>Cell__LB
                         ? (O2G<O2G__Hy ? beta2*(1.0-O2G) : 0.0) : 0.0;
        //
        Crd(11) -= T_O2G*TCH;
        Crd(11) += H_O2G*TCN;
        //
        const Real omega2 = mat.get_biochem_param(3012)*normal_distro(1., 0.02), // @elsewhere
                   W_TCH  = TCH>Cell__LB && O2G<O2G__Nc
                          ? omega2*TCH : 0.0;
        //
        Crd(11) -= W_TCH;
        //
        const Real diff = TCH>Cell__LB && O2G>O2G__Ap
                        ? mat.get_biochem_param(2399)*normal_distro(1., 0.02) : 0.0;
        //
        Grad_Crd(11) += diff*Grad_TCH*PHI;
        //
        const Real duro = TCH>Cell__LB && O2G>O2G__Ap
                        ? mat.get_biochem_param(2398)*normal_distro(1., 0.02) : 0.0;
        //
        Grad_Crd(11) += duro*TCH*Grad_ECM*PHI;
        //
        const Real chemtx1 = mat.get_biochem_param(2397)*normal_distro(1., 0.02),
                   chemtx2 = mat.get_biochem_param(2396)*normal_distro(1., 0.02),
                     C_O2G = TCH>Cell__LB && O2G>O2G__Ap && (chemtx2+pow4(O2G))
                           ? chemtx1*(chemtx2+1.0)*pow4(O2G)/(chemtx2+pow4(O2G)) : 0.0;
        //
        Grad_Crd(11) += C_O2G*TCH*Grad_O2G*PHI;
        //
        const Real metast1_EGF = mat.get_biochem_param(2395)*normal_distro(1., 0.02),
                   metast2_EGF = mat.get_biochem_param(2394)*normal_distro(1., 0.02),
                     C_EGF = TCH>Cell__LB && O2G>O2G__Ap && (metast2_EGF+EGF)
                           ? metast1_EGF*pow2(metast2_EGF)/pow2(metast2_EGF+EGF) : 0.0;
        //
        Grad_Crd(11) -= C_EGF*TCH*Grad_EGF*PHI;
        //
    }
    // ######################################################
    // ### state-variable 12: Tumour-associated Stem Cell ###
    // ######################################################
    {
        const Real O2G__Nc = mat.get_biochem_param(2200); // @elsewhere
        const Real O2G__Ap = mat.get_biochem_param(2400);
        //
        const Real alpha1 = mat.get_biochem_param(2401)*normal_distro(1., 0.02),
                   alpha2 = mat.get_biochem_param(2402)*normal_distro(1., 0.02),
                    P_O2G = O2G>O2G__Ap && (alpha2+pow2(O2G))
                          ? alpha1*pow2(O2G)/(alpha2+pow2(O2G))*(1.0+alpha2) : 0.0;
        //
        Crd(12) += P_O2G*TASC*PHI;
        //
        const Real delta1 = mat.get_biochem_param(2403)*normal_distro(1., 0.02),
                   delta2 = mat.get_biochem_param(2404)*normal_distro(1., 0.02),
                    A_O2G = TASC>Cell__LB
                          ? (O2G>O2G__Ap ? delta1 : delta2) : 0.0;
        //
        Crd(12) -= A_O2G*TASC;
        //
        const Real beta1 = mat.get_biochem_param(2421)*normal_distro(1., 0.02),
                  T_TGFa = MSC>Cell__LB
                         ? beta1*TGFa : 0.0;
        //
        Crd(12) += T_TGFa*MSC;
        //
        const Real omega3 = mat.get_biochem_param(3013)*normal_distro(1., 0.02), // @elsewhere
                   W_TASC = TASC>Cell__LB && O2G<O2G__Nc
                          ? omega3*TASC : 0.0;
        //
        Crd(12) -= W_TASC;
        //
        const Real diff = TASC>Cell__LB && O2G>O2G__Ap
                        ? mat.get_biochem_param(2599)*normal_distro(1., 0.02) : 0.0;
        //
        Grad_Crd(12) += diff*Grad_TASC*PHI;
        //
        const Real duro = TASC>Cell__LB && O2G>O2G__Ap
                        ? mat.get_biochem_param(2598)*normal_distro(1., 0.02) : 0.0;
        //
        Grad_Crd(12) += duro*TASC*Grad_ECM*PHI;
        //
        const Real chemtx1 = mat.get_biochem_param(2597)*normal_distro(1., 0.02),
                   chemtx2 = mat.get_biochem_param(2596)*normal_distro(1., 0.02),
                     C_O2G = TASC>Cell__LB && O2G>O2G__Ap && (chemtx2+pow4(O2G))
                           ? chemtx1*(chemtx2+1.0)*pow4(O2G)/(chemtx2+pow4(O2G)) : 0.0;
        //
        Grad_Crd(12) += C_O2G*TASC*Grad_O2G*PHI;
        //
    }
    // #######################################################
    // ### state-variable 13: Tumour-associated Fibroblast ###
    // #######################################################
    {
        const Real O2G__Nc = mat.get_biochem_param(2200); // @elsewhere
        const Real O2G__Ap = mat.get_biochem_param(2600);
        //
        const Real alpha1 = mat.get_biochem_param(2601)*normal_distro(1., 0.02),
                   alpha2 = mat.get_biochem_param(2602)*normal_distro(1., 0.02),
                    P_O2G = O2G>O2G__Ap && (alpha2+pow2(O2G))
                          ? alpha1*pow2(O2G)/(alpha2+pow2(O2G))*(1.0+alpha2) : 0.0;
        //
        Crd(13) += P_O2G*TAFB*PHI;
        //
        const Real delta1 = mat.get_biochem_param(2603)*normal_distro(1., 0.02),
                   delta2 = mat.get_biochem_param(2604)*normal_distro(1., 0.02),
                    A_O2G = TAFB>Cell__LB
                          ? (O2G>O2G__Ap ? delta1 : delta2) : 0.0;
        //
        Crd(13) -= A_O2G*TAFB;
        //
        const Real beta1 = mat.get_biochem_param(2621)*normal_distro(1., 0.02),
                  T_TGFb = FB>Cell__LB
                         ? beta1*TGFb : 0.0;
        //
        Crd(13) += T_TGFb*FB;
        //
        const Real omega4 = mat.get_biochem_param(3014)*normal_distro(1., 0.02), // @elsewhere
                   W_TAFB = TAFB>Cell__LB && O2G<O2G__Nc
                          ? omega4*TAFB : 0.0;
        //
        Crd(13) -= W_TAFB;
        //
        const Real diff = TAFB>Cell__LB && O2G>O2G__Ap
                        ? mat.get_biochem_param(2799)*normal_distro(1., 0.02) : 0.0;
        //
        Grad_Crd(13) += diff*Grad_TAFB*PHI;
        //
        const Real duro = TAFB>Cell__LB && O2G>O2G__Ap
                        ? mat.get_biochem_param(2798)*normal_distro(1., 0.02) : 0.0;
        //
        Grad_Crd(13) += duro*TAFB*Grad_ECM*PHI;
        //
        const Real chemtx1 = mat.get_biochem_param(2797)*normal_distro(1., 0.02),
                   chemtx2 = mat.get_biochem_param(2796)*normal_distro(1., 0.02),
                     C_O2G = TAFB>Cell__LB && O2G>O2G__Ap && (chemtx2+pow4(O2G))
                           ? chemtx1*(chemtx2+1.0)*pow4(O2G)/(chemtx2+pow4(O2G)) : 0.0;
        //
        Grad_Crd(13) += C_O2G*TAFB*Grad_O2G*PHI;
        //
    }
    // #######################################################
    // ### state-variable 14: Tumour-associated Macrophage ###
    // #######################################################
    {
        const Real O2G__Nc = mat.get_biochem_param(2200); // @elsewhere
        const Real O2G__Ap = mat.get_biochem_param(2800);
        //
        const Real alpha1 = mat.get_biochem_param(2801)*normal_distro(1., 0.02),
                   alpha2 = mat.get_biochem_param(2802)*normal_distro(1., 0.02),
                    P_O2G = O2G>O2G__Ap && (alpha2+pow2(O2G))
                          ? alpha1*pow2(O2G)/(alpha2+pow2(O2G))*(1.0+alpha2) : 0.0;
        //
        Crd(14) += P_O2G*TAMP*PHI;
        //
        const Real delta1 = mat.get_biochem_param(2803)*normal_distro(1., 0.02),
                   delta2 = mat.get_biochem_param(2804)*normal_distro(1., 0.02),
                    A_O2G = TAMP>Cell__LB
                          ? (O2G>O2G__Ap ? delta1 : delta2) : 0.0;
        //
        Crd(14) -= A_O2G*TAMP;
        //
        const Real beta1 = mat.get_biochem_param(2821)*normal_distro(1., 0.02),
                   T_ILX = ICMP>Cell__LB
                         ? beta1*ILX : 0.0;
        //
        Crd(14) += T_ILX*ICMP;
        //
        const Real omega5 = mat.get_biochem_param(3015)*normal_distro(1., 0.02), // @elsewhere
                   W_TAMP = TAMP>Cell__LB && O2G<O2G__Nc
                          ? omega5*TAMP : 0.0;
        //
        Crd(14) -= W_TAMP;
        //
        const Real diff = TAMP>Cell__LB && O2G>O2G__Ap
                        ? mat.get_biochem_param(2999)*normal_distro(1., 0.02) : 0.0;
        //
        Grad_Crd(14) += diff*Grad_TAMP*PHI;
        //
        const Real duro = TAMP>Cell__LB && O2G>O2G__Ap
                        ? mat.get_biochem_param(2998)*normal_distro(1., 0.02) : 0.0;
        //
        Grad_Crd(14) += duro*TAMP*Grad_ECM*PHI;
        //
        const Real chemtx1 = mat.get_biochem_param(2997)*normal_distro(1., 0.02),
                   chemtx2 = mat.get_biochem_param(2996)*normal_distro(1., 0.02),
                     C_O2G = TAMP>Cell__LB && O2G>O2G__Ap && (chemtx2+pow4(O2G))
                           ? chemtx1*(chemtx2+1.0)*pow4(O2G)/(chemtx2+pow4(O2G)) : 0.0;
        //
        Grad_Crd(14) += C_O2G*TAMP*Grad_O2G*PHI;
        //
    }
    // ###############################################
    // ### state-variable 15: Necrotic Tumour Cell ###
    // ###############################################
    {
        const Real O2G__Nc = mat.get_biochem_param(2200); // @elsewhere
        const Real O2G__Ap = mat.get_biochem_param(3000);
        //
        const Real delta1 = mat.get_biochem_param(3001)*normal_distro(1., 0.02),
                   A_O2G = O2G<=O2G__Ap && NTC>Cell__LB
                         ? delta1 : 0.0;
        //
        Crd(15) -= A_O2G*NTC;
        //
        const Real omega1 = mat.get_biochem_param(3011)*normal_distro(1., 0.02),
                   W_TCN  = TCN>Cell__LB && O2G<O2G__Nc
                          ? omega1*TCN : 0.0,
                   omega2 = mat.get_biochem_param(3012)*normal_distro(1., 0.02),
                   W_TCH  = TCH>Cell__LB && O2G<O2G__Nc
                          ? omega2*TCH : 0.0,
                   omega3 = mat.get_biochem_param(3013)*normal_distro(1., 0.02),
                   W_TASC = TASC>Cell__LB && O2G<O2G__Nc
                          ? omega3*TASC : 0.0,
                   omega4 = mat.get_biochem_param(3014)*normal_distro(1., 0.02),
                   W_TAFB = TAFB>Cell__LB && O2G<O2G__Nc
                          ? omega4*TAFB : 0.0,
                   omega5 = mat.get_biochem_param(3015)*normal_distro(1., 0.02),
                   W_TAMP = TAMP>Cell__LB && O2G<O2G__Nc
                          ? omega5*TAMP : 0.0;
        //
        Crd(15) += W_TCN;
        Crd(15) += W_TCH;
        Crd(15) += W_TASC;
        Crd(15) += W_TAFB;
        Crd(15) += W_TAMP;
        //
    }
    // #############################################################
    // ### state-variable 16: Vascular Endothelial Growth Factor ###
    // #############################################################
    {
        const Real GF__Thr = mat.get_biochem_param(3200);
        //
        const Real diff = VEGF>GF__Thr
                        ? mat.get_biochem_param(3399) : 0.0;
        //
        Grad_Crd(16) += diff*Grad_VEGF;
    }
    // #########################################
    // ### state-variable 17: Angiopoietin 1 ###
    // #########################################
    {
        const Real GF__Thr = mat.get_biochem_param(3400);
        //
        const Real diff = ANG1>GF__Thr
                        ? mat.get_biochem_param(3599) : 0.0;
        //
        Grad_Crd(17) += diff*Grad_ANG1;
    }
    // #########################################
    // ### state-variable 18: Angiopoietin 2 ###
    // #########################################
    {
        const Real GF__Thr = mat.get_biochem_param(3600);
        //
        const Real diff = ANG2>GF__Thr
                        ? mat.get_biochem_param(3799) : 0.0;
        //
        Grad_Crd(18) += diff*Grad_ANG2;
    }
    // ##################################################
    // ### state-variable 19: Epidermal Growth Factor ###
    // ##################################################
    {
        const Real GF__Thr = mat.get_biochem_param(3800);
        //
        const Real diff = EGF>GF__Thr
                        ? mat.get_biochem_param(3999) : 0.0;
        //
        Grad_Crd(19) += diff*Grad_EGF;
    }
    // #############################################################
    // ### state-variable 20: Transforming Growth Factor `alpha` ###
    // #############################################################
    {
        const Real GF__Thr = mat.get_biochem_param(4000);
        //
        const Real diff = TGFa>GF__Thr
                        ? mat.get_biochem_param(4199) : 0.0;
        //
        Grad_Crd(20) += diff*Grad_TGFa;
    }
    // ############################################################
    // ### state-variable 21: Transforming Growth Factor `beta` ###
    // ############################################################
    {
        const Real GF__Thr = mat.get_biochem_param(4200);
        //
        const Real diff = TGFb>GF__Thr
                        ? mat.get_biochem_param(4399) : 0.0;
        //
        Grad_Crd(21) += diff*Grad_TGFb;
    }
    // ###########################################################
    // ### state-variable 22: Macrophage-derived Growth Factor ###
    // ###########################################################
    {
        const Real GF__Thr = mat.get_biochem_param(4400);
        //
        const Real diff = MDGF>GF__Thr
                        ? mat.get_biochem_param(4599) : 0.0;
        //
        Grad_Crd(22) += diff*Grad_MDGF;
    }
    // ###################################################
    // ### state-variable 23: Matrix Metalloproteinase ###
    // ###################################################
    {
        const Real GF__Thr = mat.get_biochem_param(4600);
        //
        const Real diff = MMP>GF__Thr
                        ? mat.get_biochem_param(4799) : 0.0;
        //
        Grad_Crd(23) += diff*Grad_MMP;
    }
    // ##########################################
    // ### state-variable 24: Interleukin `X` ###
    // ##########################################
    {
        const Real GF__Thr = mat.get_biochem_param(4800);
        //
        const Real diff = ILX>GF__Thr
                        ? mat.get_biochem_param(4999) : 0.0;
        //
        Grad_Crd(24) += diff*Grad_ILX;
    }
    // ###################################################
    // ### state-variable 25: VEGFR                    ###
    // ###################################################
    {
        const Real rcpt_MSC  = mat.get_biochem_param(5011)*normal_distro(1., 0.01),
                   rcpt_FB   = mat.get_biochem_param(5012)*normal_distro(1., 0.01),
                   rcpt_EC   = mat.get_biochem_param(5013)*normal_distro(1., 0.01),
                   rcpt_ICTL = mat.get_biochem_param(5014)*normal_distro(1., 0.01),
                   rcpt_ICMP = mat.get_biochem_param(5015)*normal_distro(1., 0.01),
                   rcpt_TCN  = mat.get_biochem_param(5016)*normal_distro(1., 0.01),
                   rcpt_TCH  = mat.get_biochem_param(5017)*normal_distro(1., 0.01),
                   rcpt_TASC = mat.get_biochem_param(5018)*normal_distro(1., 0.01),
                   rcpt_TAFB = mat.get_biochem_param(5019)*normal_distro(1., 0.01),
                   rcpt_TAMP = mat.get_biochem_param(5020)*normal_distro(1., 0.01),
                   rcpt_NTC  = mat.get_biochem_param(5021)*normal_distro(1., 0.01);
        //
        Crd(24) += rcpt_MSC*Rate_MSC + rcpt_FB*Rate_FB + rcpt_EC*Rate_EC + rcpt_ICTL*Rate_ICTL + rcpt_ICMP*Rate_ICMP
                 + rcpt_TCN*Rate_TCN + rcpt_TCH*Rate_TCH + rcpt_TASC*Rate_TASC + rcpt_TAFB*Rate_TAFB + rcpt_TAMP*Rate_TAMP + rcpt_NTC*Rate_NTC;
    }
    // ###################################################
    // ### state-variable 26: TIE1                     ###
    // ###################################################
    {
        const Real rcpt_MSC  = mat.get_biochem_param(5111)*normal_distro(1., 0.01),
                   rcpt_FB   = mat.get_biochem_param(5112)*normal_distro(1., 0.01),
                   rcpt_EC   = mat.get_biochem_param(5113)*normal_distro(1., 0.01),
                   rcpt_ICTL = mat.get_biochem_param(5114)*normal_distro(1., 0.01),
                   rcpt_ICMP = mat.get_biochem_param(5115)*normal_distro(1., 0.01),
                   rcpt_TCN  = mat.get_biochem_param(5116)*normal_distro(1., 0.01),
                   rcpt_TCH  = mat.get_biochem_param(5117)*normal_distro(1., 0.01),
                   rcpt_TASC = mat.get_biochem_param(5118)*normal_distro(1., 0.01),
                   rcpt_TAFB = mat.get_biochem_param(5119)*normal_distro(1., 0.01),
                   rcpt_TAMP = mat.get_biochem_param(5120)*normal_distro(1., 0.01),
                   rcpt_NTC  = mat.get_biochem_param(5121)*normal_distro(1., 0.01);
        //
        Crd(25) += rcpt_MSC*Rate_MSC + rcpt_FB*Rate_FB + rcpt_EC*Rate_EC + rcpt_ICTL*Rate_ICTL + rcpt_ICMP*Rate_ICMP
                 + rcpt_TCN*Rate_TCN + rcpt_TCH*Rate_TCH + rcpt_TASC*Rate_TASC + rcpt_TAFB*Rate_TAFB + rcpt_TAMP*Rate_TAMP + rcpt_NTC*Rate_NTC;
    }
    // ###################################################
    // ### state-variable 27: TIE2                     ###
    // ###################################################
    {
        const Real rcpt_MSC  = mat.get_biochem_param(5211)*normal_distro(1., 0.01),
                   rcpt_FB   = mat.get_biochem_param(5212)*normal_distro(1., 0.01),
                   rcpt_EC   = mat.get_biochem_param(5213)*normal_distro(1., 0.01),
                   rcpt_ICTL = mat.get_biochem_param(5214)*normal_distro(1., 0.01),
                   rcpt_ICMP = mat.get_biochem_param(5215)*normal_distro(1., 0.01),
                   rcpt_TCN  = mat.get_biochem_param(5216)*normal_distro(1., 0.01),
                   rcpt_TCH  = mat.get_biochem_param(5217)*normal_distro(1., 0.01),
                   rcpt_TASC = mat.get_biochem_param(5218)*normal_distro(1., 0.01),
                   rcpt_TAFB = mat.get_biochem_param(5219)*normal_distro(1., 0.01),
                   rcpt_TAMP = mat.get_biochem_param(5220)*normal_distro(1., 0.01),
                   rcpt_NTC  = mat.get_biochem_param(5221)*normal_distro(1., 0.01);
        //
        Crd(26) += rcpt_MSC*Rate_MSC + rcpt_FB*Rate_FB + rcpt_EC*Rate_EC + rcpt_ICTL*Rate_ICTL + rcpt_ICMP*Rate_ICMP
                 + rcpt_TCN*Rate_TCN + rcpt_TCH*Rate_TCH + rcpt_TASC*Rate_TASC + rcpt_TAFB*Rate_TAFB + rcpt_TAMP*Rate_TAMP + rcpt_NTC*Rate_NTC;
    }
    // ###################################################
    // ### state-variable 28: EGFR                     ###
    // ###################################################
    {
        const Real rcpt_MSC  = mat.get_biochem_param(5311)*normal_distro(1., 0.01),
                   rcpt_FB   = mat.get_biochem_param(5312)*normal_distro(1., 0.01),
                   rcpt_EC   = mat.get_biochem_param(5313)*normal_distro(1., 0.01),
                   rcpt_ICTL = mat.get_biochem_param(5314)*normal_distro(1., 0.01),
                   rcpt_ICMP = mat.get_biochem_param(5315)*normal_distro(1., 0.01),
                   rcpt_TCN  = mat.get_biochem_param(5316)*normal_distro(1., 0.01),
                   rcpt_TCH  = mat.get_biochem_param(5317)*normal_distro(1., 0.01),
                   rcpt_TASC = mat.get_biochem_param(5318)*normal_distro(1., 0.01),
                   rcpt_TAFB = mat.get_biochem_param(5319)*normal_distro(1., 0.01),
                   rcpt_TAMP = mat.get_biochem_param(5320)*normal_distro(1., 0.01),
                   rcpt_NTC  = mat.get_biochem_param(5321)*normal_distro(1., 0.01);
        //
        Crd(27) += rcpt_MSC*Rate_MSC + rcpt_FB*Rate_FB + rcpt_EC*Rate_EC + rcpt_ICTL*Rate_ICTL + rcpt_ICMP*Rate_ICMP
                 + rcpt_TCN*Rate_TCN + rcpt_TCH*Rate_TCH + rcpt_TASC*Rate_TASC + rcpt_TAFB*Rate_TAFB + rcpt_TAMP*Rate_TAMP + rcpt_NTC*Rate_NTC;
    }
    // ###################################################
    // ### state-variable 29: TGFR                     ###
    // ###################################################
    {
        const Real rcpt_MSC  = mat.get_biochem_param(5411)*normal_distro(1., 0.01),
                   rcpt_FB   = mat.get_biochem_param(5412)*normal_distro(1., 0.01),
                   rcpt_EC   = mat.get_biochem_param(5413)*normal_distro(1., 0.01),
                   rcpt_ICTL = mat.get_biochem_param(5414)*normal_distro(1., 0.01),
                   rcpt_ICMP = mat.get_biochem_param(5415)*normal_distro(1., 0.01),
                   rcpt_TCN  = mat.get_biochem_param(5416)*normal_distro(1., 0.01),
                   rcpt_TCH  = mat.get_biochem_param(5417)*normal_distro(1., 0.01),
                   rcpt_TASC = mat.get_biochem_param(5418)*normal_distro(1., 0.01),
                   rcpt_TAFB = mat.get_biochem_param(5419)*normal_distro(1., 0.01),
                   rcpt_TAMP = mat.get_biochem_param(5420)*normal_distro(1., 0.01),
                   rcpt_NTC  = mat.get_biochem_param(5421)*normal_distro(1., 0.01);
        //
        Crd(28) += rcpt_MSC*Rate_MSC + rcpt_FB*Rate_FB + rcpt_EC*Rate_EC + rcpt_ICTL*Rate_ICTL + rcpt_ICMP*Rate_ICMP
                 + rcpt_TCN*Rate_TCN + rcpt_TCH*Rate_TCH + rcpt_TASC*Rate_TASC + rcpt_TAFB*Rate_TAFB + rcpt_TAMP*Rate_TAMP + rcpt_NTC*Rate_NTC;
    }
    // ###################################################
    // ### state-variable 30: WNT                      ###
    // ###################################################
    {
        const Real rcpt_MSC  = mat.get_biochem_param(5511)*normal_distro(1., 0.01),
                   rcpt_FB   = mat.get_biochem_param(5512)*normal_distro(1., 0.01),
                   rcpt_EC   = mat.get_biochem_param(5513)*normal_distro(1., 0.01),
                   rcpt_ICTL = mat.get_biochem_param(5514)*normal_distro(1., 0.01),
                   rcpt_ICMP = mat.get_biochem_param(5515)*normal_distro(1., 0.01),
                   rcpt_TCN  = mat.get_biochem_param(5516)*normal_distro(1., 0.01),
                   rcpt_TCH  = mat.get_biochem_param(5517)*normal_distro(1., 0.01),
                   rcpt_TASC = mat.get_biochem_param(5518)*normal_distro(1., 0.01),
                   rcpt_TAFB = mat.get_biochem_param(5519)*normal_distro(1., 0.01),
                   rcpt_TAMP = mat.get_biochem_param(5520)*normal_distro(1., 0.01),
                   rcpt_NTC  = mat.get_biochem_param(5521)*normal_distro(1., 0.01);
        //
        Crd(29) += rcpt_MSC*Rate_MSC + rcpt_FB*Rate_FB + rcpt_EC*Rate_EC + rcpt_ICTL*Rate_ICTL + rcpt_ICMP*Rate_ICMP
                 + rcpt_TCN*Rate_TCN + rcpt_TCH*Rate_TCH + rcpt_TASC*Rate_TASC + rcpt_TAFB*Rate_TAFB + rcpt_TAMP*Rate_TAMP + rcpt_NTC*Rate_NTC;
    }
    // ###################################################
    // ### state-variable 31: HDGH                     ###
    // ###################################################
    {
        const Real rcpt_MSC  = mat.get_biochem_param(5611)*normal_distro(1., 0.01),
                   rcpt_FB   = mat.get_biochem_param(5612)*normal_distro(1., 0.01),
                   rcpt_EC   = mat.get_biochem_param(5613)*normal_distro(1., 0.01),
                   rcpt_ICTL = mat.get_biochem_param(5614)*normal_distro(1., 0.01),
                   rcpt_ICMP = mat.get_biochem_param(5615)*normal_distro(1., 0.01),
                   rcpt_TCN  = mat.get_biochem_param(5616)*normal_distro(1., 0.01),
                   rcpt_TCH  = mat.get_biochem_param(5617)*normal_distro(1., 0.01),
                   rcpt_TASC = mat.get_biochem_param(5618)*normal_distro(1., 0.01),
                   rcpt_TAFB = mat.get_biochem_param(5619)*normal_distro(1., 0.01),
                   rcpt_TAMP = mat.get_biochem_param(5620)*normal_distro(1., 0.01),
                   rcpt_NTC  = mat.get_biochem_param(5621)*normal_distro(1., 0.01);
        //
        Crd(30) += rcpt_MSC*Rate_MSC + rcpt_FB*Rate_FB + rcpt_EC*Rate_EC + rcpt_ICTL*Rate_ICTL + rcpt_ICMP*Rate_ICMP
                 + rcpt_TCN*Rate_TCN + rcpt_TCH*Rate_TCH + rcpt_TASC*Rate_TASC + rcpt_TAFB*Rate_TAFB + rcpt_TAMP*Rate_TAMP + rcpt_NTC*Rate_NTC;
    }
    // ###################################################
    // ### state-variable 32: NTCH                     ###
    // ###################################################
    {
        const Real rcpt_MSC  = mat.get_biochem_param(5711)*normal_distro(1., 0.01),
                   rcpt_FB   = mat.get_biochem_param(5712)*normal_distro(1., 0.01),
                   rcpt_EC   = mat.get_biochem_param(5713)*normal_distro(1., 0.01),
                   rcpt_ICTL = mat.get_biochem_param(5714)*normal_distro(1., 0.01),
                   rcpt_ICMP = mat.get_biochem_param(5715)*normal_distro(1., 0.01),
                   rcpt_TCN  = mat.get_biochem_param(5716)*normal_distro(1., 0.01),
                   rcpt_TCH  = mat.get_biochem_param(5717)*normal_distro(1., 0.01),
                   rcpt_TASC = mat.get_biochem_param(5718)*normal_distro(1., 0.01),
                   rcpt_TAFB = mat.get_biochem_param(5719)*normal_distro(1., 0.01),
                   rcpt_TAMP = mat.get_biochem_param(5720)*normal_distro(1., 0.01),
                   rcpt_NTC  = mat.get_biochem_param(5721)*normal_distro(1., 0.01);
        //
        Crd(31) += rcpt_MSC*Rate_MSC + rcpt_FB*Rate_FB + rcpt_EC*Rate_EC + rcpt_ICTL*Rate_ICTL + rcpt_ICMP*Rate_ICMP
                 + rcpt_TCN*Rate_TCN + rcpt_TCH*Rate_TCH + rcpt_TASC*Rate_TASC + rcpt_TAFB*Rate_TAFB + rcpt_TAMP*Rate_TAMP + rcpt_NTC*Rate_NTC;
    }
}
//=================================================================================================
} // end of namespace invitro_tumoroid_growth...
//=================================================================================================
//
//
//
//=================================================================================================
namespace radiotherapy_tumour_fibrosis {
//=================================================================================================
void damaged_region_311 ( _LOCAL_ARGUMENTS_ ) {
    const unsigned int nvars = 3;
    Crd.resize(nvars);
    Crd = 0.0;
    Grad_Crd.resize(nvars);
    Grad_Crd = RealGradient(0.0, 0.0, 0.0);
    //
    if ( -1 == fld.a_flag ) return;
    // check which version of this model is operated
    const int model_version = mat.get_biochem_param(1000);
    //
    if ( 0 == model_version ) {
        //
        const Real tcell = fld.biochem[0],
                   fibrs = fld.biochem[1],
                   ecm   = fld.biochem[2],
                   rtd   = fld.auxiliary[0];
        const RealGradient Grad_tcell = fld.grad_biochem[0],
                           Grad_fibrs = fld.grad_biochem[1],
                           Grad_ecm   = fld.grad_biochem[2],
                           Grad_rtd   = fld.grad_auxiliary[0];
        //
        const Real tcell_rate = fld.biochem_rate[0],
                   fibrs_rate = fld.biochem_rate[1],
                   ecm_rate   = fld.biochem_rate[2];
        //
        const Real day  = fld.current_time/86400.0,
                   week = day/7.0,
                   //
                   rtd_max = mat.get_biochem_param(0) * _VARIABILITY_1pct_,
                   R_ = (rtd>0.0 ? (rtd_max>0.0 ? rtd/rtd_max : 1.0) : 0.0),
                   //
                   fraction = mat.get_biochem_param(1),
                   duration = mat.get_biochem_param(2),
                   //
                   alpha    = mat.get_biochem_param(11),
                   beta     = mat.get_biochem_param(12),
                   delta_t  = mat.get_biochem_param(13) * _VARIABILITY_5pct_,
                   //
                   D_t      = mat.get_biochem_param(14) * _VARIABILITY_2pct_,
                   //
                   gamma    = mat.get_biochem_param(21),
                   lambda_fr= mat.get_biochem_param(22) * _VARIABILITY_5pct_,
                   lambda_fe= mat.get_biochem_param(23) * _VARIABILITY_5pct_,
                   //
                   D_f      = mat.get_biochem_param(24) * _VARIABILITY_2pct_,
                   C_f      = mat.get_biochem_param(25) * _VARIABILITY_2pct_,
                   X_f      = mat.get_biochem_param(26) * _VARIABILITY_2pct_,
                   //
                   HU       = apply_bounds(-1000.0, ecm, +1000.0),
                   lambda_et= ((mat.get_biochem_param(31)<HU && HU<mat.get_biochem_param(32)) ? mat.get_biochem_param(33) : 0.0) * _VARIABILITY_5pct_,
                   lambda_ef= ((mat.get_biochem_param(34)<HU && HU<mat.get_biochem_param(35)) ? mat.get_biochem_param(36) : 0.0) * _VARIABILITY_5pct_,
                   //
                   ac0 = mat.get_biochem_param(41),
                   ac1 = mat.get_biochem_param(42),
                   //
                   H_ = apply_bounds(0.0, 1.0-tcell-fibrs, 1.0);
        //
        const Real TCSF = exp(-alpha*rtd-beta*rtd*rtd*fraction); // tumour cells survival fraction
        //
        // const Real FRIP = R_ * exp(-gamma*week); // fibroblasts radiation-induced production
        const Real FRIP = R_ * exp(-gamma*(duration/fraction)*week); // fibroblasts radiation-induced production
        const Real ECMr = (-1.0) / apply_bounds(-900.0, ecm, -20.0); // ECM reciprocal
        //
        Real ECMR_tcell = 0.0;
        if      ( 0.0 == ac0 ) ;
        else if (-1.0 == ac0 ) ECMR_tcell = tcell_rate;
        else if ( tcell>0.0 ) ECMR_tcell = (exp(-ac0*tcell)-1.0)/(exp(-ac0)-1.0);
        //
        Real ECMR_fibrs = 0.0;
        if      ( 0.0 == ac1 ) ;
        else if (-1.0 == ac1 ) ECMR_fibrs = fibrs_rate;
        else if ( fibrs>0.0 && fibrs<2*ac1 ) ECMR_fibrs = (2.0*fibrs/ac1-pow2(fibrs/ac1));
        //
        Crd(0) = - delta_t * tcell * (1.0-TCSF);
        Crd(1) = lambda_fr * (1.0-fibrs) * FRIP + lambda_fe * fibrs * ECMr;
        Crd(2) = lambda_et * ECMR_tcell + lambda_ef * ECMR_fibrs;
        //
        Grad_Crd(0) = H_ * D_t * Grad_tcell;
        Grad_Crd(1) = H_ * D_f * Grad_fibrs + H_ * R_ * C_f * Grad_fibrs + H_ * R_ * X_f * fibrs * Grad_ecm.unit();
        //
    } else if ( 1 == model_version ) {
        //
        const Real tcell = fld.biochem[0],
                   fibrs = fld.biochem[1],
                   ecm   = fld.biochem[2];
        const RealGradient Grad_tcell = fld.grad_biochem[0],
                           Grad_fibrs = fld.grad_biochem[1],
                           Grad_ecm   = fld.grad_biochem[2];
        //
        const Real T = fld.current_time; // time is in days
        //
        const Real Reference_Vol = (mat.get_biochem_param(0)>0.0 ? mat.get_biochem_param(0) : 1.0),
                   Stroma_VolFr     = mat.get_biochem_param(1),
                   Parenchyma_VolFr = mat.get_biochem_param(2),
                   tcell_VolFr = mat.get_biochem_param(3)/Reference_Vol,
                   fibrs_VolFr = mat.get_biochem_param(4)/Reference_Vol;
        //
        const Real cells_VolFr  = (tcell_VolFr*tcell) + (fibrs_VolFr*fibrs);
        const Real VolFr_exponent = mat.get_biochem_param(8);
        Real vacant_VolFr = pow(1.0-Stroma_VolFr-Parenchyma_VolFr-cells_VolFr, VolFr_exponent);
        if ( vacant_VolFr <= mat.get_biochem_param(9) ) vacant_VolFr = 0.0;
        if ( vacant_VolFr >= 1.0e+0 ) libmesh_error();
        //
        const int RT_fractions = static_cast<int>( mat.get_biochem_param(10) );
        if ( RT_fractions < 1 ) libmesh_error();
        const Real alpha = mat.get_biochem_param(11),
                   beta  = mat.get_biochem_param(12),
                   gamma = mat.get_biochem_param(13),
                   delta = mat.get_biochem_param(14);
        //
        const Real R = fld.auxiliary[0], // total radiation dose
                   r = R / RT_fractions; // radiation dose per fraction
        //
        const Real P__tcell = (uniform_distro(0.0, 1.0)<=exp(-gamma*R) ? 1.0 : 0.0);
        const Real P__fibrs = (uniform_distro(0.0, 1.0)<=exp(-delta*T) ? 1.0 : 0.0);
        const Real D__tcell = 1.0
                            - (uniform_distro(0.0, 1.0)<=exp(-alpha*R-beta*R*r) ? 1.0 : 0.0);
        const Real D__fibrs = 1.0;
        //
        const Real lambda_0 = mat.get_biochem_param(20)*normal_distro(1., 0.01),
                   tcell_r  = mat.get_biochem_param(21),
                   delta_0  = mat.get_biochem_param(22)*normal_distro(1., 0.01);
        const Real lambda_1 = mat.get_biochem_param(30)*normal_distro(1., 0.01),
                   fibrs_r  = mat.get_biochem_param(31),
                   delta_1  = mat.get_biochem_param(32)*normal_distro(1., 0.01);
        const Real phi_0 = mat.get_biochem_param(40)*normal_distro(1., 0.01),
                   phi_1 = mat.get_biochem_param(41)*normal_distro(1., 0.01);
        //
        const int TC_ecm__type = static_cast<int>( mat.get_biochem_param(50) );
        Real TC_ecm;
        if ( 0 == TC_ecm__type ) {
            TC_ecm = 1.0;
        } else if ( 1 == TC_ecm__type ) {
            const Real ecm_r = mat.get_biochem_param(51);
            if ( ecm_r == 0.0e+0 ) libmesh_error();
            TC_ecm = (ecm_r/ecm);
        } else libmesh_error();
        //
        const int TC_R__type = static_cast<int>( mat.get_biochem_param(60) );
        Real TC_R;
        if ( 0 == TC_R__type ) {
            TC_R = 1.0;
        } else if ( 1 == TC_R__type ) {
            const Real R_r = mat.get_biochem_param(61);
            if ( R_r == 0.0e+0 ) libmesh_error();
            TC_R = 1.0-(R/R_r);
        } else libmesh_error();
        //
        const int FB_ecm__type = static_cast<int>( mat.get_biochem_param(70) );
        Real FB_ecm;
        if ( 0 == FB_ecm__type ) {
            FB_ecm = 1.0;
        } else if ( 1 == FB_ecm__type ) {
            const Real ecm_r = mat.get_biochem_param(71);
            if ( ecm_r == 0.0e+0 ) libmesh_error();
            FB_ecm = (ecm_r/ecm);
        } else libmesh_error();
        //
        const int FB_R__type = static_cast<int>( mat.get_biochem_param(80) );
        Real FB_R;
        if ( 0 == FB_R__type ) {
            FB_R = 1.0;
        } else if ( 1 == FB_R__type ) {
            const Real R_r = mat.get_biochem_param(81);
            if ( R_r == 0.0e+0 ) libmesh_error();
            FB_R = (R/R_r);
        } else libmesh_error();
        //
        const int ECM__ranges = static_cast<int>( mat.get_biochem_param(90) );
        Real ECM_rm = 1.0;
        if ( ECM__ranges <= 3 ) {
            for (int I=0; I<ECM__ranges; I++) {
                const int J = 91+(3*I);
                const Real ecm_min = mat.get_biochem_param(J+0),
                           ecm_max = mat.get_biochem_param(J+1);
                if ( ! ( ecm_min <= ecm && ecm <= ecm_max ) ) continue;
                const Real ecm_r = mat.get_biochem_param(J+2);
                if ( ecm_r == 0.0e+0 ) libmesh_error();
                ECM_rm = (ecm/ecm_r);
            }
        } else libmesh_error();
        //
        const Real Prod_tcell = (4.0*tcell/tcell_r) * ((tcell_r-tcell)/tcell_r) * TC_ecm * TC_R;
        const Real Prod_fibrs = (4.0*fibrs/fibrs_r) * ((fibrs_r-fibrs)/fibrs_r) * FB_ecm * FB_R;
        const Real Prod_ecm_tcell = pow(tcell/tcell_r, mat.get_biochem_param(100)) * pow(ECM_rm, mat.get_biochem_param(102));
        const Real Prod_ecm_fibrs = pow(fibrs/fibrs_r, mat.get_biochem_param(101)) * pow(ECM_rm, mat.get_biochem_param(103));
        //
        const Real Diff_tcell_ = (tcell>0.0 && mat.get_biochem_param(110) ? mat.get_biochem_param(110) : 0.0),
                   Diff_fibrs_ = (fibrs>0.0 && mat.get_biochem_param(111) ? mat.get_biochem_param(111) : 0.0),
                   Hapto_tcell = (tcell>0.0 && mat.get_biochem_param(112) ? mat.get_biochem_param(112)*normal_distro(1., 0.025) : 0.0),
                   Hapto_fibrs = (fibrs>0.0 && mat.get_biochem_param(113) ? mat.get_biochem_param(113)*normal_distro(1., 0.025) : 0.0);
        //
        Dyadic3D Diff_tcell;
        for (int i=0; i<3; i++) Diff_tcell.comp(i,i) = Diff_tcell_*normal_distro(1., 0.025);
        Dyadic3D Diff_fibrs;
        for (int i=0; i<3; i++) Diff_fibrs.comp(i,i) = Diff_fibrs_*normal_distro(1., 0.025);
        //
        Crd(0) = lambda_0 * Prod_tcell * vacant_VolFr * P__tcell - delta_0 * tcell * D__tcell;
        //
        Crd(1) = lambda_1 * Prod_fibrs * vacant_VolFr * P__fibrs - delta_1 * fibrs * D__fibrs;
        //
        Crd(2) = phi_0 * Prod_ecm_tcell + phi_1 * Prod_ecm_fibrs;
        //
        if ( Diff_tcell_ ) Grad_Crd(0) += Diff_tcell * Grad_tcell * vacant_VolFr;
        if ( Hapto_tcell ) Grad_Crd(0) += Hapto_tcell * tcell * Grad_ecm * vacant_VolFr;
        //
        if ( Diff_fibrs_ ) Grad_Crd(1) += Diff_fibrs * Grad_fibrs * vacant_VolFr;
        if ( Hapto_fibrs ) Grad_Crd(1) += Hapto_fibrs * fibrs * Grad_ecm * vacant_VolFr;
        //
    } else libmesh_error();
}
//=================================================================================================
void healthy_region_311 ( _LOCAL_ARGUMENTS_ ) {
    radiotherapy_tumour_fibrosis::damaged_region_311(mat, fld, Crd, Grad_Crd, Ksm, Velc, Diff, Taxs);
}
//=================================================================================================
} // end of namespace radiotherapy_tumour_fibrosis...
//=================================================================================================
//
//
//
//=================================================================================================
namespace tumour_angiogenesis {
//=================================================================================================
void damaged_tissue_498 ( _LOCAL_ARGUMENTS_ ) {
    const unsigned int nvars = 4;
    Crd.resize(nvars);
    Crd = 0.0;
    Grad_Crd.resize(nvars);
    Grad_Crd = RealGradient(0.0, 0.0, 0.0);
    Ksm.resize(nvars, nvars);
    Ksm = 0.0;
    Velc.resize(nvars);
    Velc = Point(0.0, 0.0, 0.0);
    Diff.resize(nvars, nvars);
    Diff = 0.0;
    Taxs.resize(nvars, nvars);
    Taxs = RealGradient(0.0, 0.0, 0.0);
    //
    const Real taf  = fld.biochem[0],
               oxyg = fld.biochem[1],
               mde  = fld.biochem[2],
               ecm  = fld.biochem[3];
    const Real //blv__dens = fld.biochem[nvars+0], // discontinuous field variables
               //lyv__dens = fld.biochem[nvars+1],
               blv__tip  = fld.biochem[nvars+2],
               lyv__tip  = fld.biochem[nvars+3],
               //blv__func = fld.biochem[nvars+4],
               //lyv__func = fld.biochem[nvars+5],
               blv__perf = fld.biochem[nvars+6],
               lyv__perf = fld.biochem[nvars+7],
               temp_load = fld.biochem[nvars+8];
    const RealGradient Grad_taf  = fld.grad_biochem[0],
                       Grad_oxyg = fld.grad_biochem[1],
                       Grad_mde  = fld.grad_biochem[2],
                       Grad_ecm  = fld.grad_biochem[3];
    //
    const Point Veloc_taf  = ( 1==mat.get_biochem_param(20) || 3==mat.get_biochem_param(20) || 5==mat.get_biochem_param(20) || 7==mat.get_biochem_param(20) )
                           ? coordinates(fld.fluid_velocity) : zero_coordinates;
    const Point Veloc_oxyg = ( 2==mat.get_biochem_param(20) || 3==mat.get_biochem_param(20) || 6==mat.get_biochem_param(20) || 7==mat.get_biochem_param(20) )
                           ? coordinates(fld.fluid_velocity) : zero_coordinates;
    const Point Veloc_mde  = ( 4==mat.get_biochem_param(20) || 5==mat.get_biochem_param(20) || 6==mat.get_biochem_param(20) || 7==mat.get_biochem_param(20) )
                           ? coordinates(fld.fluid_velocity) : zero_coordinates;
    //
    const Real drug0 = apply_lbound(0.0, fld.auxiliary[3]),
               drug1 = apply_lbound(0.0, fld.auxiliary[7]);
    //
    if ( -1 == fld.a_flag ) return;
    //
    const Real vsc = apply_lbound(0.0, blv__perf), // normalised density of perfused vessels
               tip = apply_lbound(0.0, blv__tip), // normalised density of tip vessels
               //
               taf__LB  = mat.get_biochem_param(180), // lowest bound value
               taf__MX  = mat.get_biochem_param(181), // maximum value
               oxyg__LB = mat.get_biochem_param(182), // lowest bound value
               oxyg__MX = mat.get_biochem_param(183), // maximum value
               mde__LB  = mat.get_biochem_param(184), // lowest bound value
               mde__MX  = mat.get_biochem_param(185), // maximum value
               ecm__LB  = mat.get_biochem_param(186), // lowest bound value
               ecm__MX  = mat.get_biochem_param(187), // maximum value
               //
               oxyg_taf_cutoff = mat.get_biochem_param(6), // O2 threshold for TAF production
               ecm_cutoff      = mat.get_biochem_param(7),  // ECM threshold for remodelling
               ecm_taf_cutoff  = mat.get_biochem_param(19), // ECM threshold for TAF production
               //
               ecm_drug0_cutoff = mat.get_biochem_param(143), // ECM threshold to allow matrix/cells depletion from drug
               ecm_drug1_cutoff = mat.get_biochem_param(153), // ECM threshold to allow matrix/cells depletion from drug
               //
               lambda_taf   = mat.get_biochem_param(0),
               delta_taf    = ( taf>taf__LB ? mat.get_biochem_param(1) : 0.0 ),
               Diff_taf     = mat.get_biochem_param(2) * normal_distro(1., 0.20), // add some variability to this...
               //
               lambda_oxyg  = mat.get_biochem_param(3),
               delta_oxyg   = ( oxyg>oxyg__LB ? mat.get_biochem_param(4) : 0.0 ),
               Diff_oxyg    = mat.get_biochem_param(5) * normal_distro(1., 0.10), // add some variability to this...
               //
               Q_oxyg       = ( oxyg>oxyg__LB ? exp(-2.0*oxyg/oxyg_taf_cutoff) : 1.0 ),
               dQ_oxyg      = ( oxyg>oxyg__LB ? (-2.0/oxyg_taf_cutoff)*Q_oxyg  : 0.0 ),
               //
               Q_ecm        = ( ecm>ecm__LB ? pow(ecm, ecm_taf_cutoff) : 0.0 ),
               dQ_ecm       = ( ecm>ecm__LB ? pow(ecm, ecm_taf_cutoff-1)*ecm_taf_cutoff : 0.0 ),
               //
               O2_vsc       = ( oxyg<oxyg__MX ? vsc : 0.0 ),
               //
               lambda_mde_c = ( ecm>ecm__LB ? mat.get_biochem_param(8) : 0.0 ),
               lambda_mde_v = mat.get_biochem_param(9),
               delta_mde    = ( mde>mde__LB ? mat.get_biochem_param(10) : 0.0 ),
               Diff_mde     = mat.get_biochem_param(11) * normal_distro(1., 0.20), // add some variability to this...
               //
               lambda_ecm   = mat.get_biochem_param(12),
               delta_ecm    = ( ecm>ecm__LB && mde>0.0 ? mat.get_biochem_param(13) : 0.0 ),
               delta_temp   = ( ecm>ecm__LB            ? mat.get_biochem_param(14) : 0.0 ),
               delta_drug0  = ( ecm>ecm__LB ? mat.get_biochem_param(140) : 0.0 ) * normal_distro(1., mat.get_biochem_param(144)), // add some variability to this...
               delta_drug1  = ( ecm>ecm__LB ? mat.get_biochem_param(150) : 0.0 ) * normal_distro(1., mat.get_biochem_param(154)), // add some variability to this...
               //
               // C_drug0 = mat.get_biochem_param(141)/(1.0-mat.get_biochem_param(141)), // 141: minimum damage rate when internalised-drug concentration is minimal!
               // C_drug1 = mat.get_biochem_param(151)/(1.0-mat.get_biochem_param(151)), // 151: minimum damage rate when internalised-drug concentration is minimal!
               drug0_cutoff = mat.get_biochem_param(142), // internalised-drug threshold above which matrix/cells are depleted
               drug1_cutoff = mat.get_biochem_param(152), // internalised-drug threshold above which matrix/cells are depleted
               //
               H_ecm        = ( ecm>ecm__LB && ecm<ecm__MX ? exp(-2.0*ecm/ecm_cutoff) : 0.0 ),
               dH_ecm       = ( ecm>ecm__LB && ecm<ecm__MX ? (-2.0/ecm_cutoff)*H_ecm  : 0.0 ),
               //
               H_oxyg       = apply_bounds(oxyg__LB, oxyg, oxyg__MX),
               dH_oxyg      = ( H_oxyg>0.0 ? 1.0 : 0.0 ),
               //
               // H_drug0      = ( ecm>ecm_drug0_cutoff && drug0>0.0 ? C_drug0/(C_drug0+exp(-mat.get_biochem_param(142)*drug0)) : 0.0 ),
               // H_drug1      = ( ecm>ecm_drug1_cutoff && drug1>0.0 ? C_drug1/(C_drug1+exp(-mat.get_biochem_param(152)*drug1)) : 0.0 ),
               H_drug0      = ( ecm>ecm_drug0_cutoff && drug0>drug0_cutoff ? pow(drug0-drug0_cutoff, mat.get_biochem_param(141)) : 0.0 ),
               H_drug1      = ( ecm>ecm_drug1_cutoff && drug1>drug1_cutoff ? pow(drug1-drug1_cutoff, mat.get_biochem_param(151)) : 0.0 ),
               //
               U_temp       = ( ecm>ecm__LB ? delta_temp*temp_load : 0.0 );
    //
    Crd(0) = lambda_taf*Q_oxyg*Q_ecm - delta_taf*taf - Veloc_taf*Grad_taf;
    Crd(1) = lambda_oxyg*O2_vsc - delta_oxyg*oxyg - Veloc_oxyg*Grad_oxyg;
    Crd(2) = lambda_mde_c*ecm + lambda_mde_v*tip - delta_mde*mde - Veloc_mde*Grad_mde;
    Crd(3) = lambda_ecm*H_ecm*H_oxyg - delta_ecm*mde*ecm - (delta_drug0*H_drug0+delta_drug1*H_drug1)*ecm - U_temp*ecm;
    //
    Grad_Crd(0) = Diff_taf*Grad_taf;
    Grad_Crd(1) = Diff_oxyg*Grad_oxyg;
    Grad_Crd(2) = Diff_mde*Grad_mde;
    //
    Ksm(0,0) = - delta_taf;
    Ksm(0,1) = lambda_taf*dQ_oxyg*Q_ecm;
    Ksm(0,3) = lambda_taf*Q_oxyg*dQ_ecm;
    //
    Ksm(1,1) = - delta_oxyg;
    //
    Ksm(2,2) = - delta_mde;
    Ksm(2,3) = lambda_mde_c;
    //
    Ksm(3,1) = lambda_ecm*H_ecm*dH_oxyg;
    Ksm(3,2) = - delta_ecm*ecm;
    Ksm(3,3) = lambda_ecm*dH_ecm*H_oxyg - delta_ecm*mde - (delta_drug0*H_drug0+delta_drug1*H_drug1) - U_temp;
    //
    Velc(0) = Veloc_taf;
    Velc(1) = Veloc_oxyg;
    Velc(2) = Veloc_mde;
    //
    Diff(0,0) = Diff_taf;
    Diff(1,1) = Diff_oxyg;
    Diff(2,2) = Diff_mde;
}
//=================================================================================================
void healthy_tissue_498 ( _LOCAL_ARGUMENTS_ ) {
    tumour_angiogenesis::damaged_tissue_498(mat, fld, Crd, Grad_Crd, Ksm, Velc, Diff, Taxs);
}
//=================================================================================================
} // end of namespace tumour_angiogenesis...
//=================================================================================================
//
//
//
//=================================================================================================
namespace wound_healing {
//=================================================================================================
void damaged_tissue_511 ( _LOCAL_ARGUMENTS_ ) {
    const unsigned int nvars = 5;
    Crd.resize(nvars);
    Crd = 0.0;
    Grad_Crd.resize(nvars);
    Grad_Crd = RealGradient(0.0, 0.0, 0.0);
    //
    const Real cell = fld.biochem[0],
               ecgf = fld.biochem[1],
               oxyg = fld.biochem[2],
               mdgf = fld.biochem[3],
               vasc = fld.biochem[4];
    const RealGradient Grad_cell = fld.grad_biochem[0],
                       Grad_ecgf = fld.grad_biochem[1],
                       Grad_oxyg = fld.grad_biochem[2],
                       Grad_mdgf = fld.grad_biochem[3],
                       Grad_vasc = fld.grad_biochem[4];
    //
    if ( -1 == fld.a_flag ) return;
    //
    const bool is_activator = mat.get_biochem_param(0);
    //
    const Real kappa     = mat.get_biochem_param(1),  // 1 / day
               lambda    = mat.get_biochem_param(2),  // 1 / day
               h         = mat.get_biochem_param(3),
               Diff_cell = mat.get_biochem_param(4),  // m^2 / day
               Diff_ecgf = mat.get_biochem_param(5),  // m^2 / day
               alpha     = mat.get_biochem_param(6),
               ecgf_m    = mat.get_biochem_param(7),
               beta      = is_activator ? (1.0+pow2(ecgf_m)-2.0*h*ecgf_m)/pow2(1.0-ecgf_m) : 0.0;
    const Real oxyg_m    = mat.get_biochem_param(10),
               Diff_oxyg = mat.get_biochem_param(11), // m^2 / day
               ksi_vasc  = mat.get_biochem_param(12), // 1 / day
               ksi       = mat.get_biochem_param(13), // 1 / day
               Diff_mdgf = mat.get_biochem_param(14), // m^2 / day
               mu_oxyg   = mat.get_biochem_param(15), // 1 / day
               mu        = mat.get_biochem_param(16), // 1 / day
               Diff_vasc = mat.get_biochem_param(17), // m^2 / day
               delta     = mat.get_biochem_param(18), // 1 / day
               r         = mat.get_biochem_param(19);
    const Real s_oxyg    = mat.get_biochem_param(20);
    //
    const Real S_ = is_activator ? ((2.0*ecgf_m*(h-beta)*ecgf)/(pow2(ecgf_m)+pow2(ecgf))+beta) : (((h-1.0)*ecgf+h)/(2.0*(h-1.0)*ecgf+1.0)),
               G_ = is_activator ? (cell*(1.0+pow2(alpha))/(pow2(cell)+pow2(alpha))) : cell;
    const Real Q_ = ( oxyg < oxyg_m ) ? (1.0-oxyg/oxyg_m) : 0.0,
               H_ = ((r-1.0)*mdgf+r)/(2.0*(r-1.0)*mdgf+1.0);
    const Real K_ = (oxyg < s_oxyg) ? (kappa*apply_bounds(0.0, oxyg, 1.0)/s_oxyg) : kappa;
    //
    Crd(0) = K_*S_*cell*(2.0-cell) - kappa*cell;
    Crd(1) = lambda*G_ - lambda*ecgf;
    Crd(2) = ksi_vasc*vasc - ksi*oxyg;
    Crd(3) = mu_oxyg*Q_ - mu*mdgf;
    Crd(4) = delta*H_*vasc*(1.0-vasc);
    //
    Grad_Crd(0) = Diff_cell*Grad_cell;
    Grad_Crd(1) = Diff_ecgf*Grad_ecgf;
    Grad_Crd(2) = Diff_oxyg*Grad_oxyg;
    Grad_Crd(3) = Diff_mdgf*Grad_mdgf;
    Grad_Crd(4) = Diff_vasc*Grad_vasc;
}
//=================================================================================================
void healthy_tissue_511 ( _LOCAL_ARGUMENTS_ ) {
    const unsigned int nvars = 5;
    Crd.resize(nvars);
    Crd = 0.0;
    Grad_Crd.resize(nvars);
    Grad_Crd = RealGradient(0.0, 0.0, 0.0);
    //
    const Real cell = fld.biochem[0],
               ecgf = fld.biochem[1],
               oxyg = fld.biochem[2],
               mdgf = fld.biochem[3],
               vasc = fld.biochem[4];
    const RealGradient Grad_cell = fld.grad_biochem[0],
                       Grad_ecgf = fld.grad_biochem[1],
                       Grad_oxyg = fld.grad_biochem[2],
                       Grad_mdgf = fld.grad_biochem[3],
                       Grad_vasc = fld.grad_biochem[4];
    //
    if ( -1 == fld.a_flag ) return;
    //
    const bool is_activator = mat.get_biochem_param(0);
    //
    const Real kappa     = mat.get_biochem_param(1),  // 1 / day
               lambda    = mat.get_biochem_param(2),  // 1 / day
               h         = mat.get_biochem_param(3),
               Diff_cell = mat.get_biochem_param(4),  // m^2 / day
               Diff_ecgf = mat.get_biochem_param(5),  // m^2 / day
               alpha     = mat.get_biochem_param(6),
               ecgf_m    = mat.get_biochem_param(7),
               beta      = is_activator ? (1.0+pow2(ecgf_m)-2.0*h*ecgf_m)/pow2(1.0-ecgf_m) : 0.0;
    const Real //oxyg_m    = mat.get_biochem_param(10),
               Diff_oxyg = mat.get_biochem_param(11), // m^2 / day
               ksi_vasc  = mat.get_biochem_param(12), // 1 / day
               ksi       = mat.get_biochem_param(13), // 1 / day
               Diff_mdgf = mat.get_biochem_param(14), // m^2 / day
               //mu_oxyg   = mat.get_biochem_param(15), // 1 / day
               mu        = mat.get_biochem_param(16), // 1 / day
               Diff_vasc = mat.get_biochem_param(17), // m^2 / day
               delta     = mat.get_biochem_param(18), // 1 / day
               r         = mat.get_biochem_param(19);
    const Real s_oxyg    = mat.get_biochem_param(20);
    //
    const Real S_ = is_activator ? ((2.0*ecgf_m*(h-beta)*ecgf)/(pow2(ecgf_m)+pow2(ecgf))+beta) : (((h-1.0)*ecgf+h)/(2.0*(h-1.0)*ecgf+1.0)),
               G_ = is_activator ? (cell*(1.0+pow2(alpha))/(pow2(cell)+pow2(alpha))) : cell;
    const Real H_ = ((r-1.0)*mdgf+r)/(2.0*(r-1.0)*mdgf+1.0);
    const Real K_ = (oxyg < s_oxyg) ? (kappa*apply_bounds(0.0, oxyg, 1.0)/s_oxyg) : kappa;
    //
    Crd(0) = K_*S_*cell*(2.0-cell) - kappa*cell;
    Crd(1) = lambda*G_ - lambda*ecgf;
    Crd(2) = ksi_vasc*vasc - ksi*oxyg;
    Crd(3) = - mu*mdgf;
    Crd(4) = delta*H_*vasc*(1.0-vasc);
    //
    Grad_Crd(0) = Diff_cell*Grad_cell;
    Grad_Crd(1) = Diff_ecgf*Grad_ecgf;
    Grad_Crd(2) = Diff_oxyg*Grad_oxyg;
    Grad_Crd(3) = Diff_mdgf*Grad_mdgf;
    Grad_Crd(4) = Diff_vasc*Grad_vasc;
}
//=================================================================================================
void damaged_tissue_611 ( _LOCAL_ARGUMENTS_ ) {
    const unsigned int nvars = 6;
    Crd.resize(nvars);
    Crd = 0.0;
    Grad_Crd.resize(nvars);
    Grad_Crd = RealGradient(0.0, 0.0, 0.0);
    //
    const Real cell = fld.biochem[0],
               ecgf = fld.biochem[1],
               oxyg = fld.biochem[2],
               mdgf = fld.biochem[3],
               vasc = fld.biochem[4],
               ecm  = fld.biochem[5];
    const RealGradient Grad_cell = fld.grad_biochem[0],
                       Grad_ecgf = fld.grad_biochem[1],
                       Grad_oxyg = fld.grad_biochem[2],
                       Grad_mdgf = fld.grad_biochem[3],
                       Grad_vasc = fld.grad_biochem[4],
                       Grad_ecm  = fld.grad_biochem[5];
    //
    if ( -1 == fld.a_flag ) return;
    //
    const bool is_activator = mat.get_biochem_param(0);
    //
    const Real kappa     = mat.get_biochem_param(1),  // 1 / day
               lambda    = mat.get_biochem_param(2),  // 1 / day
               h         = mat.get_biochem_param(3),
               Diff_cell = mat.get_biochem_param(4),  // m^2 / day
               Diff_ecgf = mat.get_biochem_param(5),  // m^2 / day
               alpha     = mat.get_biochem_param(6),
               ecgf_m    = mat.get_biochem_param(7),
               beta      = is_activator ? (1.0+pow2(ecgf_m)-2.0*h*ecgf_m)/pow2(1.0-ecgf_m) : 0.0;
    const Real oxyg_m    = mat.get_biochem_param(10),
               Diff_oxyg = mat.get_biochem_param(11), // m^2 / day
               ksi_vasc  = mat.get_biochem_param(12), // 1 / day
               ksi       = mat.get_biochem_param(13), // 1 / day
               Diff_mdgf = mat.get_biochem_param(14), // m^2 / day
               mu_oxyg   = mat.get_biochem_param(15), // 1 / day
               mu        = mat.get_biochem_param(16), // 1 / day
               Diff_vasc = mat.get_biochem_param(17), // m^2 / day
               delta     = mat.get_biochem_param(18), // 1 / day
               r         = mat.get_biochem_param(19);
    const Real s_oxyg       = mat.get_biochem_param(20),
               epsilon      = mat.get_biochem_param(22), // 1 / day
               epsilon_oxyg = mat.get_biochem_param(23);
    //
    const Real S_ = is_activator ? ((2.0*ecgf_m*(h-beta)*ecgf)/(pow2(ecgf_m)+pow2(ecgf))+beta) : (((h-1.0)*ecgf+h)/(2.0*(h-1.0)*ecgf+1.0)),
               G_ = is_activator ? (cell*(1.0+pow2(alpha))/(pow2(cell)+pow2(alpha))) : cell;
    const Real Q_ = ( oxyg < oxyg_m ) ? (1.0-oxyg/oxyg_m) : 0.0,
               H_ = ((r-1.0)*mdgf+r)/(2.0*(r-1.0)*mdgf+1.0);
    const Real K_ = (oxyg < s_oxyg) ? (kappa*apply_bounds(0.0, oxyg, 1.0)/s_oxyg) : kappa;
    //
    Crd(0) = K_*S_*cell*(2.0-cell) - kappa*cell;
    Crd(1) = lambda*G_ - lambda*ecgf;
    Crd(2) = ksi_vasc*vasc - ksi*oxyg;
    Crd(3) = mu_oxyg*Q_ - mu*mdgf;
    Crd(4) = delta*H_*vasc*(1.0-vasc);
    Crd(5) = epsilon*vasc*cell*(epsilon_oxyg*oxyg-ecm);
    //
    Grad_Crd(0) = Diff_cell*Grad_cell;
    Grad_Crd(1) = Diff_ecgf*Grad_ecgf;
    Grad_Crd(2) = Diff_oxyg*Grad_oxyg;
    Grad_Crd(3) = Diff_mdgf*Grad_mdgf;
    Grad_Crd(4) = Diff_vasc*Grad_vasc;
    Grad_Crd(5) = RealGradient(0.0, 0.0, 0.0);
}
//=================================================================================================
void healthy_tissue_611 ( _LOCAL_ARGUMENTS_ ) {
    const unsigned int nvars = 6;
    Crd.resize(nvars);
    Crd = 0.0;
    Grad_Crd.resize(nvars);
    Grad_Crd = RealGradient(0.0, 0.0, 0.0);
    //
    const Real cell = fld.biochem[0],
               ecgf = fld.biochem[1],
               oxyg = fld.biochem[2],
               mdgf = fld.biochem[3],
               vasc = fld.biochem[4],
               ecm  = fld.biochem[5];
    const RealGradient Grad_cell = fld.grad_biochem[0],
                       Grad_ecgf = fld.grad_biochem[1],
                       Grad_oxyg = fld.grad_biochem[2],
                       Grad_mdgf = fld.grad_biochem[3],
                       Grad_vasc = fld.grad_biochem[4],
                       Grad_ecm  = fld.grad_biochem[5];
    //
    if ( -1 == fld.a_flag ) return;
    //
    const bool is_activator = mat.get_biochem_param(0);
    //
    const Real kappa     = mat.get_biochem_param(1),  // 1 / day
               lambda    = mat.get_biochem_param(2),  // 1 / day
               h         = mat.get_biochem_param(3),
               Diff_cell = mat.get_biochem_param(4),  // m^2 / day
               Diff_ecgf = mat.get_biochem_param(5),  // m^2 / day
               alpha     = mat.get_biochem_param(6),
               ecgf_m    = mat.get_biochem_param(7),
               beta      = is_activator ? (1.0+pow2(ecgf_m)-2.0*h*ecgf_m)/pow2(1.0-ecgf_m) : 0.0;
    const Real //oxyg_m    = mat.get_biochem_param(10),
               Diff_oxyg = mat.get_biochem_param(11), // m^2 / day
               ksi_vasc  = mat.get_biochem_param(12), // 1 / day
               ksi       = mat.get_biochem_param(13), // 1 / day
               Diff_mdgf = mat.get_biochem_param(14), // m^2 / day
               //mu_oxyg   = mat.get_biochem_param(15), // 1 / day
               mu        = mat.get_biochem_param(16), // 1 / day
               Diff_vasc = mat.get_biochem_param(17), // m^2 / day
               delta     = mat.get_biochem_param(18), // 1 / day
               r         = mat.get_biochem_param(19);
    const Real s_oxyg       = mat.get_biochem_param(20),
               epsilon      = mat.get_biochem_param(22), // 1 / day
               epsilon_oxyg = mat.get_biochem_param(23);
    //
    const Real S_ = is_activator ? ((2.0*ecgf_m*(h-beta)*ecgf)/(pow2(ecgf_m)+pow2(ecgf))+beta) : (((h-1.0)*ecgf+h)/(2.0*(h-1.0)*ecgf+1.0)),
               G_ = is_activator ? (cell*(1.0+pow2(alpha))/(pow2(cell)+pow2(alpha))) : cell;
    const Real H_ = ((r-1.0)*mdgf+r)/(2.0*(r-1.0)*mdgf+1.0);
    const Real K_ = (oxyg < s_oxyg) ? (kappa*apply_bounds(0.0, oxyg, 1.0)/s_oxyg) : kappa;
    //
    Crd(0) = K_*S_*cell*(2.0-cell) - kappa*cell;
    Crd(1) = lambda*G_ - lambda*ecgf;
    Crd(2) = ksi_vasc*vasc - ksi*oxyg;
    Crd(3) = - mu*mdgf;
    Crd(4) = delta*H_*vasc*(1.0-vasc);
    Crd(5) = epsilon*vasc*cell*(epsilon_oxyg*oxyg-ecm);
    //
    Grad_Crd(0) = Diff_cell*Grad_cell;
    Grad_Crd(1) = Diff_ecgf*Grad_ecgf;
    Grad_Crd(2) = Diff_oxyg*Grad_oxyg;
    Grad_Crd(3) = Diff_mdgf*Grad_mdgf;
    Grad_Crd(4) = Diff_vasc*Grad_vasc;
    Grad_Crd(5) = RealGradient(0.0, 0.0, 0.0);
}
//=================================================================================================
} // end of namespace wound_healing...
//=================================================================================================
//
//
//
//=================================================================================================
#undef _LOCAL_ARGUMENTS_
#undef _VARIABILITY_1pct_
#undef _VARIABILITY_2pct_
#undef _VARIABILITY_5pct_
//=================================================================================================
