/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/nurbs.h"
//=================================================================================================
#define dz_dX    this->_inv_map_dzdX
#define d2z_dX2  this->_inv_map_d2zdX2
#define dN_dz    this->_dphidz
#define d2N_dz2  this->_d2phidz2
//=================================================================================================
void NURBS::calc_d2phi (const Point& crd) const {
    if ( ! this->_calculate_d2phi ) return;
    // check for NURBS space
    if ( SP1D == this->dimension() ) {
        //
        for (int a=0; a<this->size(); a++) {
            this->_d2phi[a].comp(0,0) = d2z_dX2(0,0,0)*dN_dz[0][a]
                                      + dz_dX(0,0)*dz_dX(0,0)*d2N_dz2[0][0][a];
            this->_d2phi[a].comp(1,0) = d2z_dX2(0,1,0)*dN_dz[0][a]
                                      + dz_dX(0,1)*dz_dX(0,0)*d2N_dz2[0][0][a];
            this->_d2phi[a].comp(2,0) = d2z_dX2(0,2,0)*dN_dz[0][a]
                                      + dz_dX(0,2)*dz_dX(0,0)*d2N_dz2[0][0][a];
            this->_d2phi[a].comp(0,1) = d2z_dX2(0,0,1)*dN_dz[0][a]
                                      + dz_dX(0,0)*dz_dX(0,1)*d2N_dz2[0][0][a];
            this->_d2phi[a].comp(1,1) = d2z_dX2(0,1,1)*dN_dz[0][a]
                                      + dz_dX(0,1)*dz_dX(0,1)*d2N_dz2[0][0][a];
            this->_d2phi[a].comp(2,1) = d2z_dX2(0,2,1)*dN_dz[0][a]
                                      + dz_dX(0,2)*dz_dX(0,1)*d2N_dz2[0][0][a];
            this->_d2phi[a].comp(0,2) = d2z_dX2(0,0,2)*dN_dz[0][a]
                                      + dz_dX(0,0)*dz_dX(0,2)*d2N_dz2[0][0][a];
            this->_d2phi[a].comp(1,2) = d2z_dX2(0,1,2)*dN_dz[0][a]
                                      + dz_dX(0,1)*dz_dX(0,2)*d2N_dz2[0][0][a];
            this->_d2phi[a].comp(2,2) = d2z_dX2(0,2,2)*dN_dz[0][a]
                                      + dz_dX(0,2)*dz_dX(0,2)*d2N_dz2[0][0][a];
        }
        //
    } else if ( SP2D == this->dimension() ) {
        //
        for (int a=0; a<this->size(); a++) {
            this->_d2phi[a].comp(0,0) = d2z_dX2(0,0,0)*dN_dz[0][a]
                                      + d2z_dX2(1,0,0)*dN_dz[1][a]
                                      + dz_dX(0,0)*dz_dX(0,0)*d2N_dz2[0][0][a]
                                      + dz_dX(0,0)*dz_dX(1,0)*d2N_dz2[0][1][a]
                                      + dz_dX(1,0)*dz_dX(0,0)*d2N_dz2[1][0][a]
                                      + dz_dX(1,0)*dz_dX(1,0)*d2N_dz2[1][1][a];
            this->_d2phi[a].comp(1,0) = d2z_dX2(0,1,0)*dN_dz[0][a]
                                      + d2z_dX2(1,1,0)*dN_dz[1][a]
                                      + dz_dX(0,1)*dz_dX(0,0)*d2N_dz2[0][0][a]
                                      + dz_dX(0,1)*dz_dX(1,0)*d2N_dz2[0][1][a]
                                      + dz_dX(1,1)*dz_dX(0,0)*d2N_dz2[1][0][a]
                                      + dz_dX(1,1)*dz_dX(1,0)*d2N_dz2[1][1][a];
            this->_d2phi[a].comp(2,0) = d2z_dX2(0,2,0)*dN_dz[0][a]
                                      + d2z_dX2(1,2,0)*dN_dz[1][a]
                                      + dz_dX(0,2)*dz_dX(0,0)*d2N_dz2[0][0][a]
                                      + dz_dX(0,2)*dz_dX(1,0)*d2N_dz2[0][1][a]
                                      + dz_dX(1,2)*dz_dX(0,0)*d2N_dz2[1][0][a]
                                      + dz_dX(1,2)*dz_dX(1,0)*d2N_dz2[1][1][a];
            this->_d2phi[a].comp(0,1) = d2z_dX2(0,0,1)*dN_dz[0][a]
                                      + d2z_dX2(1,0,1)*dN_dz[1][a]
                                      + dz_dX(0,0)*dz_dX(0,1)*d2N_dz2[0][0][a]
                                      + dz_dX(0,0)*dz_dX(1,1)*d2N_dz2[0][1][a]
                                      + dz_dX(1,0)*dz_dX(0,1)*d2N_dz2[1][0][a]
                                      + dz_dX(1,0)*dz_dX(1,1)*d2N_dz2[1][1][a];
            this->_d2phi[a].comp(1,1) = d2z_dX2(0,1,1)*dN_dz[0][a]
                                      + d2z_dX2(1,1,1)*dN_dz[1][a]
                                      + dz_dX(0,1)*dz_dX(0,1)*d2N_dz2[0][0][a]
                                      + dz_dX(0,1)*dz_dX(1,1)*d2N_dz2[0][1][a]
                                      + dz_dX(1,1)*dz_dX(0,1)*d2N_dz2[1][0][a]
                                      + dz_dX(1,1)*dz_dX(1,1)*d2N_dz2[1][1][a];
            this->_d2phi[a].comp(2,1) = d2z_dX2(0,2,1)*dN_dz[0][a]
                                      + d2z_dX2(1,2,1)*dN_dz[1][a]
                                      + dz_dX(0,2)*dz_dX(0,1)*d2N_dz2[0][0][a]
                                      + dz_dX(0,2)*dz_dX(1,1)*d2N_dz2[0][1][a]
                                      + dz_dX(1,2)*dz_dX(0,1)*d2N_dz2[1][0][a]
                                      + dz_dX(1,2)*dz_dX(1,1)*d2N_dz2[1][1][a];
            this->_d2phi[a].comp(0,2) = d2z_dX2(0,0,2)*dN_dz[0][a]
                                      + d2z_dX2(1,0,2)*dN_dz[1][a]
                                      + dz_dX(0,0)*dz_dX(0,2)*d2N_dz2[0][0][a]
                                      + dz_dX(0,0)*dz_dX(1,2)*d2N_dz2[0][1][a]
                                      + dz_dX(1,0)*dz_dX(0,2)*d2N_dz2[1][0][a]
                                      + dz_dX(1,0)*dz_dX(1,2)*d2N_dz2[1][1][a];
            this->_d2phi[a].comp(1,2) = d2z_dX2(0,1,2)*dN_dz[0][a]
                                      + d2z_dX2(1,1,2)*dN_dz[1][a]
                                      + dz_dX(0,1)*dz_dX(0,2)*d2N_dz2[0][0][a]
                                      + dz_dX(0,1)*dz_dX(1,2)*d2N_dz2[0][1][a]
                                      + dz_dX(1,1)*dz_dX(0,2)*d2N_dz2[1][0][a]
                                      + dz_dX(1,1)*dz_dX(1,2)*d2N_dz2[1][1][a];
            this->_d2phi[a].comp(2,2) = d2z_dX2(0,2,2)*dN_dz[0][a]
                                      + d2z_dX2(1,2,2)*dN_dz[1][a]
                                      + dz_dX(0,2)*dz_dX(0,2)*d2N_dz2[0][0][a]
                                      + dz_dX(0,2)*dz_dX(1,2)*d2N_dz2[0][1][a]
                                      + dz_dX(1,2)*dz_dX(0,2)*d2N_dz2[1][0][a]
                                      + dz_dX(1,2)*dz_dX(1,2)*d2N_dz2[1][1][a];
        }
        //
    } else if ( SP3D == this->dimension() ) {
        //
        for (int a=0; a<this->size(); a++) {
            this->_d2phi[a].comp(0,0) = d2z_dX2(0,0,0)*dN_dz[0][a]
                                      + d2z_dX2(1,0,0)*dN_dz[1][a]
                                      + d2z_dX2(2,0,0)*dN_dz[2][a]
                                      + dz_dX(0,0)*dz_dX(0,0)*d2N_dz2[0][0][a]
                                      + dz_dX(0,0)*dz_dX(1,0)*d2N_dz2[0][1][a]
                                      + dz_dX(0,0)*dz_dX(2,0)*d2N_dz2[0][2][a]
                                      + dz_dX(1,0)*dz_dX(0,0)*d2N_dz2[1][0][a]
                                      + dz_dX(1,0)*dz_dX(1,0)*d2N_dz2[1][1][a]
                                      + dz_dX(1,0)*dz_dX(2,0)*d2N_dz2[1][2][a]
                                      + dz_dX(2,0)*dz_dX(0,0)*d2N_dz2[2][0][a]
                                      + dz_dX(2,0)*dz_dX(1,0)*d2N_dz2[2][1][a]
                                      + dz_dX(2,0)*dz_dX(2,0)*d2N_dz2[2][2][a];
            this->_d2phi[a].comp(1,0) = d2z_dX2(0,1,0)*dN_dz[0][a]
                                      + d2z_dX2(1,1,0)*dN_dz[1][a]
                                      + d2z_dX2(2,1,0)*dN_dz[2][a]
                                      + dz_dX(0,1)*dz_dX(0,0)*d2N_dz2[0][0][a]
                                      + dz_dX(0,1)*dz_dX(1,0)*d2N_dz2[0][1][a]
                                      + dz_dX(0,1)*dz_dX(2,0)*d2N_dz2[0][2][a]
                                      + dz_dX(1,1)*dz_dX(0,0)*d2N_dz2[1][0][a]
                                      + dz_dX(1,1)*dz_dX(1,0)*d2N_dz2[1][1][a]
                                      + dz_dX(1,1)*dz_dX(2,0)*d2N_dz2[1][2][a]
                                      + dz_dX(2,1)*dz_dX(0,0)*d2N_dz2[2][0][a]
                                      + dz_dX(2,1)*dz_dX(1,0)*d2N_dz2[2][1][a]
                                      + dz_dX(2,1)*dz_dX(2,0)*d2N_dz2[2][2][a];
            this->_d2phi[a].comp(2,0) = d2z_dX2(0,2,0)*dN_dz[0][a]
                                      + d2z_dX2(1,2,0)*dN_dz[1][a]
                                      + d2z_dX2(2,2,0)*dN_dz[2][a]
                                      + dz_dX(0,2)*dz_dX(0,0)*d2N_dz2[0][0][a]
                                      + dz_dX(0,2)*dz_dX(1,0)*d2N_dz2[0][1][a]
                                      + dz_dX(0,2)*dz_dX(2,0)*d2N_dz2[0][2][a]
                                      + dz_dX(1,2)*dz_dX(0,0)*d2N_dz2[1][0][a]
                                      + dz_dX(1,2)*dz_dX(1,0)*d2N_dz2[1][1][a]
                                      + dz_dX(1,2)*dz_dX(2,0)*d2N_dz2[1][2][a]
                                      + dz_dX(2,2)*dz_dX(0,0)*d2N_dz2[2][0][a]
                                      + dz_dX(2,2)*dz_dX(1,0)*d2N_dz2[2][1][a]
                                      + dz_dX(2,2)*dz_dX(2,0)*d2N_dz2[2][2][a];
            this->_d2phi[a].comp(0,1) = d2z_dX2(0,0,1)*dN_dz[0][a]
                                      + d2z_dX2(1,0,1)*dN_dz[1][a]
                                      + d2z_dX2(2,0,1)*dN_dz[2][a]
                                      + dz_dX(0,0)*dz_dX(0,1)*d2N_dz2[0][0][a]
                                      + dz_dX(0,0)*dz_dX(1,1)*d2N_dz2[0][1][a]
                                      + dz_dX(0,0)*dz_dX(2,1)*d2N_dz2[0][2][a]
                                      + dz_dX(1,0)*dz_dX(0,1)*d2N_dz2[1][0][a]
                                      + dz_dX(1,0)*dz_dX(1,1)*d2N_dz2[1][1][a]
                                      + dz_dX(1,0)*dz_dX(2,1)*d2N_dz2[1][2][a]
                                      + dz_dX(2,0)*dz_dX(0,1)*d2N_dz2[2][0][a]
                                      + dz_dX(2,0)*dz_dX(1,1)*d2N_dz2[2][1][a]
                                      + dz_dX(2,0)*dz_dX(2,1)*d2N_dz2[2][2][a];
            this->_d2phi[a].comp(1,1) = d2z_dX2(0,1,1)*dN_dz[0][a]
                                      + d2z_dX2(1,1,1)*dN_dz[1][a]
                                      + d2z_dX2(2,1,1)*dN_dz[2][a]
                                      + dz_dX(0,1)*dz_dX(0,1)*d2N_dz2[0][0][a]
                                      + dz_dX(0,1)*dz_dX(1,1)*d2N_dz2[0][1][a]
                                      + dz_dX(0,1)*dz_dX(2,1)*d2N_dz2[0][2][a]
                                      + dz_dX(1,1)*dz_dX(0,1)*d2N_dz2[1][0][a]
                                      + dz_dX(1,1)*dz_dX(1,1)*d2N_dz2[1][1][a]
                                      + dz_dX(1,1)*dz_dX(2,1)*d2N_dz2[1][2][a]
                                      + dz_dX(2,1)*dz_dX(0,1)*d2N_dz2[2][0][a]
                                      + dz_dX(2,1)*dz_dX(1,1)*d2N_dz2[2][1][a]
                                      + dz_dX(2,1)*dz_dX(2,1)*d2N_dz2[2][2][a];
            this->_d2phi[a].comp(2,1) = d2z_dX2(0,2,1)*dN_dz[0][a]
                                      + d2z_dX2(1,2,1)*dN_dz[1][a]
                                      + d2z_dX2(2,2,1)*dN_dz[2][a]
                                      + dz_dX(0,2)*dz_dX(0,1)*d2N_dz2[0][0][a]
                                      + dz_dX(0,2)*dz_dX(1,1)*d2N_dz2[0][1][a]
                                      + dz_dX(0,2)*dz_dX(2,1)*d2N_dz2[0][2][a]
                                      + dz_dX(1,2)*dz_dX(0,1)*d2N_dz2[1][0][a]
                                      + dz_dX(1,2)*dz_dX(1,1)*d2N_dz2[1][1][a]
                                      + dz_dX(1,2)*dz_dX(2,1)*d2N_dz2[1][2][a]
                                      + dz_dX(2,2)*dz_dX(0,1)*d2N_dz2[2][0][a]
                                      + dz_dX(2,2)*dz_dX(1,1)*d2N_dz2[2][1][a]
                                      + dz_dX(2,2)*dz_dX(2,1)*d2N_dz2[2][2][a];
            this->_d2phi[a].comp(0,2) = d2z_dX2(0,0,2)*dN_dz[0][a]
                                      + d2z_dX2(1,0,2)*dN_dz[1][a]
                                      + d2z_dX2(2,0,2)*dN_dz[2][a]
                                      + dz_dX(0,0)*dz_dX(0,2)*d2N_dz2[0][0][a]
                                      + dz_dX(0,0)*dz_dX(1,2)*d2N_dz2[0][1][a]
                                      + dz_dX(0,0)*dz_dX(2,2)*d2N_dz2[0][2][a]
                                      + dz_dX(1,0)*dz_dX(0,2)*d2N_dz2[1][0][a]
                                      + dz_dX(1,0)*dz_dX(1,2)*d2N_dz2[1][1][a]
                                      + dz_dX(1,0)*dz_dX(2,2)*d2N_dz2[1][2][a]
                                      + dz_dX(2,0)*dz_dX(0,2)*d2N_dz2[2][0][a]
                                      + dz_dX(2,0)*dz_dX(1,2)*d2N_dz2[2][1][a]
                                      + dz_dX(2,0)*dz_dX(2,2)*d2N_dz2[2][2][a];
            this->_d2phi[a].comp(1,2) = d2z_dX2(0,1,2)*dN_dz[0][a]
                                      + d2z_dX2(1,1,2)*dN_dz[1][a]
                                      + d2z_dX2(2,1,2)*dN_dz[2][a]
                                      + dz_dX(0,1)*dz_dX(0,2)*d2N_dz2[0][0][a]
                                      + dz_dX(0,1)*dz_dX(1,2)*d2N_dz2[0][1][a]
                                      + dz_dX(0,1)*dz_dX(2,2)*d2N_dz2[0][2][a]
                                      + dz_dX(1,1)*dz_dX(0,2)*d2N_dz2[1][0][a]
                                      + dz_dX(1,1)*dz_dX(1,2)*d2N_dz2[1][1][a]
                                      + dz_dX(1,1)*dz_dX(2,2)*d2N_dz2[1][2][a]
                                      + dz_dX(2,1)*dz_dX(0,2)*d2N_dz2[2][0][a]
                                      + dz_dX(2,1)*dz_dX(1,2)*d2N_dz2[2][1][a]
                                      + dz_dX(2,1)*dz_dX(2,2)*d2N_dz2[2][2][a];
            this->_d2phi[a].comp(2,2) = d2z_dX2(0,2,2)*dN_dz[0][a]
                                      + d2z_dX2(1,2,2)*dN_dz[1][a]
                                      + d2z_dX2(2,2,2)*dN_dz[2][a]
                                      + dz_dX(0,2)*dz_dX(0,2)*d2N_dz2[0][0][a]
                                      + dz_dX(0,2)*dz_dX(1,2)*d2N_dz2[0][1][a]
                                      + dz_dX(0,2)*dz_dX(2,2)*d2N_dz2[0][2][a]
                                      + dz_dX(1,2)*dz_dX(0,2)*d2N_dz2[1][0][a]
                                      + dz_dX(1,2)*dz_dX(1,2)*d2N_dz2[1][1][a]
                                      + dz_dX(1,2)*dz_dX(2,2)*d2N_dz2[1][2][a]
                                      + dz_dX(2,2)*dz_dX(0,2)*d2N_dz2[2][0][a]
                                      + dz_dX(2,2)*dz_dX(1,2)*d2N_dz2[2][1][a]
                                      + dz_dX(2,2)*dz_dX(2,2)*d2N_dz2[2][2][a];
        }
        //
    } else libmesh_error();
}
//=================================================================================================
#undef dz_dX
#undef d2z_dX2
#undef dN_dz
#undef d2N_dz2
//=================================================================================================
void NURBS::calc_d2phidxidxi (const Point& crd) const {
    const int L = 0, M = 0;
    if ( ! this->_calculate_d2phidz2[L][M] ) return;
    //
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a);
            this->_d2phidz2[L][M][a] = this->_basef_2nd_deriv<0>(crd(0), i);
        }
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a);
            this->_d2phidz2[L][M][a] = this->_basef_2nd_deriv<0>(crd(0), i)
                                     * this->_basef_0th_deriv<1>(crd(1), j);
        }
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a),
                      k = this->index<2>(a);
            this->_d2phidz2[L][M][a] = this->_basef_2nd_deriv<0>(crd(0), i)
                                     * this->_basef_0th_deriv<1>(crd(1), j)
                                     * this->_basef_0th_deriv<2>(crd(2), k);
        }
    }
    //
    if ( this->_is_rational ) this->_rationalize_d2phi(L, M);
}
//=================================================================================================
void NURBS::calc_d2phidxideta (const Point& crd) const {
    const int L = 0, M = 1;
    if ( ! this->_calculate_d2phidz2[L][M] ) return;
    //
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a);
            this->_d2phidz2[L][M][a] = 0.0;
        }
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a);
            this->_d2phidz2[L][M][a] = this->_basef_1st_deriv<0>(crd(0), i)
                                     * this->_basef_1st_deriv<1>(crd(1), j);
        }
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a),
                      k = this->index<2>(a);
            this->_d2phidz2[L][M][a] = this->_basef_1st_deriv<0>(crd(0), i)
                                     * this->_basef_1st_deriv<1>(crd(1), j)
                                     * this->_basef_0th_deriv<2>(crd(2), k);
        }
    }
    //
    if ( this->_is_rational ) this->_rationalize_d2phi(L, M);
}
//=================================================================================================
void NURBS::calc_d2phidxidzeta (const Point& crd) const {
    const int L = 0, M = 2;
    if ( ! this->_calculate_d2phidz2[L][M] ) return;
    //
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a);
            this->_d2phidz2[L][M][a] = 0.0;
        }
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a),
            //          j = this->index<1>(a);
            this->_d2phidz2[L][M][a] = 0.0;
        }
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a),
                      k = this->index<2>(a);
            this->_d2phidz2[L][M][a] = this->_basef_1st_deriv<0>(crd(0), i)
                                     * this->_basef_0th_deriv<1>(crd(1), j)
                                     * this->_basef_1st_deriv<2>(crd(2), k);
        }
    }
    //
    if ( this->_is_rational ) this->_rationalize_d2phi(L, M);
}
//=================================================================================================
void NURBS::calc_d2phidetadxi (const Point& crd) const {
    const int L = 1, M = 0;
    if ( ! this->_calculate_d2phidz2[L][M] ) return;
    //
    this->_d2phidz2[L][M] = this->_d2phidz2[M][L];
    if ( this->_is_rational ) this->_sum__w_d2Ndz2[L][M] = this->_sum__w_d2Ndz2[M][L];
}
//=================================================================================================
void NURBS::calc_d2phidetadeta (const Point& crd) const {
    const int L = 1, M = 1;
    if ( ! this->_calculate_d2phidz2[L][M] ) return;
    //
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a);
            this->_d2phidz2[L][M][a] = 0.0;
        }
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a);
            this->_d2phidz2[L][M][a] = this->_basef_0th_deriv<0>(crd(0), i)
                                     * this->_basef_2nd_deriv<1>(crd(1), j);
        }
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a),
                      k = this->index<2>(a);
            this->_d2phidz2[L][M][a] = this->_basef_0th_deriv<0>(crd(0), i)
                                     * this->_basef_2nd_deriv<1>(crd(1), j)
                                     * this->_basef_0th_deriv<2>(crd(2), k);
        }
    }
    //
    if ( this->_is_rational ) this->_rationalize_d2phi(L, M);
}
//=================================================================================================
void NURBS::calc_d2phidetadzeta (const Point& crd) const {
    const int L = 1, M = 2;
    if ( ! this->_calculate_d2phidz2[L][M] ) return;
    //
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a);
            this->_d2phidz2[L][M][a] = 0.0;
        }
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a),
            //          j = this->index<1>(a);
            this->_d2phidz2[L][M][a] = 0.0;
        }
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a),
                      k = this->index<2>(a);
            this->_d2phidz2[L][M][a] = this->_basef_0th_deriv<0>(crd(0), i)
                                     * this->_basef_1st_deriv<1>(crd(1), j)
                                     * this->_basef_1st_deriv<2>(crd(2), k);
        }
    }
    //
    if ( this->_is_rational ) this->_rationalize_d2phi(L, M);
}
//=================================================================================================
void NURBS::calc_d2phidzetadxi (const Point& crd) const {
    const int L = 2, M = 0;
    if ( ! this->_calculate_d2phidz2[L][M] ) return;
    //
    this->_d2phidz2[L][M] = this->_d2phidz2[M][L];
    if ( this->_is_rational ) this->_sum__w_d2Ndz2[L][M] = this->_sum__w_d2Ndz2[M][L];
}
//=================================================================================================
void NURBS::calc_d2phidzetadeta (const Point& crd) const {
    const int L = 2, M = 1;
    if ( ! this->_calculate_d2phidz2[L][M] ) return;
    //
    this->_d2phidz2[L][M] = this->_d2phidz2[M][L];
    if ( this->_is_rational ) this->_sum__w_d2Ndz2[L][M] = this->_sum__w_d2Ndz2[M][L];
}
//=================================================================================================
void NURBS::calc_d2phidzetadzeta (const Point& crd) const {
    const int L = 2, M = 2;
    if ( ! this->_calculate_d2phidz2[L][M] ) return;
    //
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a);
            this->_d2phidz2[L][M][a] = 0.0;
        }
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a),
            //          j = this->index<1>(a);
            this->_d2phidz2[L][M][a] = 0.0;
        }
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a),
                      k = this->index<2>(a);
            this->_d2phidz2[L][M][a] = this->_basef_0th_deriv<0>(crd(0), i)
                                     * this->_basef_0th_deriv<1>(crd(1), j)
                                     * this->_basef_2nd_deriv<2>(crd(2), k);
        }
    }
    //
    if ( this->_is_rational ) this->_rationalize_d2phi(L, M);
}
//=================================================================================================
