/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/rt_tumour_fibrosis_solver.h"
#include "libmesh/linear_implicit_system.h"
#include "libmesh/gmsh_io.h"
#include "feb3/quadrature.h"
#include "feb3/ida.h"
#include "feb3/fem-context.h"
#include "feb3/paraview_io.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
GlobalTolerance tol;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
extern // list of all systems used in the analysis
std::vector< std::pair<EquationSystems*, std::string> > systems_ls;
//=================================================================================================
#include "feb3/domain_data_inl.h"
#include "feb3/solution_io_inl.h"
//=================================================================================================
RtTumourFibrosisSolver::RtTumourFibrosisSolver (DomainData& dmn) :
    ReactionDiffusionSolverXD(dmn),
    _log("RtTumourFibrosisSolver") {
    // set the default label
    this->_label = "RtTumourFibrosisSolver";
}
//=================================================================================================
void RtTumourFibrosisSolver::init () {
    this->_log.push("init");

    // initialize the solver
    ReactionDiffusionSolverXD::init();

    this->_log.pop("init");
}
//=================================================================================================
void RtTumourFibrosisSolver::run () {
    this->_log.push("run");

    // run the solver
    ReactionDiffusionSolverXD::run();

    this->_log.pop("run");
}
//=================================================================================================
