/*************************************************************************************************

    Copyright (C) 2009-2019    Jean de Montigny (jean.demontigny@hotmail.fr)
                               Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#ifndef __CANCER_GROWTH_TEST_H__
#define __CANCER_GROWTH_TEST_H__
//=================================================================================================
void cancer_growth () {

  // structure containing BDM_Domain
  std::vector<BDM_Domain> rve_list;
  // read the model parameters and simulation setup
  const std::string fname = "examples/biodynamo_cancer_growth/model_parameters.txt";
  // list of the input parameters read
  libMesh::Parameters params;
  // make sure we are not invoking any of FEB3's solvers here
  params.set<ProblemCase>("solver") = NO_SOLVER;
  // pre-set the following parameters...
  params.set<unsigned int>("reaction-diffusion solver: dimension") = 3;
  params.set<bool>("BioDynaMo/use_default_visualization_tools") = true;
  //
  //
  //
  // input stream to read from
  GetPot in(fname);
  if (access(fname.c_str(), F_OK) == -1) {
    std::cout << " ### cannot find \"" << fname << "\"; abort program ###" << std::endl;
    libmesh_error();
  }
  std::set<std::string> in_p;
  // now read the parameters from the input file
  read_BDM_parameters(in, in_p, params);
  //
  // print on screen all identified variables in the input file
  std::cout << std::endl
            << " ### identified variables in file \"" << in[0] << "\" = "
            << in_p.size() << " ###" << std::endl;
  for (unsigned int i=1; i<=in_p.size(); i++) {
      std::cout << " ### " << i << " :: " << in[i] << " ###" << std::endl;
  }
  // print on screen all unidentified (if any) variables in the input file
  std::vector<std::string> non_in_p = in.unidentified_variables(in_p);
  std::cout << std::endl
            << " ### unidentified variables in file \"" << in[0] << "\" = "
            << non_in_p.size() << " ###" << std::endl;
  for (unsigned int i=0; i<non_in_p.size(); i++) {
      std::cout << " ### " << i << " :: " << non_in_p[i] << " ###" << std::endl;
  }
  std::cout << std::endl;
  //
  //
  //
  std::vector<libMesh::Point> rve_initial_distribution;
  // initialize all BioDynaMo domains (RVEs) for simulation
  bdm::init_bdm_domains(rve_list, params, rve_initial_distribution);
  //
  // iterate for all time steps and for all RVEs
  for (int iter=0; iter<params.get<int>("BioDynaMo/number_of_iterations"); iter++) {
    std::cout << "\n\t*** Iteration " << (iter+1) << " ***" << std::endl;
    // run BioDynaMo simulations for all RVEs
    bdm::simulate_bdm_domains(rve_list, params);
    // new cube creation due to cell migration (if applicable)
    bdm::init_migrating_bdm_domains(rve_list, params);
  }
  //
  // finally do memory clean-up
  bdm::clean_bdm_domains(rve_list);

}
//=================================================================================================
#endif // __CANCER_GROWTH_TEST_H__
//=================================================================================================
