/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include <iomanip>
#include "feb3/utils.h"
#include "feb3/domain_data.h"
#include "libmesh/getpot.h"
#include "libmesh/libmesh_logging.h"
#include "libmesh/mesh.h"
#include "libmesh/mesh_generation.h"
#include "libmesh/gmsh_io.h"
#include "libmesh/exodusII_io.h"
#include "feb3/mesh_iga.h"
#include "feb3/quadrature.h"
#include "feb3/paraview_io.h"
#include "feb3/drug_delivery_solver.h"
#include "feb3/reaction_diffusion_solver_Xd.h"
#include "feb3/navier_cauchy_iga_solver_3d.h"
#include "feb3/navier_cauchy_solver_3d.h"
#include "feb3/explicit_navier_cauchy_solver_3d.h"
#include "feb3/wound_healing_solver.h"
#include "feb3/tumour_angiogenesis_solver.h"
#include "feb3/tumour_angiogenesis_constitutive.h"
#include "feb3/invitro_tumoroid_solver.h"
#include "feb3/rt_tumour_fibrosis_solver.h"
#include "feb3/rve.h"
#include "feb3/ida.h"
#include "libmesh/linear_implicit_system.h"
#include "libmesh/fe.h"
#include "libmesh/elem.h"
#include "libmesh/dof_map.h"
#include "libmesh/sparse_matrix.h"
#include "libmesh/numeric_vector.h"
#include "libmesh/dense_matrix.h"
#include "libmesh/dense_vector.h"
#include "libmesh/type_tensor.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
GlobalTolerance tol;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
//=================================================================================================
void my_test_8 () {
    for (int i=0; i<20; i++) {
        const int num = uniform_distro(0, 100);
        std::vector<int> num_digit;
        const unsigned int res = digitise_me(num, num_digit);
        std::cout << num << ' ' << res << ' ' << std::flush;
        for (unsigned int j=0; j<num_digit.size(); j++)
            std::cout << ' ' << num_digit[j] << std::flush;
        std::cout << std::endl;
    }
    //
    //
    //
    const Real sc = 1.0e-2;
    Dyadic3D strain;
    strain.comp = +8.2, +0.0, -3.3,
                  +0.0, -5.4, +0.9,
                  -3.3, +0.9, +6.1;
    strain.comp *= sc;
    std::vector<Vector3D> n;
    std::vector<Real> epsilon;
    eigen_solve(strain, epsilon, n);
    //
    std::cout << std::endl;
    std::cout << strain.comp << std::endl;
    std::cout << epsilon[0] << ' ' << epsilon[1] << ' ' << epsilon[2] << ' ' << std::endl;
    std::cout << sqrt(2.0*epsilon[0]+1.0) << ' ' << sqrt(2.0*epsilon[1]+1.0) << ' ' << sqrt(2.0*epsilon[2]+1.0) << ' ' << std::endl;
    std::cout << magnitude(n[0]) << ' ' << magnitude(n[1]) << ' ' << magnitude(n[2]) << ' ' << std::endl;
    std::cout << n[0].comp << ' ' << n[1].comp << ' ' << n[2].comp << ' ' << std::endl;
    std::cout << std::endl;
}
//=================================================================================================
void my_test_7 () {
    //
    VascularInfo vi;
    vi.flow = VascularInfo::BLOOD;
    vi.status = VascularInfo::NORMAL;
    // vi.status = VascularInfo::COLLAPSED;
    vi.thickness = 3.000e-6;
    vi.pores_fraction = 50.01e-4;
    vi.pore_radius = 50.00e-9;
    vi.pore_radius = 150.00e-9;
    //
    MaterialInfo mi__vnr(0, 500);
    mi__vnr.set_biochem_param(0) = 399.967e-5; // Pa*s (default constant blood viscosity: erythrocytes, platelets and plasma proteins)
    mi__vnr.set_biochem_param(1) = 385.000e-5; // Pa*s (default constant lymph viscosity)
    mi__vnr.set_biochem_param(2) = 150.000e-5; // Pa*s (default constant plasma viscosity)
    mi__vnr.set_biochem_param(3) = 350.000e-5; // Pa*s (default constant interstitial fluid viscosity)
    mi__vnr.set_biochem_param(210) = 0.500e-9; // m (mean radius of molecule/particle: drug type-1: phase-1)
    mi__vnr.set_biochem_param(211) = 0.500e-9; // m (mean radius of molecule/particle: drug type-1: phase-2)
    mi__vnr.set_biochem_param(212) = 9.999e-9; // m (mean radius of molecule/particle: drug type-2: phase-1)
    mi__vnr.set_biochem_param(213) = 9.999e-9; // m (mean radius of molecule/particle: drug type-2: phase-2)
    // mi__vnr.set_biochem_param(205) = 49.99e-9; // m (mean radius of the drug type-2 molecule/particle)
    //
    MaterialInfo mi__dmn(0, 500);
    //
    FieldInfo fld(4, 8, 6);
    //
    Dyadic3D D0 = stroma_diffusion_coefficient_2_drug_498(&mi__vnr, &mi__dmn, &fld, 10),
             D1 = stroma_diffusion_coefficient_2_drug_498(&mi__vnr, &mi__dmn, &fld, 12);
    //
    std::cout << std::endl;
    std::cout << mi__vnr.set_biochem_param(210)/vi.pore_radius << std::endl;
    std::cout << mi__vnr.set_biochem_param(212)/vi.pore_radius << std::endl;
    std::cout << std::endl;
    std::cout << vascular_hydraulic_permeability_2_drug_498(&vi, &mi__vnr, &mi__dmn, &fld, 10) << std::endl;
    std::cout << vascular_hydraulic_permeability_2_drug_498(&vi, &mi__vnr, &mi__dmn, &fld, 12) << std::endl;
    std::cout << std::endl;
    std::cout << vascular_reflection_coefficient_2_drug_498(&vi, &mi__vnr, &mi__dmn, &fld, 10) << std::endl;
    std::cout << vascular_reflection_coefficient_2_drug_498(&vi, &mi__vnr, &mi__dmn, &fld, 12) << std::endl;
    std::cout << std::endl;
    std::cout << D0.comp << std::endl;
    std::cout << D1.comp << std::endl;
    std::cout << std::endl;
    //
    //
    //
    /*
    for (Real e_=-0.0001; e_>-2.5; e_-=0.05)
    {
        Real molecule_radius = molecule_radius = mi__vnr.get_biochem_param(210);
        // molecule_radius *= normal_distro(1., 0.050); // add some variability to this...
        //
        const Real lambda = molecule_radius / vi.pore_radius;
        // if ( lambda >= 1.0 ) return 0.0;
        const Real lambda_[] = { 1.0, lambda, pow2(lambda), pow3(lambda), pow4(lambda) };
        const Real delta_[] = { 1.0, (1.0-lambda), pow2(1.0-lambda) };
        //
        const Real F = delta_[2];
        const Real a_[] = { 0, -1.216667E+00,  1.533591E+00, -2.250830E+01, -5.617000E+00, -3.363000E-01, -1.216000E+00, 1.647000E+00 };
        const Real c_ = 3.140489E+01,
                   d_ = 1.0;//(1.0+lambda);
        //
        const Real K_t = c_*(1.0+a_[1]*delta_[1]+a_[2]*delta_[2])
                       + d_*(a_[3]*lambda_[0]+a_[4]*lambda_[1]+a_[5]*lambda_[2]+a_[6]*lambda_[3]+a_[7]*lambda_[4]);
        // hydrodynamic interactions of the drug
        const Real H = m6PI*F/K_t;
        //
        const Real absolute_temperature = 310.0; // Kelvin
        const Real fluid_viscosity = mi__vnr.set_biochem_param(0);
        // Stokes-Einstein equation: drug diffusion coefficient in free_solution at 310K
        const Real D_o = Boltzmann_constant*absolute_temperature / (m6PI*fluid_viscosity*molecule_radius);
        //
        const Real P_d = vi.pores_fraction*H*D_o/vi.thickness;
        //
        std::cout << molecule_radius << ' ' << lambda << ' ' << H << ' ' << P_d << std::endl;
    }*/
    std::cout << std::endl;
    //
    /*
    for (Real e_=-0.0001; e_>-2.5; e_-=0.05)
    {
        Real molecule_radius = molecule_radius = mi__vnr.get_biochem_param(212);
        // molecule_radius *= normal_distro(1., 0.050); // add some variability to this...
        //
        const Real lambda = molecule_radius / vi.pore_radius;
        // if ( lambda >= 1.0 ) return 0.0;
        const Real lambda_[] = { 1.0, lambda, pow2(lambda), pow3(lambda), pow4(lambda) };
        const Real delta_[] = { 1.0, (1.0-lambda), pow2(1.0-lambda) };
        //
        const Real F = delta_[2];
        const Real a_[] = { 0, -1.216667E+00,  1.533591E+00, -2.250830E+01, -5.617000E+00, -3.363000E-01, -1.216000E+00, 1.647000E+00 };
        const Real c_ = 3.140489E+01,
                   d_ = 1.0;//(1.0+lambda);
        //
        const Real K_t = c_*(1.0+a_[1]*delta_[1]+a_[2]*delta_[2])
                       + d_*(a_[3]*lambda_[0]+a_[4]*lambda_[1]+a_[5]*lambda_[2]+a_[6]*lambda_[3]+a_[7]*lambda_[4]);
        // hydrodynamic interactions of the drug
        const Real H = m6PI*F/K_t;
        //
        const Real absolute_temperature = 310.0; // Kelvin
        const Real fluid_viscosity = mi__vnr.set_biochem_param(0);
        // Stokes-Einstein equation: drug diffusion coefficient in free_solution at 310K
        const Real D_o = Boltzmann_constant*absolute_temperature / (m6PI*fluid_viscosity*molecule_radius);
        //
        const Real P_d = vi.pores_fraction*H*D_o/vi.thickness;
        //
        std::cout << molecule_radius << ' ' << lambda << ' ' << H << ' ' << P_d << std::endl;
    }*/
    std::cout << std::endl;
}
//=================================================================================================
// a sample differential equation "dy/dt = F(t,y)"
static
Real dydt (const Real t, const Real y)
{
    // return ((t-y)/2.0);
    return (y/3.0);
}
//
void my_test_6 () {
    // check this out:
    // https://www.geeksforgeeks.org/runge-kutta-4th-order-method-solve-differential-equation/
    Real t0 = 0., y0 = 1.,
         h = 0.001;
    std::cout << std::endl << " Operate the Runge-Kutta 4th order ODE solver" << std::endl;
    for (Real t=t0; t<=20.0; t+=1.0) {
        y0 = runge_kutta(t0, y0, &dydt, t, h);
        t0 = t;
        std::cout << t0 << ' ' << y0 << std::endl;
    }
    std::cout << std::endl;
}
//=================================================================================================
void my_test_5 () {
    if ( 0 ) {
        std::map<int, int> norm_hist{};
        for (int n=0; n<100000; ++n)
            ++norm_hist[std::round(100* normal_distro(1.00000, 0.10) )];
        std::cout << std::endl;
        for (auto p : norm_hist)
            std::cout << std::setw(4)
                      << p.first << ' ' << std::string(p.second/50, '*') << '\n';
    }
    //
    std::map<int, int> norm_hist{};
    for (int n=0; n<10000; ++n) {
        ++norm_hist[std::round( normal_distro(0.0, 3.0) )];
    }
    std::cout << std::endl;
    for (auto p : norm_hist) {
        std::cout << std::setw(2)
                  << p.first << ' ' << std::string(p.second/50, '*') << '\n';
    }
    //
    std::map<int, int> unif_hist{};
    for (int n=0; n<10000; ++n) {
        ++unif_hist[std::round( uniform_distro(0.0, 10.0) )];
    }
    std::cout << std::endl;
    for (auto p : unif_hist) {
        std::cout << std::setw(2)
                  << p.first << ' ' << std::string(p.second/50, '*') << '\n';
    }
    //
    std::cout << std::endl;
    for (int n=1; n<=100; ++n) {
        std::cout << ' ' << std::setw(6) << std::setprecision(3) << normal_distro(5.0, 2.0) << std::flush;
        if ( 0==n%25 ) std::cout << std::endl;
    }
    std::cout << std::endl;
    //
    std::cout << std::endl;
    for (int n=1; n<=100; ++n) {
        std::cout << ' ' << std::setw(6) << std::setprecision(3) << 5.0*normal_distro(1.0, 2./5.) << std::flush;
        if ( 0==n%25 ) std::cout << std::endl;
    }
    std::cout << std::endl;
}
//=================================================================================================
// local variables
static
std::vector<RealVectorValue> u__;
static
std::vector<RealVectorValue> dudx__;
static
std::vector<RealVectorValue> dudy__;
static
std::vector<RealVectorValue> dudz__;
static
void u_field    (const Point& xyz, RealVectorValue& u ) {
    u(0) = +111.1111*xyz(0);
    u(1) = +0.0000;
    u(2) = +0.0000;
}
static
void dudx_field (const Point& xyz, RealVectorValue& du) {
    du(0) = +111.1111;
    du(1) = +0.0000;
    du(2) = +0.0000;
}
static
void dudy_field (const Point& xyz, RealVectorValue& du) {
    du(0) = +0.0000;
    du(1) = +0.0000;
    du(2) = +0.0000;
}
static
void dudz_field (const Point& xyz, RealVectorValue& du) {
    du(0) = +0.0000;
    du(1) = +0.0000;
    du(2) = +0.0000;
}
static
void my_test_4___assembly_func (EquationSystems& es, const std::string& sname) {
    const MeshBase& msh = es.get_mesh();
    libmesh_assert( SP1D == msh.mesh_dimension() );
    LinearImplicitSystem& sys = es.get_system<LinearImplicitSystem>(sname);
    libmesh_assert( 3 == sys.n_vars() );
    //
    const MeshBase::const_element_iterator begin_el = msh.active_local_elements_begin();
    const MeshBase::const_element_iterator end_el = msh.active_local_elements_end();
    //
    Real system_matrix_max_local  = 0.0,
         system_matrix_max_global = 0.0;
    // zero-out the system matrix and RHS vector
    sys.matrix->zero();
    sys.rhs->zero();
    //
    std::vector<Number> displacement;
    sys.update_global_solution(displacement);
    //
    const unsigned int nvars = 3;
    const unsigned int u_var[] = { sys.variable_number("u_X") ,
                                   sys.variable_number("u_Y") ,
                                   sys.variable_number("u_Z") };
    //
    const Real alpha = es.parameters.get<Real>("IDA exponent");
    const RVE* rve_ptr = es.parameters.get<RVE*>("representative volume element");
    const Real Af = es.parameters.get<Real>("fibril cross section area"),
               Ef_T = es.parameters.get<Real>("fibril stiffness (tension)"),
               Ef_C = es.parameters.get<Real>("fibril stiffness (compression)");
    //
    // loop over all active local 1-D finite elements
    for (MeshBase::const_element_iterator ci_elem=begin_el; ci_elem!=end_el; ci_elem++) {
        const Elem* elem = (*ci_elem);
        if (libMesh::EDGE2!=elem->type()) libmesh_error();
        //
        std::vector<unsigned int> dof_indices;
        sys.get_dof_map().dof_indices(elem, dof_indices);
        //
        std::vector<unsigned int> dof_indices_u[nvars];
        for (unsigned int ivar=0; ivar<nvars; ivar++)
            sys.get_dof_map().dof_indices(elem, dof_indices_u[ivar], u_var[ivar]);
        //
        const unsigned long var_dofs[] = { dof_indices_u[0].size() ,
                                           dof_indices_u[1].size() ,
                                           dof_indices_u[2].size() };
        const unsigned long offset[] = { 0                                   ,
                                         var_dofs[0]                         ,
                                         var_dofs[0]+var_dofs[1]             ,
                                         var_dofs[0]+var_dofs[1]+var_dofs[2] };
        //
        DenseMatrix<Number> Ke(dof_indices.size(), dof_indices.size());
        std::vector< DenseSubMatrix<Number> > Kesub(nvars*nvars, Ke);
        DenseVector<Number> Fe(dof_indices.size());
        std::vector< DenseSubVector<Number> > Fesub(nvars, Fe);
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            for (unsigned int jvar=0; jvar<nvars; jvar++)
                Kesub[ivar*nvars+jvar].reposition(offset[ivar], offset[jvar], var_dofs[ivar], var_dofs[jvar]);
            Fesub[ivar].reposition(offset[ivar], var_dofs[ivar]);
        }
        //
        const Point e[] = { Point(1., 0., 0.) , Point(0., 1., 0.) , Point(0., 0., 1.) },
                    l = Point(elem->point(1)-elem->point(0)).unit(),
                    u[] = { Point( displacement[elem->node_ref(0).dof_number(sys.number(), 0, 0)] ,
                                   displacement[elem->node_ref(0).dof_number(sys.number(), 1, 0)] ,
                                   displacement[elem->node_ref(0).dof_number(sys.number(), 2, 0)] ) ,
                            Point( displacement[elem->node_ref(1).dof_number(sys.number(), 0, 0)] ,
                                   displacement[elem->node_ref(1).dof_number(sys.number(), 1, 0)] ,
                                   displacement[elem->node_ref(1).dof_number(sys.number(), 2, 0)] ) };
        //
        const Real L = (elem->point(1)     -elem->point(0)     ).norm(),
                   s = (elem->point(1)+u[1]-elem->point(0)-u[0]).norm() / L, // stretch ratio
                   Ef = ( s >= 1.0 ? Ef_T : Ef_C );
        const Real C[] = { (e[0]*l),  (e[1]*l),  (e[2]*l) };
        //
        for (unsigned int a=0; a<var_dofs[0]; a++) {
            for (unsigned int b=0; b<var_dofs[0]; b++) {
                const Real coef = (Af*Ef/L)*( a == b ? +1 : -1 );
                Kesub[0*nvars+0](a,b) += coef*C[0]*C[0];
                Kesub[0*nvars+1](a,b) += coef*C[0]*C[1];
                Kesub[0*nvars+2](a,b) += coef*C[0]*C[2];
                Kesub[1*nvars+0](a,b) += coef*C[1]*C[0];
                Kesub[1*nvars+1](a,b) += coef*C[1]*C[1];
                Kesub[1*nvars+2](a,b) += coef*C[1]*C[2];
                Kesub[2*nvars+0](a,b) += coef*C[2]*C[0];
                Kesub[2*nvars+1](a,b) += coef*C[2]*C[1];
                Kesub[2*nvars+2](a,b) += coef*C[2]*C[2];
            }
        }
        // find the maximum value of the stiffness matrix
        system_matrix_max_local = std::max(system_matrix_max_local, Ke.max());
        // update the system stiffness matrix and RHS vector
        sys.matrix->add_matrix(Ke, dof_indices);
        sys.rhs->add_vector(Fe, dof_indices);
        // end of active local 1-D finite elements loop...
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // evaluate the penalty parameter for the stiffness matrix
    MPI_Allreduce(&system_matrix_max_local, &system_matrix_max_global, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // set the penalty parameter for the system matrix
    const Real penalty = system_matrix_max_global
                       * es.parameters.get<Real>("system matrix penalty parameter");
    //
    // loop over all active local 1-D finite elements
    for (MeshBase::const_element_iterator ci_elem=begin_el; ci_elem!=end_el; ci_elem++) {
        const Elem* elem = (*ci_elem);
        if (libMesh::EDGE2!=elem->type()) libmesh_error();
        //
        std::vector<unsigned int> dof_indices;
        sys.get_dof_map().dof_indices(elem, dof_indices);
        //
        std::vector<unsigned int> dof_indices_u[nvars];
        for (unsigned int ivar=0; ivar<nvars; ivar++)
            sys.get_dof_map().dof_indices(elem, dof_indices_u[ivar], u_var[ivar]);
        //
        const unsigned long var_dofs[] = { dof_indices_u[0].size() ,
                                           dof_indices_u[1].size() ,
                                           dof_indices_u[2].size() };
        const unsigned long offset[] = { 0                                   ,
                                         var_dofs[0]                         ,
                                         var_dofs[0]+var_dofs[1]             ,
                                         var_dofs[0]+var_dofs[1]+var_dofs[2] };
        //
        DenseMatrix<Number> Ke(dof_indices.size(), dof_indices.size());
        std::vector< DenseSubMatrix<Number> > Kesub(nvars*nvars, Ke);
        DenseVector<Number> Fe(dof_indices.size());
        std::vector< DenseSubVector<Number> > Fesub(nvars, Fe);
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            for (unsigned int jvar=0; jvar<nvars; jvar++)
                Kesub[ivar*nvars+jvar].reposition(offset[ivar], offset[jvar], var_dofs[ivar], var_dofs[jvar]);
            Fesub[ivar].reposition(offset[ivar], var_dofs[ivar]);
        }
        //
        InverseDistanceAlgorithm ida(alpha);
        ida.init(rve_ptr->vertices());
        //
        int face =-1;
        for (unsigned int ne=0; ne<elem->n_nodes(); ne++)
        if ( ! elem->neighbor_ptr(ne) ||
             rve_ptr->node_is_boundary(elem->node_id(ne), face) ) {
            // ignore the following if edge-point is not a "hanging" node
            RealVectorValue u;
            ida.interp<RealVectorValue>(elem->point(ne), u__, u);
            //
            Kesub[0*nvars+0](ne,ne) += penalty;
            Kesub[1*nvars+1](ne,ne) += penalty;
            Kesub[2*nvars+2](ne,ne) += penalty;
            Fesub[0](ne) += penalty * u(0);
            Fesub[1](ne) += penalty * u(1);
            Fesub[2](ne) += penalty * u(2);
            // end of element nodes loop...
        }
        // update the system stiffness matrix and RHS vector
        sys.matrix->add_matrix(Ke, dof_indices);
        sys.rhs->add_vector(Fe, dof_indices);
        // end of active local 1-D finite elements loop...
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // now close the system matrix and RHS vector
    sys.matrix->close();
    sys.rhs->close();
}
static
void my_test_4___post_process (EquationSystems& es) {
    const MeshBase& msh = es.get_mesh();
    libmesh_assert( SP1D == msh.mesh_dimension() );
    LinearImplicitSystem& sys = es.get_system<LinearImplicitSystem>("RVE_Balance");
    libmesh_assert( SP3D == sys.n_vars() );
    //
    const MeshBase::const_element_iterator begin_el = msh.active_elements_begin();
    const MeshBase::const_element_iterator end_el = msh.active_elements_end();
    //
    std::vector<Real> displacement;
    sys.update_global_solution(displacement);
    //
    const Real alpha = es.parameters.get<Real>("IDA exponent");
    const RVE* rve_ptr = es.parameters.get<RVE*>("representative volume element");
    const Real Af = es.parameters.get<Real>("fibril cross section area"),
               Ef_T = es.parameters.get<Real>("fibril stiffness (tension)"),
               Ef_C = es.parameters.get<Real>("fibril stiffness (compression)");
    const Real theta = es.parameters.get<Real>("fibre volume fraction"),
               LDF = es.parameters.get<Real>("length downscaling factor");
    //
    // macroscopic stress tensor
    RealTensorValue S_ij(0., 0., 0.,
                         0., 0., 0.,
                         0., 0., 0.);
    // microscopic stress, space vector, U-field spatial derivatives, traction
    // ...all expressed @boundary points of the RVE per [ faceID ]
    std::vector<RealTensorValue> sigma[6]; // NOTE: RVE has 6 faces by default!!!
    std::vector<Point> xyz[6]; // NOTE: RVE has 6 faces by default!!!
    std::vector<RealVectorValue> dudx[6], dudy[6], dudz[6]; // NOTE: RVE has 6 faces by default!!!
    std::vector<RealVectorValue> traction[6]; // NOTE: RVE has 6 faces by default!!!
    //
    // loop over all active local 1-D finite elements
    for (MeshBase::const_element_iterator ci_elem=begin_el; ci_elem!=end_el; ci_elem++) {
        const Elem* elem = (*ci_elem);
        if (libMesh::EDGE2!=elem->type()) libmesh_error();
        //
        const Point e[] = { Point(1., 0., 0.) , Point(0., 1., 0.) , Point(0., 0., 1.) },
                    l = Point(elem->point(1)-elem->point(0)).unit(),
                    u[] = { Point( displacement[elem->node_ref(0).dof_number(sys.number(), 0, 0)] ,
                                   displacement[elem->node_ref(0).dof_number(sys.number(), 1, 0)] ,
                                   displacement[elem->node_ref(0).dof_number(sys.number(), 2, 0)] ) ,
                            Point( displacement[elem->node_ref(1).dof_number(sys.number(), 0, 0)] ,
                                   displacement[elem->node_ref(1).dof_number(sys.number(), 1, 0)] ,
                                   displacement[elem->node_ref(1).dof_number(sys.number(), 2, 0)] ) };
        //
        const Real L = (elem->point(1)     -elem->point(0)     ).norm(),
                   s = (elem->point(1)+u[1]-elem->point(0)-u[0]).norm() / L, // stretch ratio
                   Ef = ( s >= 1.0 ? Ef_T : Ef_C );
        const Real F = (Af*Ef)*(s-1.0);
        //
        int face =-1;
        for (unsigned int ne=0; ne<elem->n_nodes(); ne++)
        if ( rve_ptr->node_is_boundary(elem->node_id(ne), face) ) {
            if (face<0) libmesh_error();
            //
            const Point& eta_B = rve_ptr->normal(face);
            //
            Point eta_F;
            if      ( 0 == ne && s >= 1.0 ) eta_F = Point(elem->point(0)-elem->point(1)).unit();
            else if ( 0 == ne && s  < 1.0 ) eta_F = Point(elem->point(1)-elem->point(0)).unit();
            else if ( 1 == ne && s >= 1.0 ) eta_F = Point(elem->point(1)-elem->point(0)).unit();
            else if ( 1 == ne && s  < 1.0 ) eta_F = Point(elem->point(0)-elem->point(1)).unit();
            else libmesh_error();
            //
            const Point pnt = elem->point(ne) - rve_ptr->centre();
            // const Point pnt = elem->point(ne);
            //
            for (unsigned int i=0; i<SP3D; i++)
                for (unsigned int j=0; j<SP3D; j++)
                    S_ij(i,j) += (1.0/rve_ptr->volume())
                               * pnt(i)*F*eta_F(j);
            //
            RealTensorValue sm_ij;
            for (unsigned int i=0; i<SP3D; i++)
                for (unsigned int j=0; j<SP3D; j++)
                    sm_ij(i,j) = eta_F(i)*F*eta_B(j);
            //
            std::vector<Point> xyz__fv;  // face vertices
            std::vector<RealVectorValue> dudx__fv; // face vertices
            std::vector<RealVectorValue> dudy__fv; // face vertices
            std::vector<RealVectorValue> dudz__fv; // face vertices
            if        ( 0 == face ) {
                xyz__fv.push_back(rve_ptr->vertex(0)); dudx__fv.push_back(dudx__[0]); dudy__fv.push_back(dudy__[0]); dudz__fv.push_back(dudz__[0]);
                xyz__fv.push_back(rve_ptr->vertex(1)); dudx__fv.push_back(dudx__[1]); dudy__fv.push_back(dudy__[1]); dudz__fv.push_back(dudz__[1]);
                xyz__fv.push_back(rve_ptr->vertex(2)); dudx__fv.push_back(dudx__[2]); dudy__fv.push_back(dudy__[2]); dudz__fv.push_back(dudz__[2]);
                xyz__fv.push_back(rve_ptr->vertex(3)); dudx__fv.push_back(dudx__[3]); dudy__fv.push_back(dudy__[3]); dudz__fv.push_back(dudz__[3]);
            } else if ( 1 == face ) {
                xyz__fv.push_back(rve_ptr->vertex(0)); dudx__fv.push_back(dudx__[0]); dudy__fv.push_back(dudy__[0]); dudz__fv.push_back(dudz__[0]);
                xyz__fv.push_back(rve_ptr->vertex(1)); dudx__fv.push_back(dudx__[1]); dudy__fv.push_back(dudy__[1]); dudz__fv.push_back(dudz__[1]);
                xyz__fv.push_back(rve_ptr->vertex(4)); dudx__fv.push_back(dudx__[4]); dudy__fv.push_back(dudy__[4]); dudz__fv.push_back(dudz__[4]);
                xyz__fv.push_back(rve_ptr->vertex(5)); dudx__fv.push_back(dudx__[5]); dudy__fv.push_back(dudy__[5]); dudz__fv.push_back(dudz__[5]);
            } else if ( 2 == face ) {
                xyz__fv.push_back(rve_ptr->vertex(1)); dudx__fv.push_back(dudx__[1]); dudy__fv.push_back(dudy__[1]); dudz__fv.push_back(dudz__[1]);
                xyz__fv.push_back(rve_ptr->vertex(2)); dudx__fv.push_back(dudx__[2]); dudy__fv.push_back(dudy__[2]); dudz__fv.push_back(dudz__[2]);
                xyz__fv.push_back(rve_ptr->vertex(5)); dudx__fv.push_back(dudx__[5]); dudy__fv.push_back(dudy__[5]); dudz__fv.push_back(dudz__[5]);
                xyz__fv.push_back(rve_ptr->vertex(6)); dudx__fv.push_back(dudx__[6]); dudy__fv.push_back(dudy__[6]); dudz__fv.push_back(dudz__[6]);
            } else if ( 3 == face ) {
                xyz__fv.push_back(rve_ptr->vertex(2)); dudx__fv.push_back(dudx__[2]); dudy__fv.push_back(dudy__[2]); dudz__fv.push_back(dudz__[2]);
                xyz__fv.push_back(rve_ptr->vertex(3)); dudx__fv.push_back(dudx__[3]); dudy__fv.push_back(dudy__[3]); dudz__fv.push_back(dudz__[3]);
                xyz__fv.push_back(rve_ptr->vertex(6)); dudx__fv.push_back(dudx__[6]); dudy__fv.push_back(dudy__[6]); dudz__fv.push_back(dudz__[6]);
                xyz__fv.push_back(rve_ptr->vertex(7)); dudx__fv.push_back(dudx__[7]); dudy__fv.push_back(dudy__[7]); dudz__fv.push_back(dudz__[7]);
            } else if ( 4 == face ) {
                xyz__fv.push_back(rve_ptr->vertex(0)); dudx__fv.push_back(dudx__[0]); dudy__fv.push_back(dudy__[0]); dudz__fv.push_back(dudz__[0]);
                xyz__fv.push_back(rve_ptr->vertex(3)); dudx__fv.push_back(dudx__[3]); dudy__fv.push_back(dudy__[3]); dudz__fv.push_back(dudz__[3]);
                xyz__fv.push_back(rve_ptr->vertex(4)); dudx__fv.push_back(dudx__[4]); dudy__fv.push_back(dudy__[4]); dudz__fv.push_back(dudz__[4]);
                xyz__fv.push_back(rve_ptr->vertex(7)); dudx__fv.push_back(dudx__[7]); dudy__fv.push_back(dudy__[7]); dudz__fv.push_back(dudz__[7]);
            } else if ( 5 == face ) {
                xyz__fv.push_back(rve_ptr->vertex(4)); dudx__fv.push_back(dudx__[4]); dudy__fv.push_back(dudy__[4]); dudz__fv.push_back(dudz__[4]);
                xyz__fv.push_back(rve_ptr->vertex(5)); dudx__fv.push_back(dudx__[5]); dudy__fv.push_back(dudy__[5]); dudz__fv.push_back(dudz__[5]);
                xyz__fv.push_back(rve_ptr->vertex(6)); dudx__fv.push_back(dudx__[6]); dudy__fv.push_back(dudy__[6]); dudz__fv.push_back(dudz__[6]);
                xyz__fv.push_back(rve_ptr->vertex(7)); dudx__fv.push_back(dudx__[7]); dudy__fv.push_back(dudy__[7]); dudz__fv.push_back(dudz__[7]);
            } else libmesh_error();
            //
            InverseDistanceAlgorithm ida(alpha);
            ida.init(xyz__fv);
            RealVectorValue dud_;
            // push the coordinates of this boundary point
            xyz[face].push_back( elem->point(ne) );
            // push the stresses at this boundary point
            sigma[face].push_back( sm_ij );
            // now interpolate & push the gradient component of this boundary point
            ida.interp<RealVectorValue>(elem->point(ne), dudx__fv, dud_);
            dudx[face].push_back( dud_ );
            // now interpolate & push the gradient component of this boundary point
            ida.interp<RealVectorValue>(elem->point(ne), dudy__fv, dud_);
            dudy[face].push_back( dud_ );
            // now interpolate & push the gradient component of this boundary point
            ida.interp<RealVectorValue>(elem->point(ne), dudz__fv, dud_);
            dudz[face].push_back( dud_ );
            // push the traction at this boundary point
            traction[face].push_back( eta_F*F );
            // end of element nodes loop...
        }
        // end of active local 1-D finite elements loop...
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    // declare the micro/macro source term
    RealVectorValue Q_source(0., 0., 0.);
    RealVectorValue T_face[6] = { RealVectorValue(0., 0., 0.) ,
                                  RealVectorValue(0., 0., 0.) ,
                                  RealVectorValue(0., 0., 0.) ,
                                  RealVectorValue(0., 0., 0.) ,
                                  RealVectorValue(0., 0., 0.) ,
                                  RealVectorValue(0., 0., 0.) };
    // loop over all 6 faces of the RVE cube
    for (unsigned int face=0; face<6; face++) {
        const Point& eta_B = rve_ptr->normal(face);
        //
        InverseDistanceAlgorithm ida(alpha);
        ida.init(xyz[face]);
        //
        for (unsigned int qp=0; qp<4; qp++) {
            // interpolate to obtain the (microscopic) stresses at the quadrature point
            RealTensorValue sigma__qp;
            ida.interp<RealTensorValue>(rve_ptr->quadrature(face)[qp], sigma[face], sigma__qp);
            // interpolate to obtain the U-field gradients at the quadrature point
            RealVectorValue dudx__qp;
            ida.interp<RealVectorValue>(rve_ptr->quadrature(face)[qp], dudx[face], dudx__qp);
            RealVectorValue dudy__qp;
            ida.interp<RealVectorValue>(rve_ptr->quadrature(face)[qp], dudy[face], dudy__qp);
            RealVectorValue dudz__qp;
            ida.interp<RealVectorValue>(rve_ptr->quadrature(face)[qp], dudz[face], dudz__qp);
            RealVectorValue traction__qp;
            ida.interp<RealVectorValue>(rve_ptr->quadrature(face)[qp], traction[face], traction__qp);
            //
            RealVectorValue n__du;
            n__du(0) = eta_B(0)*dudx__qp(0) + eta_B(1)*dudx__qp(1) + eta_B(2)*dudx__qp(2);
            n__du(1) = eta_B(0)*dudy__qp(0) + eta_B(1)*dudy__qp(1) + eta_B(2)*dudy__qp(2);
            n__du(2) = eta_B(0)*dudz__qp(0) + eta_B(1)*dudz__qp(1) + eta_B(2)*dudz__qp(2);
            //
            const RealTensorValue Ds = (sigma__qp - S_ij).transpose();
            //
            // now compute the "Q" (vector) source term
            Q_source += rve_ptr->jacobian()
                      * (1.0/rve_ptr->volume())
                      * (Ds * n__du);
            //
            T_face[face] += rve_ptr->jacobian()
                          * (1.0/rve_ptr->side_surface())
                          * traction__qp;
        }
        // end of RVE faces loop...
    }
    //
    const Real J = (theta/(Af*LDF*LDF))
                 * (rve_ptr->volume()/rve_ptr->fibres_length());
    //
    // std::cout
    // << std::endl
    // << ' ' << J*S_ij(0,0) << ' ' << J*S_ij(0,1) << ' ' << J*S_ij(0,2) << std::endl
    // << ' ' << J*S_ij(1,0) << ' ' << J*S_ij(1,1) << ' ' << J*S_ij(1,2) << std::endl
    // << ' ' << J*S_ij(2,0) << ' ' << J*S_ij(2,1) << ' ' << J*S_ij(2,2) << std::endl
    // << std::endl
    // << " F2(+X) " << T_face[2] << std::endl
    // << " F4(-X) " << T_face[4] << std::endl
    // << " F3(+Y) " << T_face[3] << std::endl
    // << " F1(-Y) " << T_face[1] << std::endl
    // << " F5(+Z) " << T_face[5] << std::endl
    // << " F0(-Z) " << T_face[0] << std::endl
    // << std::endl;
    //
    RealTensorValue S_ij_Sym(0., 0., 0.,
                             0., 0., 0.,
                             0., 0., 0.);
    S_ij_Sym(0,0) = J*S_ij(0,0);
    S_ij_Sym(1,1) = J*S_ij(1,1);
    S_ij_Sym(2,2) = J*S_ij(2,2);
    S_ij_Sym(0,1) = J*(S_ij(0,1)+S_ij(1,0))/2.0;
    S_ij_Sym(1,0) = J*(S_ij(0,1)+S_ij(1,0))/2.0;
    S_ij_Sym(0,2) = J*(S_ij(0,2)+S_ij(2,0))/2.0;
    S_ij_Sym(2,0) = J*(S_ij(0,2)+S_ij(2,0))/2.0;
    S_ij_Sym(1,2) = J*(S_ij(1,2)+S_ij(2,1))/2.0;
    S_ij_Sym(2,1) = J*(S_ij(1,2)+S_ij(2,1))/2.0;
    //
    RealVectorValue Q_i(0., 0., 0.);
    Q_i(0) = J*Q_source(0);
    Q_i(1) = J*Q_source(1);
    Q_i(2) = J*Q_source(2);
    //
    std::cout
    << " macroscopic stress tensor (in symmetric & Voigt form) " << std::endl
    << ' ' << S_ij_Sym(0,0) << ' ' << S_ij_Sym(1,1) << ' ' << S_ij_Sym(2,2)
    << ' ' << S_ij_Sym(1,2) << ' ' << S_ij_Sym(0,2) << ' ' << S_ij_Sym(0,1)
    << std::endl
    << " source term vector " << std::endl
    << ' ' << Q_i(0) << ' ' << Q_i(1) << ' ' << Q_i(2)
    << std::endl
    << std::endl;
}
//
void my_test_4 () {
    const unsigned int n_fibre = 1000; // max number of fibres in RVE
    const Real y_min = 0.0, // min dimension of RVE, in [μm]
               y_max = 3.0; // max dimension of RVE, in [μm]

    // declare the RVE mesh, generate it and save it in GMSH format
    RVE rve(*gcomm_ptr);
    GmshIO gmsh(rve);
    ExodusII_IO ex2(rve);
    rve.generate(n_fibre, y_min, y_max); // RVE cubic domain, in [μm^3]
    rve.print_info();
    rve.write("my_test_4.rve.msh");
    gmsh.write("my_test_4.rve.libmesh.msh");
    rve.reinit();

    // initialise field values & spatial derivatives
    u__.resize(8);
    dudx__.resize(8);
    dudy__.resize(8);
    dudz__.resize(8);
    for (int i=0; i<8; i++) {
        u_field   (rve.vertex(i), u__[i]   );
        dudx_field(rve.vertex(i), dudx__[i]);
        dudy_field(rve.vertex(i), dudy__[i]);
        dudz_field(rve.vertex(i), dudz__[i]);
    }

    // setup the PETSc linear solver to direct mode
    PetscOptionsSetValue(NULL, "-ksp_type", "preonly");
    PetscOptionsSetValue(NULL, "-pc_type", "lu");
    PetscOptionsSetValue(NULL, "-pc_factor_mat_solver_package", "superlu_dist");
    PetscOptionsSetValue(NULL, "-pc_factor_mat_solver_package", "mumps");

    EquationSystems es(rve);
    LinearImplicitSystem& sys = es.add_system<LinearImplicitSystem>("RVE_Balance");
    sys.add_variable("u_X", FIRST, LAGRANGE);
    sys.add_variable("u_Y", FIRST, LAGRANGE);
    sys.add_variable("u_Z", FIRST, LAGRANGE);
    sys.attach_assemble_function(my_test_4___assembly_func);
    es.init();
    es.print_info();

    es.parameters.set<Real>("length downscaling factor") = 1.0e-6; // from [m] to [μm]
    es.parameters.set<Real>("fibril cross section area") = 0.3; // in [μm^2]
    es.parameters.set<Real>("fibril stiffness (tension)") = 30.0e-6; // in [N/μm^2]
    es.parameters.set<Real>("fibril stiffness (compression)") = 3.0e-6; // in [N/μm^2]
    es.parameters.set<Real>("fibre volume fraction") = 0.3; // dimensionless
    es.parameters.set<Real>("system matrix penalty parameter") = 1.0e+5;
    es.parameters.set<Real>("IDA exponent") = 2.5;
    es.parameters.set<RVE*>("representative volume element") = &rve;

    es.get_system<LinearImplicitSystem>("RVE_Balance").solve();
    ex2.write_equation_systems("my_test_4.results.ex2", es);
    gmsh.write_equation_systems("my_test_4.results.post", es);
    try { my_test_4___post_process(es); }
    catch (...) { }
}
//=================================================================================================
void my_test_3 () {
    Mesh msh(*gcomm_ptr, SP3D);
    EquationSystems es(msh);
    std::map<dof_id_type, Mesh> msh_m;

    gcomm_ptr->barrier();

    std::cout << " vector of meshes - step #0\r";
    es.parameters.set<std::map<dof_id_type, Mesh>*>("pointer to RVE parametric mesh(es)") = &msh_m;

    gcomm_ptr->barrier();

    std::map<dof_id_type, Mesh>& msh_m_r = *(es.parameters.set<std::map<dof_id_type, Mesh>*>("pointer to RVE parametric mesh(es)"));

    std::cout << " vector of meshes - step #1\r";

    for (unsigned int i=0; i<5; i++) {
        msh_m_r.insert( std::make_pair(i, Mesh(*gcomm_ptr, SP1D)) );
        //
        GmshIO(msh_m_r.find(i)->second).read("./src/tests/rve.msh");
        msh_m_r.find(i)->second.prepare_for_use();
        // msh_m_r.find(i)->second.print_info();
    }

    gcomm_ptr->barrier();

    std::cout << " vector of meshes - step #2\r";

    for (unsigned int i=5; i<100; i++) {
        msh_m_r.insert( std::make_pair(i, Mesh(*gcomm_ptr, SP1D)) );
        //
        GmshIO(msh_m_r.find(i)->second).read("./src/tests/rve.msh");
        msh_m_r.find(i)->second.prepare_for_use();
        // msh_m_r.find(i)->second.print_info();
    }

    gcomm_ptr->barrier();
}
//=================================================================================================
static
Real u_field (const Point& coords, int i) {
    const Real u_[] = { 1.8 * coords(0) - 0.5 * coords(1) + 0.4,
                        0.2 * coords(0) + 0.1 * coords(2) - 1.3,
                      - 0.1 * coords(1) - 0.7 * coords(2) - 0.2 };
    return u_[i];
}
//
void my_test_2 () {
    Mesh msh(*gcomm_ptr, SP3D);
    const Real L = 30.0;
    MeshTools::Generation::build_cube( msh, 1, 1, 1,
                                       0.0, L, 0.0, L, 0.0, L, HEX8);
    msh.print_info();
    //
    EquationSystems es(msh);
    es.add_system<System>("a_system");
    es.get_system<System>("a_system").add_variable("u_0", FIRST, LAGRANGE);
    es.get_system<System>("a_system").add_variable("u_1", FIRST, LAGRANGE);
    es.get_system<System>("a_system").add_variable("u_2", FIRST, LAGRANGE);
    es.init();
    es.print_info();
    //
    System& sys = es.get_system<System>("a_system");
    //
    for (const auto & elem : msh.active_local_element_ptr_range()) {
        std::vector<dof_id_type> dof_indices;
        sys.get_dof_map().dof_indices(elem, dof_indices);
        //
        // QGauss qrule(SP3D, CONSTANT);
        QGauss qrule(SP3D, SECOND);
        //
        std::unique_ptr<FEBase> fe(FEBase::build(SP3D, sys.get_dof_map().variable_type(0)));
        fe->attach_quadrature_rule(&qrule);
        //
        const std::vector<Point>& pnt = fe->get_xyz();
        const std::vector<std::vector<Real>> & phi = fe->get_phi();
        const std::vector<std::vector<RealGradient>>& dphi_dX = fe->get_dphi();
        const std::vector<std::vector<Real>> & dphi_dZ0 = fe->get_dphidxi();
        const std::vector<std::vector<Real>> & dphi_dZ1 = fe->get_dphideta();
        const std::vector<std::vector<Real>> & dphi_dZ2 = fe->get_dphidzeta();
        const std::vector<Real>& dZ0_dX0 = fe->get_dxidx();
        const std::vector<Real>& dZ0_dX1 = fe->get_dxidy();
        const std::vector<Real>& dZ0_dX2 = fe->get_dxidz();
        const std::vector<Real>& dZ1_dX0 = fe->get_detadx();
        const std::vector<Real>& dZ1_dX1 = fe->get_detady();
        const std::vector<Real>& dZ1_dX2 = fe->get_detadz();
        const std::vector<Real>& dZ2_dX0 = fe->get_dzetadx();
        const std::vector<Real>& dZ2_dX1 = fe->get_dzetady();
        const std::vector<Real>& dZ2_dX2 = fe->get_dzetadz();
        fe->reinit(elem);
        const std::vector<Point>& q_pnt = qrule.get_points();
        //
        for (unsigned int qp=0; qp<qrule.n_points(); qp++) {
            Real u[SP3D] = { 0.0 ,
                             0.0 ,
                             0.0 };
            Vector3D du_dX[SP3D];
            for (unsigned int i=0; i<phi.size(); i++) {
                for (unsigned int s=0; s<SP3D; s++)
                    u[s] += phi[i][qp] * u_field(elem->point(i), s);
                //
                for (unsigned int s=0; s<SP3D; s++)
                    du_dX[s].comp += vec(dphi_dX[i][qp]).comp * u_field(elem->point(i), s);
            }
            //
            Vector3D du_dZ[SP3D];
            for (unsigned int i=0; i<phi.size(); i++) {
                for (unsigned int s=0; s<SP3D; s++)
                    du_dZ[s].comp(0) += dphi_dZ0[i][qp] * u_field(elem->point(i), s);
                //
                for (unsigned int s=0; s<SP3D; s++)
                    du_dZ[s].comp(1) += dphi_dZ1[i][qp] * u_field(elem->point(i), s);
                //
                for (unsigned int s=0; s<SP3D; s++)
                    du_dZ[s].comp(2) += dphi_dZ2[i][qp] * u_field(elem->point(i), s);
            }
            //
            Dyadic3D dZ_dX;
            dZ_dX.comp = dZ0_dX0[qp], dZ0_dX1[qp], dZ0_dX2[qp],
                         dZ1_dX0[qp], dZ1_dX1[qp], dZ1_dX2[qp],
                         dZ2_dX0[qp], dZ2_dX1[qp], dZ2_dX2[qp];
            //
            Vector3D du_dX__[SP3D];
            for (unsigned int s=0; s<SP3D; s++)
                du_dX__[s] = mul_m_mi(du_dZ[s], dZ_dX);
            //
            Vector3D X;
            X = set_vector(pnt[qp](0), pnt[qp](1), pnt[qp](2));
            //
            Vector3D Z;
            Z.comp = (dZ_dX*X).comp - 1.0;
            //
            Vector3D dZ;
            dZ.comp = vec(q_pnt[qp]).comp - Z.comp;
            //
            std::cout << " ***************** " << std::flush;
            std::cout << " QP = " << qp << " \n" << std::endl;
            std::cout << ' ' << X.comp << std::endl;
            // std::cout << ' ' << vec(q_pnt[qp]).comp << std::endl;
            // std::cout << ' ' << Z.comp << std::endl;
            std::cout << ' ' << dZ.comp << std::endl;
            // std::cout << ' ' << u[0] << ' ' << u[1] << ' ' << u[2] << std::endl;
            // std::cout << ' ' << u_field(pnt[qp], 0) << ' ' << u_field(pnt[qp], 1) << ' ' << u_field(pnt[qp], 2) << std::endl;
            // std::cout << std::endl;
            // std::cout << ' ' << u[0] << std::endl;
            // std::cout << ' ' << du_dX[0](0) << ' ' << du_dX[0](1) << ' ' << du_dX[0](2) << std::endl;
            // std::cout << ' ' << du_dX__[0](0) << ' ' << du_dX__[0](1) << ' ' << du_dX__[0](2) << std::endl;
            // std::cout << ' ' << du_dZ[0](0) << ' ' << du_dZ[0](1) << ' ' << du_dZ[0](2) << std::endl;
            // std::cout << std::endl;
            // std::cout << ' ' << u[1] << std::endl;
            // std::cout << ' ' << du_dX[1](0) << ' ' << du_dX[1](1) << ' ' << du_dX[1](2) << std::endl;
            // std::cout << ' ' << du_dX__[1](0) << ' ' << du_dX__[1](1) << ' ' << du_dX__[1](2) << std::endl;
            // std::cout << ' ' << du_dZ[1](0) << ' ' << du_dZ[1](1) << ' ' << du_dZ[1](2) << std::endl;
            // std::cout << std::endl;
            // std::cout << ' ' << u[2] << std::endl;
            // std::cout << ' ' << du_dX[2](0) << ' ' << du_dX[2](1) << ' ' << du_dX[2](2) << std::endl;
            // std::cout << ' ' << du_dX__[2](0) << ' ' << du_dX__[2](1) << ' ' << du_dX__[2](2) << std::endl;
            // std::cout << ' ' << du_dZ[2](0) << ' ' << du_dZ[2](1) << ' ' << du_dZ[2](2) << std::endl;
            std::cout << std::endl;
            //
        }
        //
    }
    //
}
//=================================================================================================
void my_test_1 () {
    Dyadic3D F;
    F.comp =  2.9,  0.1, -0.5,
              0.3,  1.4,  0.2,
             -0.1,  0.0, -1.2;
    const Dyadic3D E = GreenLagrange_strain(F);
    // std::cout << F.comp << std::endl;
    std::cout << E.comp << std::endl;
    //
    std::vector<Real> e;
    std::vector<Vector3D> E_a;
    eigen_solve(E, e, E_a);
    std::vector<Real> l;
    for (int i=0; i<3; i++) l.push_back(sqrt(2.0*e[i]+1.0));
    //
    for (int i=0; i<3; i++) std::cout << l[i] << " *** \n" << E_a[i].comp << std::endl;
    std::cout << std::endl;
    //
    MaterialInfo mi__vnr(500, 500);
    mi__vnr.set_biochem_param(3) = 350.000e-5; // Pa*s
    mi__vnr.set_biochem_param(210) = 0.500e-9; // m (mean radius of the drug)
    mi__vnr.set_biochem_param(211) = 5.000e-9; // m (mean radius of the drug)
    mi__vnr.set_biochem_param(212) = 50.00e-9; // m (mean radius of the drug)
    //
    MaterialInfo mi__dmn(500, 500);
    mi__dmn.set_biochem_param(70) = 0.064e-13; // m^2/(Pa*s) (interstitial hydraulic conductivity)
    mi__dmn.set_biochem_param(71) = -100.0;    // -- (scaling parameter for the interstitial hydraulic conductivity, expressed explicitly with respect to ECM density)
    mi__dmn.set_biochem_param(72) = 0.006e-13; // m^2/(Pa*s) (interstitial hydraulic conductivity - max tension)
    mi__dmn.set_biochem_param(73) = 0.001e-13; // m^2/(Pa*s) (interstitial hydraulic conductivity - max compression)
    mi__dmn.set_biochem_param(74) = 0;         // -- (describe dependency of hydraulic conductivity in the stroma; 0= no dependency!)
    mi__dmn.set_biochem_param(300) = 0.300;    // -- (fibre volume fraction)
    mi__dmn.set_biochem_param(301) = 9.999e-9; // nm (fibre radius)
    mi__dmn.set_biochem_param(302) = 100.0;    // -- (diffusion factor coefficient)
    mi__dmn.set_biochem_param(303) = 0.;       // -- (Kozeny factor coefficient)
    //
    FieldInfo fld(4, 8, 6);
    fld.biochem[3] = 1.0;
    fld.strain = E;
    //
    VascularInfo vi;
    vi.status = VascularInfo::NORMAL;
    vi.radius = 5.00e-6;
    vi.thickness = 3.000e-6;
    vi.pore_radius = 90.00e-9;
    vi.pores_fraction = 50.00e-4;
    //
    mi__dmn.set_biochem_param(304) = 0;        // -- (describe dependency of drug diffusion in the stroma; 0= no dependency!)
    std::cout << stroma_diffusion_coefficient_2_drug_498(&mi__vnr, &mi__dmn, &fld, 10).comp << std::endl;
    std::cout << stroma_diffusion_coefficient_2_drug_498(&mi__vnr, &mi__dmn, &fld, 11).comp << std::endl;
    std::cout << stroma_diffusion_coefficient_2_drug_498(&mi__vnr, &mi__dmn, &fld, 12).comp << std::endl;
    std::cout << std::endl;
    std::cout << trace(stroma_diffusion_coefficient_2_drug_498(&mi__vnr, &mi__dmn, &fld, 10))/3.0 << std::endl;
    std::cout << trace(stroma_diffusion_coefficient_2_drug_498(&mi__vnr, &mi__dmn, &fld, 11))/3.0 << std::endl;
    std::cout << trace(stroma_diffusion_coefficient_2_drug_498(&mi__vnr, &mi__dmn, &fld, 12))/3.0 << std::endl;
    std::cout << std::endl;
    //
    mi__dmn.set_biochem_param(304) = 1;        // -- (describe dependency of drug diffusion in the stroma; 0= no dependency!)
    std::cout << stroma_diffusion_coefficient_2_drug_498(&mi__vnr, &mi__dmn, &fld, 10).comp << std::endl;
    std::cout << stroma_diffusion_coefficient_2_drug_498(&mi__vnr, &mi__dmn, &fld, 11).comp << std::endl;
    std::cout << stroma_diffusion_coefficient_2_drug_498(&mi__vnr, &mi__dmn, &fld, 12).comp << std::endl;
    std::cout << std::endl;
    std::cout << trace(stroma_diffusion_coefficient_2_drug_498(&mi__vnr, &mi__dmn, &fld, 10))/3.0 << std::endl;
    std::cout << trace(stroma_diffusion_coefficient_2_drug_498(&mi__vnr, &mi__dmn, &fld, 11))/3.0 << std::endl;
    std::cout << trace(stroma_diffusion_coefficient_2_drug_498(&mi__vnr, &mi__dmn, &fld, 12))/3.0 << std::endl;
    std::cout << std::endl;
    //
    std::cout << 1.0-vascular_reflection_coefficient_2_drug_498(&vi, &mi__vnr, &mi__dmn, &fld, 10) << std::endl;
    std::cout << 1.0-vascular_reflection_coefficient_2_drug_498(&vi, &mi__vnr, &mi__dmn, &fld, 11) << std::endl;
    std::cout << 1.0-vascular_reflection_coefficient_2_drug_498(&vi, &mi__vnr, &mi__dmn, &fld, 12) << std::endl;
    std::cout << std::endl;
    //
    /*
    Dyadic3D D;
    //
    D = stroma_hydraulic_conductivity_498(&mi__vnr, &mi__dmn, &fld);
    std::cout << magnitude(mul_im_m(D, unit(E_a[0]))) << " --- \n"
              << E_a[0].comp << D.comp << std::endl;
    //
    D = stroma_hydraulic_conductivity_498(&mi__vnr, &mi__dmn, &fld);
    std::cout << magnitude(mul_im_m(D, unit(E_a[1]))) << " --- \n"
              << E_a[1].comp << D.comp << std::endl;
    //
    D = stroma_hydraulic_conductivity_498(&mi__vnr, &mi__dmn, &fld);
    std::cout << magnitude(mul_im_m(D, unit(E_a[2]))) << " --- \n"
              << E_a[2].comp << D.comp << std::endl;
    //
    for (int i=0; i<5; i++) {
        D = stroma_hydraulic_conductivity_498(&mi__vnr, &mi__dmn, &fld);
        const Point p = Point( uniform_distro(-1.,+1.) ,
                               uniform_distro(-1.,+1.) ,
                               uniform_distro(-1.,+1.) ).unit();
        const Vector3D v = vec(p);
        std::cout << (D*p).norm() << " --- \n"
                  << v.comp << D.comp << std::endl;
        // const Point p_ = -p;
        // const Vector3D v_ = vec(p_);
        // std::cout << (D*p_).norm() << " --- \n"
        //           << v_.comp << D.comp << std::endl;
    }
    */
}
//=================================================================================================
