/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/utils.h"
#include "libmesh/getpot.h"
#include "libmesh/libmesh_logging.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
GlobalTolerance tol;
//=================================================================================================
// external routines
extern
void inspect_system (const System&, std::ofstream& );
extern
void inspect_all_systems (const EquationSystems&, std::ofstream& );
extern
void eigenproblem_test ();
// extern
// void IGA_test (const DomainData& dmn);
extern
void BDM_test ();
extern
void my_test_1 ();
extern
void my_test_2 ();
extern
void my_test_3 ();
extern
void my_test_4 ();
extern
void my_test_5 ();
extern
void my_test_6 ();
extern
void my_test_7 ();
extern
void my_test_8 ();
//=================================================================================================
void run_a_test (const char* arg) {
    std::cout << "\n *** Running a test with FEB3...\r" << std::flush;

    flog.open("run_a_test.log", true);
    ferr.open("run_a_test.err", true);
    if ( ! arg ) ERROR_("input argument is not properly initialised");
    else if ( 0 == strcmp(arg, "BDM") ) BDM_test();
    else if ( 0 == strcmp(arg, "1") ) my_test_1();
    else if ( 0 == strcmp(arg, "2") ) my_test_2();
    else if ( 0 == strcmp(arg, "3") ) my_test_3();
    else if ( 0 == strcmp(arg, "4") ) my_test_4();
    else if ( 0 == strcmp(arg, "5") ) my_test_5();
    else if ( 0 == strcmp(arg, "6") ) my_test_6();
    else if ( 0 == strcmp(arg, "7") ) my_test_7();
    else if ( 0 == strcmp(arg, "8") ) my_test_8();
    else if ( 0 == strcmp(arg, "eigen") ) eigenproblem_test();
    else ERROR_("exception caught for the kind of test to operate");

    std::cout << " *** Running a test with FEB3... [  OK  ]\n" << std::endl;
}
//=================================================================================================
