/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/nurbs.h"
//=================================================================================================
#define dz_dX    this->_inv_map_dzdX
#define d2z_dX2  this->_inv_map_d2zdX2
#define d3z_dX3  this->_inv_map_d3zdX3
#define dN_dz    this->_dphidz
#define d2N_dz2  this->_d2phidz2
#define d3N_dz3  this->_d3phidz3
//=================================================================================================
void NURBS::calc_d3phi (const Point& crd) const {
    if ( ! this->_calculate_d3phi ) return;
    // check for NURBS space
    if ( SP1D == this->dimension() ) {
        //
        for (int a=0; a<this->size(); a++) {
            this->_d3phi[a].comp(0,0,0) = d3z_dX3(0,0,0,0)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,0)*dz_dX(0,0)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(1,0,0) = d3z_dX3(0,1,0,0)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,0)*dz_dX(0,0)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(2,0,0) = d3z_dX3(0,2,0,0)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,0)*dz_dX(0,0)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(0,1,0) = d3z_dX3(0,0,1,0)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,1)*dz_dX(0,0)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(1,1,0) = d3z_dX3(0,1,1,0)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,1)*dz_dX(0,0)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(2,1,0) = d3z_dX3(0,2,1,0)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,1)*dz_dX(0,0)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(0,2,0) = d3z_dX3(0,0,2,0)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,2)*dz_dX(0,0)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(1,2,0) = d3z_dX3(0,1,2,0)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,2)*dz_dX(0,0)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(2,2,0) = d3z_dX3(0,2,2,0)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,2)*dz_dX(0,0)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(0,0,1) = d3z_dX3(0,0,0,1)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,0)*dz_dX(0,1)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(1,0,1) = d3z_dX3(0,1,0,1)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,0)*dz_dX(0,1)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(2,0,1) = d3z_dX3(0,2,0,1)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,0)*dz_dX(0,1)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(0,1,1) = d3z_dX3(0,0,1,1)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,1)*dz_dX(0,1)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(1,1,1) = d3z_dX3(0,1,1,1)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,1)*dz_dX(0,1)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(2,1,1) = d3z_dX3(0,2,1,1)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,1)*dz_dX(0,1)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(0,2,1) = d3z_dX3(0,0,2,1)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,2)*dz_dX(0,1)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(1,2,1) = d3z_dX3(0,1,2,1)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,2)*dz_dX(0,1)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(2,2,1) = d3z_dX3(0,2,2,1)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,2)*dz_dX(0,1)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(0,0,2) = d3z_dX3(0,0,0,2)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,0)*dz_dX(0,2)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(1,0,2) = d3z_dX3(0,1,0,2)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,0)*dz_dX(0,2)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(2,0,2) = d3z_dX3(0,2,0,2)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,0)*dz_dX(0,2)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(0,1,2) = d3z_dX3(0,0,1,2)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,1)*dz_dX(0,2)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(1,1,2) = d3z_dX3(0,1,1,2)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,1)*dz_dX(0,2)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(2,1,2) = d3z_dX3(0,2,1,2)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,1)*dz_dX(0,2)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(0,2,2) = d3z_dX3(0,0,2,2)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,2)*dz_dX(0,2)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(1,2,2) = d3z_dX3(0,1,2,2)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,2)*dz_dX(0,2)*d3N_dz3[0][0][0][a];
            this->_d3phi[a].comp(2,2,2) = d3z_dX3(0,2,2,2)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,2)*dz_dX(0,2)*d3N_dz3[0][0][0][a];
        }
        //
    } else if ( SP2D == this->dimension() ) {
        //
        for (int a=0; a<this->size(); a++) {
            this->_d3phi[a].comp(0,0,0) = d3z_dX3(0,0,0,0)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,0)*dz_dX(0,0)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,0)*dz_dX(1,0)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,0,0,0)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,0,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,0,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,0,0)*d2N_dz2[0][1][a]
                                        + dz_dX(0,0)*dz_dX(1,0)*dz_dX(0,0)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,0)*dz_dX(1,0)*dz_dX(1,0)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,0,0,0)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,0,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,0,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,0,0)*d2N_dz2[1][0][a]
                                        + dz_dX(1,0)*dz_dX(0,0)*dz_dX(0,0)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,0)*dz_dX(0,0)*dz_dX(1,0)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,0,0,0)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,0,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,0,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,0,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*dz_dX(1,0)*dz_dX(0,0)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,0)*dz_dX(1,0)*dz_dX(1,0)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(1,0,0) = d3z_dX3(0,1,0,0)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,0)*dz_dX(0,0)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,0)*dz_dX(1,0)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,1,0,0)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,0,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,1,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,1,0)*d2N_dz2[0][1][a]
                                        + dz_dX(0,1)*dz_dX(1,0)*dz_dX(0,0)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,1)*dz_dX(1,0)*dz_dX(1,0)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,1,0,0)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,0,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,1,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,1,0)*d2N_dz2[1][0][a]
                                        + dz_dX(1,1)*dz_dX(0,0)*dz_dX(0,0)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,1)*dz_dX(0,0)*dz_dX(1,0)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,1,0,0)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,0,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,1,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,1,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*dz_dX(1,0)*dz_dX(0,0)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,1)*dz_dX(1,0)*dz_dX(1,0)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(2,0,0) = d3z_dX3(0,2,0,0)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,0)*dz_dX(0,0)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,0)*dz_dX(1,0)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,2,0,0)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,0,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,2,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,2,0)*d2N_dz2[0][1][a]
                                        + dz_dX(0,2)*dz_dX(1,0)*dz_dX(0,0)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,2)*dz_dX(1,0)*dz_dX(1,0)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,2,0,0)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,0,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,2,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,2,0)*d2N_dz2[1][0][a]
                                        + dz_dX(1,2)*dz_dX(0,0)*dz_dX(0,0)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,2)*dz_dX(0,0)*dz_dX(1,0)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,2,0,0)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,0,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,2,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,2,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*dz_dX(1,0)*dz_dX(0,0)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,2)*dz_dX(1,0)*dz_dX(1,0)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(0,1,0) = d3z_dX3(0,0,1,0)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,1)*dz_dX(0,0)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,1)*dz_dX(1,0)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,0,1,0)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,1,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,0,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,0,1)*d2N_dz2[0][1][a]
                                        + dz_dX(0,0)*dz_dX(1,1)*dz_dX(0,0)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,0)*dz_dX(1,1)*dz_dX(1,0)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,0,1,0)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,1,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,0,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,0,1)*d2N_dz2[1][0][a]
                                        + dz_dX(1,0)*dz_dX(0,1)*dz_dX(0,0)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,0)*dz_dX(0,1)*dz_dX(1,0)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,0,1,0)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,1,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,0,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,0,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*dz_dX(1,1)*dz_dX(0,0)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,0)*dz_dX(1,1)*dz_dX(1,0)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(1,1,0) = d3z_dX3(0,1,1,0)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,1)*dz_dX(0,0)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,1)*dz_dX(1,0)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,1,1,0)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,1,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,1,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,1,1)*d2N_dz2[0][1][a]
                                        + dz_dX(0,1)*dz_dX(1,1)*dz_dX(0,0)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,1)*dz_dX(1,1)*dz_dX(1,0)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,1,1,0)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,1,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,1,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,1,1)*d2N_dz2[1][0][a]
                                        + dz_dX(1,1)*dz_dX(0,1)*dz_dX(0,0)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,1)*dz_dX(0,1)*dz_dX(1,0)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,1,1,0)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,1,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,1,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,1,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*dz_dX(1,1)*dz_dX(0,0)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,1)*dz_dX(1,1)*dz_dX(1,0)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(2,1,0) = d3z_dX3(0,2,1,0)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,1)*dz_dX(0,0)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,1)*dz_dX(1,0)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,2,1,0)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,1,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,2,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,2,1)*d2N_dz2[0][1][a]
                                        + dz_dX(0,2)*dz_dX(1,1)*dz_dX(0,0)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,2)*dz_dX(1,1)*dz_dX(1,0)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,2,1,0)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,1,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,2,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,2,1)*d2N_dz2[1][0][a]
                                        + dz_dX(1,2)*dz_dX(0,1)*dz_dX(0,0)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,2)*dz_dX(0,1)*dz_dX(1,0)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,2,1,0)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,1,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,2,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,2,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*dz_dX(1,1)*dz_dX(0,0)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,2)*dz_dX(1,1)*dz_dX(1,0)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(0,2,0) = d3z_dX3(0,0,2,0)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,2)*dz_dX(0,0)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,2)*dz_dX(1,0)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,0,2,0)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,2,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,0,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,0,2)*d2N_dz2[0][1][a]
                                        + dz_dX(0,0)*dz_dX(1,2)*dz_dX(0,0)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,0)*dz_dX(1,2)*dz_dX(1,0)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,0,2,0)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,2,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,0,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,0,2)*d2N_dz2[1][0][a]
                                        + dz_dX(1,0)*dz_dX(0,2)*dz_dX(0,0)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,0)*dz_dX(0,2)*dz_dX(1,0)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,0,2,0)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,2,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,0,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,0,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*dz_dX(1,2)*dz_dX(0,0)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,0)*dz_dX(1,2)*dz_dX(1,0)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(1,2,0) = d3z_dX3(0,1,2,0)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,2)*dz_dX(0,0)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,2)*dz_dX(1,0)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,1,2,0)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,2,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,1,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,1,2)*d2N_dz2[0][1][a]
                                        + dz_dX(0,1)*dz_dX(1,2)*dz_dX(0,0)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,1)*dz_dX(1,2)*dz_dX(1,0)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,1,2,0)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,2,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,1,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,1,2)*d2N_dz2[1][0][a]
                                        + dz_dX(1,1)*dz_dX(0,2)*dz_dX(0,0)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,1)*dz_dX(0,2)*dz_dX(1,0)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,1,2,0)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,2,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,1,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,1,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*dz_dX(1,2)*dz_dX(0,0)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,1)*dz_dX(1,2)*dz_dX(1,0)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(2,2,0) = d3z_dX3(0,2,2,0)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,2)*dz_dX(0,0)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,2)*dz_dX(1,0)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,2,2,0)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,2,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,2,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,2,2)*d2N_dz2[0][1][a]
                                        + dz_dX(0,2)*dz_dX(1,2)*dz_dX(0,0)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,2)*dz_dX(1,2)*dz_dX(1,0)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,2,2,0)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,2,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,2,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,2,2)*d2N_dz2[1][0][a]
                                        + dz_dX(1,2)*dz_dX(0,2)*dz_dX(0,0)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,2)*dz_dX(0,2)*dz_dX(1,0)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,2,2,0)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,2,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,2,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,2,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*dz_dX(1,2)*dz_dX(0,0)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,2)*dz_dX(1,2)*dz_dX(1,0)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(0,0,1) = d3z_dX3(0,0,0,1)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,0)*dz_dX(0,1)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,0)*dz_dX(1,1)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,0,0,1)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,0,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,0,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,0,0)*d2N_dz2[0][1][a]
                                        + dz_dX(0,0)*dz_dX(1,0)*dz_dX(0,1)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,0)*dz_dX(1,0)*dz_dX(1,1)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,0,0,1)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,0,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,0,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,0,0)*d2N_dz2[1][0][a]
                                        + dz_dX(1,0)*dz_dX(0,0)*dz_dX(0,1)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,0)*dz_dX(0,0)*dz_dX(1,1)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,0,0,1)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,0,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,0,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,0,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*dz_dX(1,0)*dz_dX(0,1)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,0)*dz_dX(1,0)*dz_dX(1,1)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(1,0,1) = d3z_dX3(0,1,0,1)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,0)*dz_dX(0,1)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,0)*dz_dX(1,1)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,1,0,1)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,0,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,1,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,1,0)*d2N_dz2[0][1][a]
                                        + dz_dX(0,1)*dz_dX(1,0)*dz_dX(0,1)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,1)*dz_dX(1,0)*dz_dX(1,1)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,1,0,1)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,0,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,1,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,1,0)*d2N_dz2[1][0][a]
                                        + dz_dX(1,1)*dz_dX(0,0)*dz_dX(0,1)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,1)*dz_dX(0,0)*dz_dX(1,1)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,1,0,1)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,0,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,1,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,1,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*dz_dX(1,0)*dz_dX(0,1)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,1)*dz_dX(1,0)*dz_dX(1,1)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(2,0,1) = d3z_dX3(0,2,0,1)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,0)*dz_dX(0,1)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,0)*dz_dX(1,1)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,2,0,1)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,0,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,2,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,2,0)*d2N_dz2[0][1][a]
                                        + dz_dX(0,2)*dz_dX(1,0)*dz_dX(0,1)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,2)*dz_dX(1,0)*dz_dX(1,1)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,2,0,1)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,0,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,2,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,2,0)*d2N_dz2[1][0][a]
                                        + dz_dX(1,2)*dz_dX(0,0)*dz_dX(0,1)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,2)*dz_dX(0,0)*dz_dX(1,1)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,2,0,1)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,0,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,2,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,2,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*dz_dX(1,0)*dz_dX(0,1)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,2)*dz_dX(1,0)*dz_dX(1,1)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(0,1,1) = d3z_dX3(0,0,1,1)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,1)*dz_dX(0,1)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,1)*dz_dX(1,1)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,0,1,1)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,1,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,0,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,0,1)*d2N_dz2[0][1][a]
                                        + dz_dX(0,0)*dz_dX(1,1)*dz_dX(0,1)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,0)*dz_dX(1,1)*dz_dX(1,1)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,0,1,1)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,1,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,0,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,0,1)*d2N_dz2[1][0][a]
                                        + dz_dX(1,0)*dz_dX(0,1)*dz_dX(0,1)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,0)*dz_dX(0,1)*dz_dX(1,1)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,0,1,1)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,1,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,0,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,0,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*dz_dX(1,1)*dz_dX(0,1)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,0)*dz_dX(1,1)*dz_dX(1,1)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(1,1,1) = d3z_dX3(0,1,1,1)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,1)*dz_dX(0,1)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,1)*dz_dX(1,1)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,1,1,1)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,1,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,1,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,1,1)*d2N_dz2[0][1][a]
                                        + dz_dX(0,1)*dz_dX(1,1)*dz_dX(0,1)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,1)*dz_dX(1,1)*dz_dX(1,1)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,1,1,1)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,1,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,1,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,1,1)*d2N_dz2[1][0][a]
                                        + dz_dX(1,1)*dz_dX(0,1)*dz_dX(0,1)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,1)*dz_dX(0,1)*dz_dX(1,1)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,1,1,1)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,1,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,1,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,1,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*dz_dX(1,1)*dz_dX(0,1)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,1)*dz_dX(1,1)*dz_dX(1,1)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(2,1,1) = d3z_dX3(0,2,1,1)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,1)*dz_dX(0,1)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,1)*dz_dX(1,1)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,2,1,1)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,1,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,2,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,2,1)*d2N_dz2[0][1][a]
                                        + dz_dX(0,2)*dz_dX(1,1)*dz_dX(0,1)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,2)*dz_dX(1,1)*dz_dX(1,1)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,2,1,1)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,1,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,2,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,2,1)*d2N_dz2[1][0][a]
                                        + dz_dX(1,2)*dz_dX(0,1)*dz_dX(0,1)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,2)*dz_dX(0,1)*dz_dX(1,1)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,2,1,1)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,1,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,2,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,2,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*dz_dX(1,1)*dz_dX(0,1)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,2)*dz_dX(1,1)*dz_dX(1,1)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(0,2,1) = d3z_dX3(0,0,2,1)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,2)*dz_dX(0,1)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,2)*dz_dX(1,1)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,0,2,1)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,2,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,0,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,0,2)*d2N_dz2[0][1][a]
                                        + dz_dX(0,0)*dz_dX(1,2)*dz_dX(0,1)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,0)*dz_dX(1,2)*dz_dX(1,1)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,0,2,1)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,2,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,0,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,0,2)*d2N_dz2[1][0][a]
                                        + dz_dX(1,0)*dz_dX(0,2)*dz_dX(0,1)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,0)*dz_dX(0,2)*dz_dX(1,1)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,0,2,1)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,2,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,0,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,0,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*dz_dX(1,2)*dz_dX(0,1)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,0)*dz_dX(1,2)*dz_dX(1,1)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(1,2,1) = d3z_dX3(0,1,2,1)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,2)*dz_dX(0,1)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,2)*dz_dX(1,1)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,1,2,1)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,2,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,1,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,1,2)*d2N_dz2[0][1][a]
                                        + dz_dX(0,1)*dz_dX(1,2)*dz_dX(0,1)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,1)*dz_dX(1,2)*dz_dX(1,1)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,1,2,1)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,2,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,1,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,1,2)*d2N_dz2[1][0][a]
                                        + dz_dX(1,1)*dz_dX(0,2)*dz_dX(0,1)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,1)*dz_dX(0,2)*dz_dX(1,1)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,1,2,1)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,2,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,1,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,1,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*dz_dX(1,2)*dz_dX(0,1)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,1)*dz_dX(1,2)*dz_dX(1,1)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(2,2,1) = d3z_dX3(0,2,2,1)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,2)*dz_dX(0,1)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,2)*dz_dX(1,1)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,2,2,1)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,2,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,2,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,2,2)*d2N_dz2[0][1][a]
                                        + dz_dX(0,2)*dz_dX(1,2)*dz_dX(0,1)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,2)*dz_dX(1,2)*dz_dX(1,1)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,2,2,1)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,2,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,2,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,2,2)*d2N_dz2[1][0][a]
                                        + dz_dX(1,2)*dz_dX(0,2)*dz_dX(0,1)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,2)*dz_dX(0,2)*dz_dX(1,1)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,2,2,1)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,2,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,2,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,2,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*dz_dX(1,2)*dz_dX(0,1)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,2)*dz_dX(1,2)*dz_dX(1,1)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(0,0,2) = d3z_dX3(0,0,0,2)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,0)*dz_dX(0,2)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,0)*dz_dX(1,2)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,0,0,2)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,0,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,0,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,0,0)*d2N_dz2[0][1][a]
                                        + dz_dX(0,0)*dz_dX(1,0)*dz_dX(0,2)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,0)*dz_dX(1,0)*dz_dX(1,2)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,0,0,2)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,0,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,0,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,0,0)*d2N_dz2[1][0][a]
                                        + dz_dX(1,0)*dz_dX(0,0)*dz_dX(0,2)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,0)*dz_dX(0,0)*dz_dX(1,2)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,0,0,2)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,0,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,0,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,0,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*dz_dX(1,0)*dz_dX(0,2)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,0)*dz_dX(1,0)*dz_dX(1,2)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(1,0,2) = d3z_dX3(0,1,0,2)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,0)*dz_dX(0,2)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,0)*dz_dX(1,2)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,1,0,2)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,0,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,1,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,1,0)*d2N_dz2[0][1][a]
                                        + dz_dX(0,1)*dz_dX(1,0)*dz_dX(0,2)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,1)*dz_dX(1,0)*dz_dX(1,2)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,1,0,2)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,0,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,1,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,1,0)*d2N_dz2[1][0][a]
                                        + dz_dX(1,1)*dz_dX(0,0)*dz_dX(0,2)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,1)*dz_dX(0,0)*dz_dX(1,2)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,1,0,2)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,0,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,1,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,1,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*dz_dX(1,0)*dz_dX(0,2)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,1)*dz_dX(1,0)*dz_dX(1,2)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(2,0,2) = d3z_dX3(0,2,0,2)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,0)*dz_dX(0,2)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,0)*dz_dX(1,2)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,2,0,2)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,0,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,2,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,2,0)*d2N_dz2[0][1][a]
                                        + dz_dX(0,2)*dz_dX(1,0)*dz_dX(0,2)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,2)*dz_dX(1,0)*dz_dX(1,2)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,2,0,2)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,0,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,2,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,2,0)*d2N_dz2[1][0][a]
                                        + dz_dX(1,2)*dz_dX(0,0)*dz_dX(0,2)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,2)*dz_dX(0,0)*dz_dX(1,2)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,2,0,2)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,0,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,2,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,2,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*dz_dX(1,0)*dz_dX(0,2)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,2)*dz_dX(1,0)*dz_dX(1,2)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(0,1,2) = d3z_dX3(0,0,1,2)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,1)*dz_dX(0,2)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,1)*dz_dX(1,2)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,0,1,2)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,1,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,0,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,0,1)*d2N_dz2[0][1][a]
                                        + dz_dX(0,0)*dz_dX(1,1)*dz_dX(0,2)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,0)*dz_dX(1,1)*dz_dX(1,2)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,0,1,2)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,1,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,0,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,0,1)*d2N_dz2[1][0][a]
                                        + dz_dX(1,0)*dz_dX(0,1)*dz_dX(0,2)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,0)*dz_dX(0,1)*dz_dX(1,2)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,0,1,2)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,1,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,0,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,0,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*dz_dX(1,1)*dz_dX(0,2)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,0)*dz_dX(1,1)*dz_dX(1,2)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(1,1,2) = d3z_dX3(0,1,1,2)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,1)*dz_dX(0,2)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,1)*dz_dX(1,2)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,1,1,2)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,1,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,1,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,1,1)*d2N_dz2[0][1][a]
                                        + dz_dX(0,1)*dz_dX(1,1)*dz_dX(0,2)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,1)*dz_dX(1,1)*dz_dX(1,2)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,1,1,2)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,1,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,1,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,1,1)*d2N_dz2[1][0][a]
                                        + dz_dX(1,1)*dz_dX(0,1)*dz_dX(0,2)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,1)*dz_dX(0,1)*dz_dX(1,2)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,1,1,2)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,1,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,1,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,1,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*dz_dX(1,1)*dz_dX(0,2)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,1)*dz_dX(1,1)*dz_dX(1,2)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(2,1,2) = d3z_dX3(0,2,1,2)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,1)*dz_dX(0,2)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,1)*dz_dX(1,2)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,2,1,2)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,1,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,2,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,2,1)*d2N_dz2[0][1][a]
                                        + dz_dX(0,2)*dz_dX(1,1)*dz_dX(0,2)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,2)*dz_dX(1,1)*dz_dX(1,2)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,2,1,2)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,1,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,2,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,2,1)*d2N_dz2[1][0][a]
                                        + dz_dX(1,2)*dz_dX(0,1)*dz_dX(0,2)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,2)*dz_dX(0,1)*dz_dX(1,2)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,2,1,2)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,1,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,2,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,2,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*dz_dX(1,1)*dz_dX(0,2)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,2)*dz_dX(1,1)*dz_dX(1,2)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(0,2,2) = d3z_dX3(0,0,2,2)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,2)*dz_dX(0,2)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,2)*dz_dX(1,2)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,0,2,2)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,2,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,0,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,0,2)*d2N_dz2[0][1][a]
                                        + dz_dX(0,0)*dz_dX(1,2)*dz_dX(0,2)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,0)*dz_dX(1,2)*dz_dX(1,2)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,0,2,2)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,2,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,0,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,0,2)*d2N_dz2[1][0][a]
                                        + dz_dX(1,0)*dz_dX(0,2)*dz_dX(0,2)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,0)*dz_dX(0,2)*dz_dX(1,2)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,0,2,2)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,2,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,0,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,0,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*dz_dX(1,2)*dz_dX(0,2)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,0)*dz_dX(1,2)*dz_dX(1,2)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(1,2,2) = d3z_dX3(0,1,2,2)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,2)*dz_dX(0,2)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,2)*dz_dX(1,2)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,1,2,2)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,2,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,1,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,1,2)*d2N_dz2[0][1][a]
                                        + dz_dX(0,1)*dz_dX(1,2)*dz_dX(0,2)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,1)*dz_dX(1,2)*dz_dX(1,2)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,1,2,2)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,2,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,1,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,1,2)*d2N_dz2[1][0][a]
                                        + dz_dX(1,1)*dz_dX(0,2)*dz_dX(0,2)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,1)*dz_dX(0,2)*dz_dX(1,2)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,1,2,2)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,2,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,1,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,1,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*dz_dX(1,2)*dz_dX(0,2)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,1)*dz_dX(1,2)*dz_dX(1,2)*d3N_dz3[1][1][1][a];
            this->_d3phi[a].comp(2,2,2) = d3z_dX3(0,2,2,2)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,2)*dz_dX(0,2)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,2)*dz_dX(1,2)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,2,2,2)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,2,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,2,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,2,2)*d2N_dz2[0][1][a]
                                        + dz_dX(0,2)*dz_dX(1,2)*dz_dX(0,2)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,2)*dz_dX(1,2)*dz_dX(1,2)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(1,2,2,2)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,2,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,2,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,2,2)*d2N_dz2[1][0][a]
                                        + dz_dX(1,2)*dz_dX(0,2)*dz_dX(0,2)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,2)*dz_dX(0,2)*dz_dX(1,2)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,2,2,2)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,2,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,2,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,2,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*dz_dX(1,2)*dz_dX(0,2)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,2)*dz_dX(1,2)*dz_dX(1,2)*d3N_dz3[1][1][1][a];
        }
        //
    } else if ( SP3D == this->dimension() ) {
        //
        for (int a=0; a<this->size(); a++) {
            this->_d3phi[a].comp(0,0,0) = d3z_dX3(0,0,0,0)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,0)*dz_dX(0,0)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,0)*dz_dX(1,0)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,0,0,0)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,0,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,0,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,0,0)*d2N_dz2[0][1][a]
                                        + dz_dX(0,0)*dz_dX(1,0)*dz_dX(0,0)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,0)*dz_dX(1,0)*dz_dX(1,0)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,0,0,0)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,0,0)*d2N_dz2[0][2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,0,0)*d2N_dz2[0][2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,0,0)*d2N_dz2[0][2][a]
                                        + dz_dX(0,0)*dz_dX(2,0)*dz_dX(0,0)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,0)*dz_dX(2,0)*dz_dX(1,0)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,0,0,0)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,0,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,0,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,0,0)*d2N_dz2[1][0][a]
                                        + dz_dX(1,0)*dz_dX(0,0)*dz_dX(0,0)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,0)*dz_dX(0,0)*dz_dX(1,0)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,0,0,0)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,0,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,0,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,0,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*dz_dX(1,0)*dz_dX(0,0)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,0)*dz_dX(1,0)*dz_dX(1,0)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,0,0,0)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,0,0)*d2N_dz2[1][2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,0,0)*d2N_dz2[1][2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,0,0)*d2N_dz2[1][2][a]
                                        + dz_dX(1,0)*dz_dX(2,0)*dz_dX(0,0)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,0)*dz_dX(2,0)*dz_dX(1,0)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,0,0,0)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,0,0)*d2N_dz2[2][0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,0,0)*d2N_dz2[2][0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,0,0)*d2N_dz2[2][0][a]
                                        + dz_dX(2,0)*dz_dX(0,0)*dz_dX(0,0)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,0)*dz_dX(0,0)*dz_dX(1,0)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,0,0,0)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,0,0)*d2N_dz2[2][1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,0,0)*d2N_dz2[2][1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,0,0)*d2N_dz2[2][1][a]
                                        + dz_dX(2,0)*dz_dX(1,0)*dz_dX(0,0)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,0)*dz_dX(1,0)*dz_dX(1,0)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,0,0,0)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,0,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,0,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,0,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*dz_dX(2,0)*dz_dX(0,0)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,0)*dz_dX(2,0)*dz_dX(1,0)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(1,0,0) = d3z_dX3(0,1,0,0)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,0)*dz_dX(0,0)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,0)*dz_dX(1,0)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,1,0,0)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,0,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,1,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,1,0)*d2N_dz2[0][1][a]
                                        + dz_dX(0,1)*dz_dX(1,0)*dz_dX(0,0)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,1)*dz_dX(1,0)*dz_dX(1,0)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,1,0,0)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,0,0)*d2N_dz2[0][2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,1,0)*d2N_dz2[0][2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,1,0)*d2N_dz2[0][2][a]
                                        + dz_dX(0,1)*dz_dX(2,0)*dz_dX(0,0)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,1)*dz_dX(2,0)*dz_dX(1,0)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,1,0,0)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,0,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,1,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,1,0)*d2N_dz2[1][0][a]
                                        + dz_dX(1,1)*dz_dX(0,0)*dz_dX(0,0)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,1)*dz_dX(0,0)*dz_dX(1,0)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,1,0,0)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,0,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,1,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,1,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*dz_dX(1,0)*dz_dX(0,0)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,1)*dz_dX(1,0)*dz_dX(1,0)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,1,0,0)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,0,0)*d2N_dz2[1][2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,1,0)*d2N_dz2[1][2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,1,0)*d2N_dz2[1][2][a]
                                        + dz_dX(1,1)*dz_dX(2,0)*dz_dX(0,0)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,1)*dz_dX(2,0)*dz_dX(1,0)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,1,0,0)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,0,0)*d2N_dz2[2][0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,1,0)*d2N_dz2[2][0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,1,0)*d2N_dz2[2][0][a]
                                        + dz_dX(2,1)*dz_dX(0,0)*dz_dX(0,0)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,1)*dz_dX(0,0)*dz_dX(1,0)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,1,0,0)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,0,0)*d2N_dz2[2][1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,1,0)*d2N_dz2[2][1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,1,0)*d2N_dz2[2][1][a]
                                        + dz_dX(2,1)*dz_dX(1,0)*dz_dX(0,0)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,1)*dz_dX(1,0)*dz_dX(1,0)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,1,0,0)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,0,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,1,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,1,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*dz_dX(2,0)*dz_dX(0,0)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,1)*dz_dX(2,0)*dz_dX(1,0)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(2,0,0) = d3z_dX3(0,2,0,0)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,0)*dz_dX(0,0)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,0)*dz_dX(1,0)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,2,0,0)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,0,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,2,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,2,0)*d2N_dz2[0][1][a]
                                        + dz_dX(0,2)*dz_dX(1,0)*dz_dX(0,0)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,2)*dz_dX(1,0)*dz_dX(1,0)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,2,0,0)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,0,0)*d2N_dz2[0][2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,2,0)*d2N_dz2[0][2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,2,0)*d2N_dz2[0][2][a]
                                        + dz_dX(0,2)*dz_dX(2,0)*dz_dX(0,0)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,2)*dz_dX(2,0)*dz_dX(1,0)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,2,0,0)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,0,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,2,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,2,0)*d2N_dz2[1][0][a]
                                        + dz_dX(1,2)*dz_dX(0,0)*dz_dX(0,0)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,2)*dz_dX(0,0)*dz_dX(1,0)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,2,0,0)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,0,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,2,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,2,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*dz_dX(1,0)*dz_dX(0,0)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,2)*dz_dX(1,0)*dz_dX(1,0)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,2,0,0)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,0,0)*d2N_dz2[1][2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,2,0)*d2N_dz2[1][2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,2,0)*d2N_dz2[1][2][a]
                                        + dz_dX(1,2)*dz_dX(2,0)*dz_dX(0,0)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,2)*dz_dX(2,0)*dz_dX(1,0)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,2,0,0)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,0,0)*d2N_dz2[2][0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,2,0)*d2N_dz2[2][0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,2,0)*d2N_dz2[2][0][a]
                                        + dz_dX(2,2)*dz_dX(0,0)*dz_dX(0,0)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,2)*dz_dX(0,0)*dz_dX(1,0)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,2,0,0)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,0,0)*d2N_dz2[2][1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,2,0)*d2N_dz2[2][1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,2,0)*d2N_dz2[2][1][a]
                                        + dz_dX(2,2)*dz_dX(1,0)*dz_dX(0,0)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,2)*dz_dX(1,0)*dz_dX(1,0)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,2,0,0)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,0,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,2,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,2,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*dz_dX(2,0)*dz_dX(0,0)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,2)*dz_dX(2,0)*dz_dX(1,0)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(0,1,0) = d3z_dX3(0,0,1,0)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,1)*dz_dX(0,0)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,1)*dz_dX(1,0)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,0,1,0)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,1,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,0,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,0,1)*d2N_dz2[0][1][a]
                                        + dz_dX(0,0)*dz_dX(1,1)*dz_dX(0,0)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,0)*dz_dX(1,1)*dz_dX(1,0)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,0,1,0)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,1,0)*d2N_dz2[0][2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,0,0)*d2N_dz2[0][2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,0,1)*d2N_dz2[0][2][a]
                                        + dz_dX(0,0)*dz_dX(2,1)*dz_dX(0,0)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,0)*dz_dX(2,1)*dz_dX(1,0)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,0,1,0)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,1,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,0,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,0,1)*d2N_dz2[1][0][a]
                                        + dz_dX(1,0)*dz_dX(0,1)*dz_dX(0,0)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,0)*dz_dX(0,1)*dz_dX(1,0)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,0,1,0)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,1,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,0,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,0,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*dz_dX(1,1)*dz_dX(0,0)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,0)*dz_dX(1,1)*dz_dX(1,0)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,0,1,0)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,1,0)*d2N_dz2[1][2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,0,0)*d2N_dz2[1][2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,0,1)*d2N_dz2[1][2][a]
                                        + dz_dX(1,0)*dz_dX(2,1)*dz_dX(0,0)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,0)*dz_dX(2,1)*dz_dX(1,0)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,0,1,0)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,1,0)*d2N_dz2[2][0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,0,0)*d2N_dz2[2][0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,0,1)*d2N_dz2[2][0][a]
                                        + dz_dX(2,0)*dz_dX(0,1)*dz_dX(0,0)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,0)*dz_dX(0,1)*dz_dX(1,0)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,0,1,0)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,1,0)*d2N_dz2[2][1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,0,0)*d2N_dz2[2][1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,0,1)*d2N_dz2[2][1][a]
                                        + dz_dX(2,0)*dz_dX(1,1)*dz_dX(0,0)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,0)*dz_dX(1,1)*dz_dX(1,0)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,0,1,0)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,1,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,0,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,0,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*dz_dX(2,1)*dz_dX(0,0)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,0)*dz_dX(2,1)*dz_dX(1,0)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(1,1,0) = d3z_dX3(0,1,1,0)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,1)*dz_dX(0,0)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,1)*dz_dX(1,0)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,1,1,0)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,1,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,1,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,1,1)*d2N_dz2[0][1][a]
                                        + dz_dX(0,1)*dz_dX(1,1)*dz_dX(0,0)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,1)*dz_dX(1,1)*dz_dX(1,0)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,1,1,0)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,1,0)*d2N_dz2[0][2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,1,0)*d2N_dz2[0][2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,1,1)*d2N_dz2[0][2][a]
                                        + dz_dX(0,1)*dz_dX(2,1)*dz_dX(0,0)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,1)*dz_dX(2,1)*dz_dX(1,0)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,1,1,0)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,1,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,1,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,1,1)*d2N_dz2[1][0][a]
                                        + dz_dX(1,1)*dz_dX(0,1)*dz_dX(0,0)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,1)*dz_dX(0,1)*dz_dX(1,0)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,1,1,0)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,1,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,1,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,1,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*dz_dX(1,1)*dz_dX(0,0)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,1)*dz_dX(1,1)*dz_dX(1,0)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,1,1,0)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,1,0)*d2N_dz2[1][2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,1,0)*d2N_dz2[1][2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,1,1)*d2N_dz2[1][2][a]
                                        + dz_dX(1,1)*dz_dX(2,1)*dz_dX(0,0)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,1)*dz_dX(2,1)*dz_dX(1,0)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,1,1,0)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,1,0)*d2N_dz2[2][0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,1,0)*d2N_dz2[2][0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,1,1)*d2N_dz2[2][0][a]
                                        + dz_dX(2,1)*dz_dX(0,1)*dz_dX(0,0)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,1)*dz_dX(0,1)*dz_dX(1,0)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,1,1,0)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,1,0)*d2N_dz2[2][1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,1,0)*d2N_dz2[2][1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,1,1)*d2N_dz2[2][1][a]
                                        + dz_dX(2,1)*dz_dX(1,1)*dz_dX(0,0)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,1)*dz_dX(1,1)*dz_dX(1,0)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,1,1,0)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,1,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,1,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,1,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*dz_dX(2,1)*dz_dX(0,0)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,1)*dz_dX(2,1)*dz_dX(1,0)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(2,1,0) = d3z_dX3(0,2,1,0)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,1)*dz_dX(0,0)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,1)*dz_dX(1,0)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,2,1,0)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,1,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,2,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,2,1)*d2N_dz2[0][1][a]
                                        + dz_dX(0,2)*dz_dX(1,1)*dz_dX(0,0)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,2)*dz_dX(1,1)*dz_dX(1,0)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,2,1,0)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,1,0)*d2N_dz2[0][2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,2,0)*d2N_dz2[0][2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,2,1)*d2N_dz2[0][2][a]
                                        + dz_dX(0,2)*dz_dX(2,1)*dz_dX(0,0)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,2)*dz_dX(2,1)*dz_dX(1,0)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,2,1,0)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,1,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,2,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,2,1)*d2N_dz2[1][0][a]
                                        + dz_dX(1,2)*dz_dX(0,1)*dz_dX(0,0)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,2)*dz_dX(0,1)*dz_dX(1,0)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,2,1,0)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,1,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,2,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,2,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*dz_dX(1,1)*dz_dX(0,0)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,2)*dz_dX(1,1)*dz_dX(1,0)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,2,1,0)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,1,0)*d2N_dz2[1][2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,2,0)*d2N_dz2[1][2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,2,1)*d2N_dz2[1][2][a]
                                        + dz_dX(1,2)*dz_dX(2,1)*dz_dX(0,0)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,2)*dz_dX(2,1)*dz_dX(1,0)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,2,1,0)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,1,0)*d2N_dz2[2][0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,2,0)*d2N_dz2[2][0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,2,1)*d2N_dz2[2][0][a]
                                        + dz_dX(2,2)*dz_dX(0,1)*dz_dX(0,0)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,2)*dz_dX(0,1)*dz_dX(1,0)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,2,1,0)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,1,0)*d2N_dz2[2][1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,2,0)*d2N_dz2[2][1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,2,1)*d2N_dz2[2][1][a]
                                        + dz_dX(2,2)*dz_dX(1,1)*dz_dX(0,0)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,2)*dz_dX(1,1)*dz_dX(1,0)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,2,1,0)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,1,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,2,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,2,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*dz_dX(2,1)*dz_dX(0,0)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,2)*dz_dX(2,1)*dz_dX(1,0)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(0,2,0) = d3z_dX3(0,0,2,0)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,2)*dz_dX(0,0)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,2)*dz_dX(1,0)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,0,2,0)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,2,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,0,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,0,2)*d2N_dz2[0][1][a]
                                        + dz_dX(0,0)*dz_dX(1,2)*dz_dX(0,0)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,0)*dz_dX(1,2)*dz_dX(1,0)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,0,2,0)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,2,0)*d2N_dz2[0][2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,0,0)*d2N_dz2[0][2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,0,2)*d2N_dz2[0][2][a]
                                        + dz_dX(0,0)*dz_dX(2,2)*dz_dX(0,0)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,0)*dz_dX(2,2)*dz_dX(1,0)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,0,2,0)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,2,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,0,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,0,2)*d2N_dz2[1][0][a]
                                        + dz_dX(1,0)*dz_dX(0,2)*dz_dX(0,0)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,0)*dz_dX(0,2)*dz_dX(1,0)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,0,2,0)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,2,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,0,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,0,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*dz_dX(1,2)*dz_dX(0,0)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,0)*dz_dX(1,2)*dz_dX(1,0)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,0,2,0)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,2,0)*d2N_dz2[1][2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,0,0)*d2N_dz2[1][2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,0,2)*d2N_dz2[1][2][a]
                                        + dz_dX(1,0)*dz_dX(2,2)*dz_dX(0,0)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,0)*dz_dX(2,2)*dz_dX(1,0)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,0,2,0)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,2,0)*d2N_dz2[2][0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,0,0)*d2N_dz2[2][0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,0,2)*d2N_dz2[2][0][a]
                                        + dz_dX(2,0)*dz_dX(0,2)*dz_dX(0,0)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,0)*dz_dX(0,2)*dz_dX(1,0)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,0,2,0)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,2,0)*d2N_dz2[2][1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,0,0)*d2N_dz2[2][1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,0,2)*d2N_dz2[2][1][a]
                                        + dz_dX(2,0)*dz_dX(1,2)*dz_dX(0,0)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,0)*dz_dX(1,2)*dz_dX(1,0)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,0,2,0)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,2,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,0,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,0,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*dz_dX(2,2)*dz_dX(0,0)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,0)*dz_dX(2,2)*dz_dX(1,0)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(1,2,0) = d3z_dX3(0,1,2,0)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,2)*dz_dX(0,0)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,2)*dz_dX(1,0)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,1,2,0)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,2,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,1,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,1,2)*d2N_dz2[0][1][a]
                                        + dz_dX(0,1)*dz_dX(1,2)*dz_dX(0,0)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,1)*dz_dX(1,2)*dz_dX(1,0)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,1,2,0)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,2,0)*d2N_dz2[0][2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,1,0)*d2N_dz2[0][2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,1,2)*d2N_dz2[0][2][a]
                                        + dz_dX(0,1)*dz_dX(2,2)*dz_dX(0,0)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,1)*dz_dX(2,2)*dz_dX(1,0)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,1,2,0)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,2,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,1,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,1,2)*d2N_dz2[1][0][a]
                                        + dz_dX(1,1)*dz_dX(0,2)*dz_dX(0,0)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,1)*dz_dX(0,2)*dz_dX(1,0)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,1,2,0)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,2,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,1,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,1,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*dz_dX(1,2)*dz_dX(0,0)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,1)*dz_dX(1,2)*dz_dX(1,0)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,1,2,0)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,2,0)*d2N_dz2[1][2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,1,0)*d2N_dz2[1][2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,1,2)*d2N_dz2[1][2][a]
                                        + dz_dX(1,1)*dz_dX(2,2)*dz_dX(0,0)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,1)*dz_dX(2,2)*dz_dX(1,0)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,1,2,0)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,2,0)*d2N_dz2[2][0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,1,0)*d2N_dz2[2][0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,1,2)*d2N_dz2[2][0][a]
                                        + dz_dX(2,1)*dz_dX(0,2)*dz_dX(0,0)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,1)*dz_dX(0,2)*dz_dX(1,0)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,1,2,0)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,2,0)*d2N_dz2[2][1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,1,0)*d2N_dz2[2][1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,1,2)*d2N_dz2[2][1][a]
                                        + dz_dX(2,1)*dz_dX(1,2)*dz_dX(0,0)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,1)*dz_dX(1,2)*dz_dX(1,0)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,1,2,0)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,2,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,1,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,1,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*dz_dX(2,2)*dz_dX(0,0)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,1)*dz_dX(2,2)*dz_dX(1,0)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(2,2,0) = d3z_dX3(0,2,2,0)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,2)*dz_dX(0,0)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,2)*dz_dX(1,0)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,2,2,0)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,2,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,2,0)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,2,2)*d2N_dz2[0][1][a]
                                        + dz_dX(0,2)*dz_dX(1,2)*dz_dX(0,0)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,2)*dz_dX(1,2)*dz_dX(1,0)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,2,2,0)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,2,0)*d2N_dz2[0][2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,2,0)*d2N_dz2[0][2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,2,2)*d2N_dz2[0][2][a]
                                        + dz_dX(0,2)*dz_dX(2,2)*dz_dX(0,0)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,2)*dz_dX(2,2)*dz_dX(1,0)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,2,2,0)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,2,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,2,0)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,2,2)*d2N_dz2[1][0][a]
                                        + dz_dX(1,2)*dz_dX(0,2)*dz_dX(0,0)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,2)*dz_dX(0,2)*dz_dX(1,0)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,2,2,0)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,2,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,2,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,2,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*dz_dX(1,2)*dz_dX(0,0)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,2)*dz_dX(1,2)*dz_dX(1,0)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,2,2,0)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,2,0)*d2N_dz2[1][2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,2,0)*d2N_dz2[1][2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,2,2)*d2N_dz2[1][2][a]
                                        + dz_dX(1,2)*dz_dX(2,2)*dz_dX(0,0)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,2)*dz_dX(2,2)*dz_dX(1,0)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,2,2,0)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,2,0)*d2N_dz2[2][0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,2,0)*d2N_dz2[2][0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,2,2)*d2N_dz2[2][0][a]
                                        + dz_dX(2,2)*dz_dX(0,2)*dz_dX(0,0)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,2)*dz_dX(0,2)*dz_dX(1,0)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,2,2,0)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,2,0)*d2N_dz2[2][1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,2,0)*d2N_dz2[2][1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,2,2)*d2N_dz2[2][1][a]
                                        + dz_dX(2,2)*dz_dX(1,2)*dz_dX(0,0)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,2)*dz_dX(1,2)*dz_dX(1,0)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,2,2,0)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,2,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,2,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,2,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*dz_dX(2,2)*dz_dX(0,0)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,2)*dz_dX(2,2)*dz_dX(1,0)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(0,0,1) = d3z_dX3(0,0,0,1)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,0)*dz_dX(0,1)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,0)*dz_dX(1,1)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,0,0,1)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,0,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,0,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,0,0)*d2N_dz2[0][1][a]
                                        + dz_dX(0,0)*dz_dX(1,0)*dz_dX(0,1)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,0)*dz_dX(1,0)*dz_dX(1,1)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,0,0,1)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,0,1)*d2N_dz2[0][2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,0,1)*d2N_dz2[0][2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,0,0)*d2N_dz2[0][2][a]
                                        + dz_dX(0,0)*dz_dX(2,0)*dz_dX(0,1)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,0)*dz_dX(2,0)*dz_dX(1,1)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,0,0,1)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,0,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,0,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,0,0)*d2N_dz2[1][0][a]
                                        + dz_dX(1,0)*dz_dX(0,0)*dz_dX(0,1)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,0)*dz_dX(0,0)*dz_dX(1,1)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,0,0,1)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,0,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,0,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,0,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*dz_dX(1,0)*dz_dX(0,1)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,0)*dz_dX(1,0)*dz_dX(1,1)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,0,0,1)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,0,1)*d2N_dz2[1][2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,0,1)*d2N_dz2[1][2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,0,0)*d2N_dz2[1][2][a]
                                        + dz_dX(1,0)*dz_dX(2,0)*dz_dX(0,1)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,0)*dz_dX(2,0)*dz_dX(1,1)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,0,0,1)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,0,1)*d2N_dz2[2][0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,0,1)*d2N_dz2[2][0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,0,0)*d2N_dz2[2][0][a]
                                        + dz_dX(2,0)*dz_dX(0,0)*dz_dX(0,1)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,0)*dz_dX(0,0)*dz_dX(1,1)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,0,0,1)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,0,1)*d2N_dz2[2][1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,0,1)*d2N_dz2[2][1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,0,0)*d2N_dz2[2][1][a]
                                        + dz_dX(2,0)*dz_dX(1,0)*dz_dX(0,1)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,0)*dz_dX(1,0)*dz_dX(1,1)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,0,0,1)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,0,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,0,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,0,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*dz_dX(2,0)*dz_dX(0,1)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,0)*dz_dX(2,0)*dz_dX(1,1)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(1,0,1) = d3z_dX3(0,1,0,1)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,0)*dz_dX(0,1)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,0)*dz_dX(1,1)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,1,0,1)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,0,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,1,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,1,0)*d2N_dz2[0][1][a]
                                        + dz_dX(0,1)*dz_dX(1,0)*dz_dX(0,1)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,1)*dz_dX(1,0)*dz_dX(1,1)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,1,0,1)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,0,1)*d2N_dz2[0][2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,1,1)*d2N_dz2[0][2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,1,0)*d2N_dz2[0][2][a]
                                        + dz_dX(0,1)*dz_dX(2,0)*dz_dX(0,1)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,1)*dz_dX(2,0)*dz_dX(1,1)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,1,0,1)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,0,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,1,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,1,0)*d2N_dz2[1][0][a]
                                        + dz_dX(1,1)*dz_dX(0,0)*dz_dX(0,1)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,1)*dz_dX(0,0)*dz_dX(1,1)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,1,0,1)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,0,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,1,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,1,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*dz_dX(1,0)*dz_dX(0,1)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,1)*dz_dX(1,0)*dz_dX(1,1)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,1,0,1)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,0,1)*d2N_dz2[1][2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,1,1)*d2N_dz2[1][2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,1,0)*d2N_dz2[1][2][a]
                                        + dz_dX(1,1)*dz_dX(2,0)*dz_dX(0,1)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,1)*dz_dX(2,0)*dz_dX(1,1)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,1,0,1)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,0,1)*d2N_dz2[2][0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,1,1)*d2N_dz2[2][0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,1,0)*d2N_dz2[2][0][a]
                                        + dz_dX(2,1)*dz_dX(0,0)*dz_dX(0,1)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,1)*dz_dX(0,0)*dz_dX(1,1)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,1,0,1)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,0,1)*d2N_dz2[2][1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,1,1)*d2N_dz2[2][1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,1,0)*d2N_dz2[2][1][a]
                                        + dz_dX(2,1)*dz_dX(1,0)*dz_dX(0,1)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,1)*dz_dX(1,0)*dz_dX(1,1)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,1,0,1)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,0,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,1,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,1,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*dz_dX(2,0)*dz_dX(0,1)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,1)*dz_dX(2,0)*dz_dX(1,1)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(2,0,1) = d3z_dX3(0,2,0,1)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,0)*dz_dX(0,1)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,0)*dz_dX(1,1)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,2,0,1)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,0,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,2,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,2,0)*d2N_dz2[0][1][a]
                                        + dz_dX(0,2)*dz_dX(1,0)*dz_dX(0,1)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,2)*dz_dX(1,0)*dz_dX(1,1)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,2,0,1)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,0,1)*d2N_dz2[0][2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,2,1)*d2N_dz2[0][2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,2,0)*d2N_dz2[0][2][a]
                                        + dz_dX(0,2)*dz_dX(2,0)*dz_dX(0,1)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,2)*dz_dX(2,0)*dz_dX(1,1)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,2,0,1)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,0,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,2,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,2,0)*d2N_dz2[1][0][a]
                                        + dz_dX(1,2)*dz_dX(0,0)*dz_dX(0,1)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,2)*dz_dX(0,0)*dz_dX(1,1)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,2,0,1)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,0,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,2,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,2,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*dz_dX(1,0)*dz_dX(0,1)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,2)*dz_dX(1,0)*dz_dX(1,1)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,2,0,1)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,0,1)*d2N_dz2[1][2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,2,1)*d2N_dz2[1][2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,2,0)*d2N_dz2[1][2][a]
                                        + dz_dX(1,2)*dz_dX(2,0)*dz_dX(0,1)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,2)*dz_dX(2,0)*dz_dX(1,1)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,2,0,1)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,0,1)*d2N_dz2[2][0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,2,1)*d2N_dz2[2][0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,2,0)*d2N_dz2[2][0][a]
                                        + dz_dX(2,2)*dz_dX(0,0)*dz_dX(0,1)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,2)*dz_dX(0,0)*dz_dX(1,1)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,2,0,1)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,0,1)*d2N_dz2[2][1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,2,1)*d2N_dz2[2][1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,2,0)*d2N_dz2[2][1][a]
                                        + dz_dX(2,2)*dz_dX(1,0)*dz_dX(0,1)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,2)*dz_dX(1,0)*dz_dX(1,1)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,2,0,1)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,0,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,2,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,2,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*dz_dX(2,0)*dz_dX(0,1)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,2)*dz_dX(2,0)*dz_dX(1,1)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(0,1,1) = d3z_dX3(0,0,1,1)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,1)*dz_dX(0,1)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,1)*dz_dX(1,1)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,0,1,1)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,1,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,0,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,0,1)*d2N_dz2[0][1][a]
                                        + dz_dX(0,0)*dz_dX(1,1)*dz_dX(0,1)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,0)*dz_dX(1,1)*dz_dX(1,1)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,0,1,1)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,1,1)*d2N_dz2[0][2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,0,1)*d2N_dz2[0][2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,0,1)*d2N_dz2[0][2][a]
                                        + dz_dX(0,0)*dz_dX(2,1)*dz_dX(0,1)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,0)*dz_dX(2,1)*dz_dX(1,1)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,0,1,1)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,1,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,0,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,0,1)*d2N_dz2[1][0][a]
                                        + dz_dX(1,0)*dz_dX(0,1)*dz_dX(0,1)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,0)*dz_dX(0,1)*dz_dX(1,1)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,0,1,1)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,1,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,0,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,0,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*dz_dX(1,1)*dz_dX(0,1)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,0)*dz_dX(1,1)*dz_dX(1,1)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,0,1,1)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,1,1)*d2N_dz2[1][2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,0,1)*d2N_dz2[1][2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,0,1)*d2N_dz2[1][2][a]
                                        + dz_dX(1,0)*dz_dX(2,1)*dz_dX(0,1)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,0)*dz_dX(2,1)*dz_dX(1,1)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,0,1,1)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,1,1)*d2N_dz2[2][0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,0,1)*d2N_dz2[2][0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,0,1)*d2N_dz2[2][0][a]
                                        + dz_dX(2,0)*dz_dX(0,1)*dz_dX(0,1)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,0)*dz_dX(0,1)*dz_dX(1,1)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,0,1,1)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,1,1)*d2N_dz2[2][1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,0,1)*d2N_dz2[2][1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,0,1)*d2N_dz2[2][1][a]
                                        + dz_dX(2,0)*dz_dX(1,1)*dz_dX(0,1)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,0)*dz_dX(1,1)*dz_dX(1,1)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,0,1,1)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,1,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,0,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,0,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*dz_dX(2,1)*dz_dX(0,1)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,0)*dz_dX(2,1)*dz_dX(1,1)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(1,1,1) = d3z_dX3(0,1,1,1)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,1)*dz_dX(0,1)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,1)*dz_dX(1,1)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,1,1,1)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,1,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,1,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,1,1)*d2N_dz2[0][1][a]
                                        + dz_dX(0,1)*dz_dX(1,1)*dz_dX(0,1)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,1)*dz_dX(1,1)*dz_dX(1,1)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,1,1,1)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,1,1)*d2N_dz2[0][2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,1,1)*d2N_dz2[0][2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,1,1)*d2N_dz2[0][2][a]
                                        + dz_dX(0,1)*dz_dX(2,1)*dz_dX(0,1)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,1)*dz_dX(2,1)*dz_dX(1,1)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,1,1,1)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,1,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,1,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,1,1)*d2N_dz2[1][0][a]
                                        + dz_dX(1,1)*dz_dX(0,1)*dz_dX(0,1)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,1)*dz_dX(0,1)*dz_dX(1,1)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,1,1,1)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,1,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,1,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,1,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*dz_dX(1,1)*dz_dX(0,1)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,1)*dz_dX(1,1)*dz_dX(1,1)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,1,1,1)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,1,1)*d2N_dz2[1][2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,1,1)*d2N_dz2[1][2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,1,1)*d2N_dz2[1][2][a]
                                        + dz_dX(1,1)*dz_dX(2,1)*dz_dX(0,1)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,1)*dz_dX(2,1)*dz_dX(1,1)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,1,1,1)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,1,1)*d2N_dz2[2][0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,1,1)*d2N_dz2[2][0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,1,1)*d2N_dz2[2][0][a]
                                        + dz_dX(2,1)*dz_dX(0,1)*dz_dX(0,1)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,1)*dz_dX(0,1)*dz_dX(1,1)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,1,1,1)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,1,1)*d2N_dz2[2][1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,1,1)*d2N_dz2[2][1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,1,1)*d2N_dz2[2][1][a]
                                        + dz_dX(2,1)*dz_dX(1,1)*dz_dX(0,1)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,1)*dz_dX(1,1)*dz_dX(1,1)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,1,1,1)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,1,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,1,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,1,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*dz_dX(2,1)*dz_dX(0,1)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,1)*dz_dX(2,1)*dz_dX(1,1)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(2,1,1) = d3z_dX3(0,2,1,1)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,1)*dz_dX(0,1)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,1)*dz_dX(1,1)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,2,1,1)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,1,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,2,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,2,1)*d2N_dz2[0][1][a]
                                        + dz_dX(0,2)*dz_dX(1,1)*dz_dX(0,1)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,2)*dz_dX(1,1)*dz_dX(1,1)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,2,1,1)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,1,1)*d2N_dz2[0][2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,2,1)*d2N_dz2[0][2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,2,1)*d2N_dz2[0][2][a]
                                        + dz_dX(0,2)*dz_dX(2,1)*dz_dX(0,1)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,2)*dz_dX(2,1)*dz_dX(1,1)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,2,1,1)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,1,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,2,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,2,1)*d2N_dz2[1][0][a]
                                        + dz_dX(1,2)*dz_dX(0,1)*dz_dX(0,1)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,2)*dz_dX(0,1)*dz_dX(1,1)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,2,1,1)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,1,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,2,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,2,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*dz_dX(1,1)*dz_dX(0,1)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,2)*dz_dX(1,1)*dz_dX(1,1)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,2,1,1)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,1,1)*d2N_dz2[1][2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,2,1)*d2N_dz2[1][2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,2,1)*d2N_dz2[1][2][a]
                                        + dz_dX(1,2)*dz_dX(2,1)*dz_dX(0,1)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,2)*dz_dX(2,1)*dz_dX(1,1)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,2,1,1)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,1,1)*d2N_dz2[2][0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,2,1)*d2N_dz2[2][0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,2,1)*d2N_dz2[2][0][a]
                                        + dz_dX(2,2)*dz_dX(0,1)*dz_dX(0,1)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,2)*dz_dX(0,1)*dz_dX(1,1)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,2,1,1)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,1,1)*d2N_dz2[2][1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,2,1)*d2N_dz2[2][1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,2,1)*d2N_dz2[2][1][a]
                                        + dz_dX(2,2)*dz_dX(1,1)*dz_dX(0,1)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,2)*dz_dX(1,1)*dz_dX(1,1)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,2,1,1)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,1,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,2,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,2,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*dz_dX(2,1)*dz_dX(0,1)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,2)*dz_dX(2,1)*dz_dX(1,1)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(0,2,1) = d3z_dX3(0,0,2,1)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,2)*dz_dX(0,1)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,2)*dz_dX(1,1)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,0,2,1)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,2,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,0,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,0,2)*d2N_dz2[0][1][a]
                                        + dz_dX(0,0)*dz_dX(1,2)*dz_dX(0,1)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,0)*dz_dX(1,2)*dz_dX(1,1)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,0,2,1)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,2,1)*d2N_dz2[0][2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,0,1)*d2N_dz2[0][2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,0,2)*d2N_dz2[0][2][a]
                                        + dz_dX(0,0)*dz_dX(2,2)*dz_dX(0,1)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,0)*dz_dX(2,2)*dz_dX(1,1)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,0,2,1)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,2,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,0,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,0,2)*d2N_dz2[1][0][a]
                                        + dz_dX(1,0)*dz_dX(0,2)*dz_dX(0,1)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,0)*dz_dX(0,2)*dz_dX(1,1)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,0,2,1)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,2,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,0,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,0,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*dz_dX(1,2)*dz_dX(0,1)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,0)*dz_dX(1,2)*dz_dX(1,1)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,0,2,1)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,2,1)*d2N_dz2[1][2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,0,1)*d2N_dz2[1][2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,0,2)*d2N_dz2[1][2][a]
                                        + dz_dX(1,0)*dz_dX(2,2)*dz_dX(0,1)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,0)*dz_dX(2,2)*dz_dX(1,1)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,0,2,1)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,2,1)*d2N_dz2[2][0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,0,1)*d2N_dz2[2][0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,0,2)*d2N_dz2[2][0][a]
                                        + dz_dX(2,0)*dz_dX(0,2)*dz_dX(0,1)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,0)*dz_dX(0,2)*dz_dX(1,1)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,0,2,1)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,2,1)*d2N_dz2[2][1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,0,1)*d2N_dz2[2][1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,0,2)*d2N_dz2[2][1][a]
                                        + dz_dX(2,0)*dz_dX(1,2)*dz_dX(0,1)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,0)*dz_dX(1,2)*dz_dX(1,1)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,0,2,1)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,2,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,0,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,0,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*dz_dX(2,2)*dz_dX(0,1)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,0)*dz_dX(2,2)*dz_dX(1,1)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(1,2,1) = d3z_dX3(0,1,2,1)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,2)*dz_dX(0,1)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,2)*dz_dX(1,1)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,1,2,1)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,2,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,1,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,1,2)*d2N_dz2[0][1][a]
                                        + dz_dX(0,1)*dz_dX(1,2)*dz_dX(0,1)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,1)*dz_dX(1,2)*dz_dX(1,1)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,1,2,1)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,2,1)*d2N_dz2[0][2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,1,1)*d2N_dz2[0][2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,1,2)*d2N_dz2[0][2][a]
                                        + dz_dX(0,1)*dz_dX(2,2)*dz_dX(0,1)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,1)*dz_dX(2,2)*dz_dX(1,1)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,1,2,1)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,2,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,1,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,1,2)*d2N_dz2[1][0][a]
                                        + dz_dX(1,1)*dz_dX(0,2)*dz_dX(0,1)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,1)*dz_dX(0,2)*dz_dX(1,1)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,1,2,1)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,2,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,1,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,1,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*dz_dX(1,2)*dz_dX(0,1)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,1)*dz_dX(1,2)*dz_dX(1,1)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,1,2,1)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,2,1)*d2N_dz2[1][2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,1,1)*d2N_dz2[1][2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,1,2)*d2N_dz2[1][2][a]
                                        + dz_dX(1,1)*dz_dX(2,2)*dz_dX(0,1)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,1)*dz_dX(2,2)*dz_dX(1,1)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,1,2,1)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,2,1)*d2N_dz2[2][0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,1,1)*d2N_dz2[2][0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,1,2)*d2N_dz2[2][0][a]
                                        + dz_dX(2,1)*dz_dX(0,2)*dz_dX(0,1)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,1)*dz_dX(0,2)*dz_dX(1,1)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,1,2,1)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,2,1)*d2N_dz2[2][1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,1,1)*d2N_dz2[2][1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,1,2)*d2N_dz2[2][1][a]
                                        + dz_dX(2,1)*dz_dX(1,2)*dz_dX(0,1)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,1)*dz_dX(1,2)*dz_dX(1,1)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,1,2,1)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,2,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,1,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,1,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*dz_dX(2,2)*dz_dX(0,1)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,1)*dz_dX(2,2)*dz_dX(1,1)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(2,2,1) = d3z_dX3(0,2,2,1)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,2)*dz_dX(0,1)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,2)*dz_dX(1,1)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,2,2,1)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,2,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,2,1)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,2,2)*d2N_dz2[0][1][a]
                                        + dz_dX(0,2)*dz_dX(1,2)*dz_dX(0,1)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,2)*dz_dX(1,2)*dz_dX(1,1)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,2,2,1)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,2,1)*d2N_dz2[0][2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,2,1)*d2N_dz2[0][2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,2,2)*d2N_dz2[0][2][a]
                                        + dz_dX(0,2)*dz_dX(2,2)*dz_dX(0,1)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,2)*dz_dX(2,2)*dz_dX(1,1)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,2,2,1)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,2,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,2,1)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,2,2)*d2N_dz2[1][0][a]
                                        + dz_dX(1,2)*dz_dX(0,2)*dz_dX(0,1)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,2)*dz_dX(0,2)*dz_dX(1,1)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,2,2,1)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,2,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,2,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,2,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*dz_dX(1,2)*dz_dX(0,1)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,2)*dz_dX(1,2)*dz_dX(1,1)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,2,2,1)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,2,1)*d2N_dz2[1][2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,2,1)*d2N_dz2[1][2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,2,2)*d2N_dz2[1][2][a]
                                        + dz_dX(1,2)*dz_dX(2,2)*dz_dX(0,1)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,2)*dz_dX(2,2)*dz_dX(1,1)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,2,2,1)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,2,1)*d2N_dz2[2][0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,2,1)*d2N_dz2[2][0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,2,2)*d2N_dz2[2][0][a]
                                        + dz_dX(2,2)*dz_dX(0,2)*dz_dX(0,1)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,2)*dz_dX(0,2)*dz_dX(1,1)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,2,2,1)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,2,1)*d2N_dz2[2][1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,2,1)*d2N_dz2[2][1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,2,2)*d2N_dz2[2][1][a]
                                        + dz_dX(2,2)*dz_dX(1,2)*dz_dX(0,1)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,2)*dz_dX(1,2)*dz_dX(1,1)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,2,2,1)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,2,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,2,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,2,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*dz_dX(2,2)*dz_dX(0,1)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,2)*dz_dX(2,2)*dz_dX(1,1)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(0,0,2) = d3z_dX3(0,0,0,2)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,0)*dz_dX(0,2)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,0)*dz_dX(1,2)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,0,0,2)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,0,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,0,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,0,0)*d2N_dz2[0][1][a]
                                        + dz_dX(0,0)*dz_dX(1,0)*dz_dX(0,2)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,0)*dz_dX(1,0)*dz_dX(1,2)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,0,0,2)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,0,2)*d2N_dz2[0][2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,0,2)*d2N_dz2[0][2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,0,0)*d2N_dz2[0][2][a]
                                        + dz_dX(0,0)*dz_dX(2,0)*dz_dX(0,2)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,0)*dz_dX(2,0)*dz_dX(1,2)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,0,0,2)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,0,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,0,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,0,0)*d2N_dz2[1][0][a]
                                        + dz_dX(1,0)*dz_dX(0,0)*dz_dX(0,2)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,0)*dz_dX(0,0)*dz_dX(1,2)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,0,0,2)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,0,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,0,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,0,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*dz_dX(1,0)*dz_dX(0,2)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,0)*dz_dX(1,0)*dz_dX(1,2)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,0,0,2)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,0,2)*d2N_dz2[1][2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,0,2)*d2N_dz2[1][2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,0,0)*d2N_dz2[1][2][a]
                                        + dz_dX(1,0)*dz_dX(2,0)*dz_dX(0,2)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,0)*dz_dX(2,0)*dz_dX(1,2)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,0,0,2)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,0,2)*d2N_dz2[2][0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,0,2)*d2N_dz2[2][0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,0,0)*d2N_dz2[2][0][a]
                                        + dz_dX(2,0)*dz_dX(0,0)*dz_dX(0,2)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,0)*dz_dX(0,0)*dz_dX(1,2)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,0,0,2)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,0,2)*d2N_dz2[2][1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,0,2)*d2N_dz2[2][1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,0,0)*d2N_dz2[2][1][a]
                                        + dz_dX(2,0)*dz_dX(1,0)*dz_dX(0,2)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,0)*dz_dX(1,0)*dz_dX(1,2)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,0,0,2)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,0,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,0,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,0,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*dz_dX(2,0)*dz_dX(0,2)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,0)*dz_dX(2,0)*dz_dX(1,2)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(1,0,2) = d3z_dX3(0,1,0,2)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,0)*dz_dX(0,2)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,0)*dz_dX(1,2)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,1,0,2)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,0,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,1,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,1,0)*d2N_dz2[0][1][a]
                                        + dz_dX(0,1)*dz_dX(1,0)*dz_dX(0,2)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,1)*dz_dX(1,0)*dz_dX(1,2)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,1,0,2)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,0,2)*d2N_dz2[0][2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,1,2)*d2N_dz2[0][2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,1,0)*d2N_dz2[0][2][a]
                                        + dz_dX(0,1)*dz_dX(2,0)*dz_dX(0,2)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,1)*dz_dX(2,0)*dz_dX(1,2)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,1,0,2)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,0,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,1,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,1,0)*d2N_dz2[1][0][a]
                                        + dz_dX(1,1)*dz_dX(0,0)*dz_dX(0,2)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,1)*dz_dX(0,0)*dz_dX(1,2)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,1,0,2)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,0,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,1,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,1,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*dz_dX(1,0)*dz_dX(0,2)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,1)*dz_dX(1,0)*dz_dX(1,2)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,1,0,2)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,0,2)*d2N_dz2[1][2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,1,2)*d2N_dz2[1][2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,1,0)*d2N_dz2[1][2][a]
                                        + dz_dX(1,1)*dz_dX(2,0)*dz_dX(0,2)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,1)*dz_dX(2,0)*dz_dX(1,2)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,1,0,2)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,0,2)*d2N_dz2[2][0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,1,2)*d2N_dz2[2][0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,1,0)*d2N_dz2[2][0][a]
                                        + dz_dX(2,1)*dz_dX(0,0)*dz_dX(0,2)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,1)*dz_dX(0,0)*dz_dX(1,2)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,1,0,2)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,0,2)*d2N_dz2[2][1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,1,2)*d2N_dz2[2][1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,1,0)*d2N_dz2[2][1][a]
                                        + dz_dX(2,1)*dz_dX(1,0)*dz_dX(0,2)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,1)*dz_dX(1,0)*dz_dX(1,2)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,1,0,2)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,0,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,1,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,1,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*dz_dX(2,0)*dz_dX(0,2)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,1)*dz_dX(2,0)*dz_dX(1,2)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(2,0,2) = d3z_dX3(0,2,0,2)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,0)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,0)*dz_dX(0,2)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,0)*dz_dX(1,2)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,2,0,2)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,0,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,2,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,2,0)*d2N_dz2[0][1][a]
                                        + dz_dX(0,2)*dz_dX(1,0)*dz_dX(0,2)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,2)*dz_dX(1,0)*dz_dX(1,2)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,2,0,2)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,0,2)*d2N_dz2[0][2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,2,2)*d2N_dz2[0][2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,2,0)*d2N_dz2[0][2][a]
                                        + dz_dX(0,2)*dz_dX(2,0)*dz_dX(0,2)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,2)*dz_dX(2,0)*dz_dX(1,2)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,2,0,2)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,0,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,2,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,2,0)*d2N_dz2[1][0][a]
                                        + dz_dX(1,2)*dz_dX(0,0)*dz_dX(0,2)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,2)*dz_dX(0,0)*dz_dX(1,2)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,2,0,2)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,0,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,2,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,2,0)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*dz_dX(1,0)*dz_dX(0,2)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,2)*dz_dX(1,0)*dz_dX(1,2)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,2,0,2)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,0,2)*d2N_dz2[1][2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,2,2)*d2N_dz2[1][2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,2,0)*d2N_dz2[1][2][a]
                                        + dz_dX(1,2)*dz_dX(2,0)*dz_dX(0,2)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,2)*dz_dX(2,0)*dz_dX(1,2)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,2,0,2)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,0,2)*d2N_dz2[2][0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,2,2)*d2N_dz2[2][0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,2,0)*d2N_dz2[2][0][a]
                                        + dz_dX(2,2)*dz_dX(0,0)*dz_dX(0,2)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,2)*dz_dX(0,0)*dz_dX(1,2)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,2,0,2)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,0,2)*d2N_dz2[2][1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,2,2)*d2N_dz2[2][1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,2,0)*d2N_dz2[2][1][a]
                                        + dz_dX(2,2)*dz_dX(1,0)*dz_dX(0,2)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,2)*dz_dX(1,0)*dz_dX(1,2)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,2,0,2)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,0,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,2,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,2,0)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*dz_dX(2,0)*dz_dX(0,2)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,2)*dz_dX(2,0)*dz_dX(1,2)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(0,1,2) = d3z_dX3(0,0,1,2)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,1)*dz_dX(0,2)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,1)*dz_dX(1,2)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,0,1,2)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,1,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,0,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,0,1)*d2N_dz2[0][1][a]
                                        + dz_dX(0,0)*dz_dX(1,1)*dz_dX(0,2)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,0)*dz_dX(1,1)*dz_dX(1,2)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,0,1,2)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,1,2)*d2N_dz2[0][2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,0,2)*d2N_dz2[0][2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,0,1)*d2N_dz2[0][2][a]
                                        + dz_dX(0,0)*dz_dX(2,1)*dz_dX(0,2)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,0)*dz_dX(2,1)*dz_dX(1,2)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,0,1,2)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,1,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,0,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,0,1)*d2N_dz2[1][0][a]
                                        + dz_dX(1,0)*dz_dX(0,1)*dz_dX(0,2)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,0)*dz_dX(0,1)*dz_dX(1,2)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,0,1,2)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,1,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,0,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,0,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*dz_dX(1,1)*dz_dX(0,2)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,0)*dz_dX(1,1)*dz_dX(1,2)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,0,1,2)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,1,2)*d2N_dz2[1][2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,0,2)*d2N_dz2[1][2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,0,1)*d2N_dz2[1][2][a]
                                        + dz_dX(1,0)*dz_dX(2,1)*dz_dX(0,2)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,0)*dz_dX(2,1)*dz_dX(1,2)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,0,1,2)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,1,2)*d2N_dz2[2][0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,0,2)*d2N_dz2[2][0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,0,1)*d2N_dz2[2][0][a]
                                        + dz_dX(2,0)*dz_dX(0,1)*dz_dX(0,2)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,0)*dz_dX(0,1)*dz_dX(1,2)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,0,1,2)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,1,2)*d2N_dz2[2][1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,0,2)*d2N_dz2[2][1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,0,1)*d2N_dz2[2][1][a]
                                        + dz_dX(2,0)*dz_dX(1,1)*dz_dX(0,2)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,0)*dz_dX(1,1)*dz_dX(1,2)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,0,1,2)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,1,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,0,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,0,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*dz_dX(2,1)*dz_dX(0,2)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,0)*dz_dX(2,1)*dz_dX(1,2)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(1,1,2) = d3z_dX3(0,1,1,2)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,1)*dz_dX(0,2)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,1)*dz_dX(1,2)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,1,1,2)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,1,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,1,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,1,1)*d2N_dz2[0][1][a]
                                        + dz_dX(0,1)*dz_dX(1,1)*dz_dX(0,2)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,1)*dz_dX(1,1)*dz_dX(1,2)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,1,1,2)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,1,2)*d2N_dz2[0][2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,1,2)*d2N_dz2[0][2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,1,1)*d2N_dz2[0][2][a]
                                        + dz_dX(0,1)*dz_dX(2,1)*dz_dX(0,2)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,1)*dz_dX(2,1)*dz_dX(1,2)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,1,1,2)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,1,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,1,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,1,1)*d2N_dz2[1][0][a]
                                        + dz_dX(1,1)*dz_dX(0,1)*dz_dX(0,2)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,1)*dz_dX(0,1)*dz_dX(1,2)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,1,1,2)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,1,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,1,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,1,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*dz_dX(1,1)*dz_dX(0,2)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,1)*dz_dX(1,1)*dz_dX(1,2)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,1,1,2)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,1,2)*d2N_dz2[1][2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,1,2)*d2N_dz2[1][2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,1,1)*d2N_dz2[1][2][a]
                                        + dz_dX(1,1)*dz_dX(2,1)*dz_dX(0,2)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,1)*dz_dX(2,1)*dz_dX(1,2)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,1,1,2)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,1,2)*d2N_dz2[2][0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,1,2)*d2N_dz2[2][0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,1,1)*d2N_dz2[2][0][a]
                                        + dz_dX(2,1)*dz_dX(0,1)*dz_dX(0,2)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,1)*dz_dX(0,1)*dz_dX(1,2)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,1,1,2)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,1,2)*d2N_dz2[2][1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,1,2)*d2N_dz2[2][1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,1,1)*d2N_dz2[2][1][a]
                                        + dz_dX(2,1)*dz_dX(1,1)*dz_dX(0,2)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,1)*dz_dX(1,1)*dz_dX(1,2)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,1,1,2)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,1,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,1,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,1,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*dz_dX(2,1)*dz_dX(0,2)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,1)*dz_dX(2,1)*dz_dX(1,2)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(2,1,2) = d3z_dX3(0,2,1,2)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,1)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,1)*dz_dX(0,2)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,1)*dz_dX(1,2)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,2,1,2)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,1,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,2,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,2,1)*d2N_dz2[0][1][a]
                                        + dz_dX(0,2)*dz_dX(1,1)*dz_dX(0,2)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,2)*dz_dX(1,1)*dz_dX(1,2)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,2,1,2)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,1,2)*d2N_dz2[0][2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,2,2)*d2N_dz2[0][2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,2,1)*d2N_dz2[0][2][a]
                                        + dz_dX(0,2)*dz_dX(2,1)*dz_dX(0,2)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,2)*dz_dX(2,1)*dz_dX(1,2)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,2,1,2)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,1,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,2,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,2,1)*d2N_dz2[1][0][a]
                                        + dz_dX(1,2)*dz_dX(0,1)*dz_dX(0,2)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,2)*dz_dX(0,1)*dz_dX(1,2)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,2,1,2)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,1,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,2,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,2,1)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*dz_dX(1,1)*dz_dX(0,2)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,2)*dz_dX(1,1)*dz_dX(1,2)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,2,1,2)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,1,2)*d2N_dz2[1][2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,2,2)*d2N_dz2[1][2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,2,1)*d2N_dz2[1][2][a]
                                        + dz_dX(1,2)*dz_dX(2,1)*dz_dX(0,2)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,2)*dz_dX(2,1)*dz_dX(1,2)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,2,1,2)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,1,2)*d2N_dz2[2][0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,2,2)*d2N_dz2[2][0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,2,1)*d2N_dz2[2][0][a]
                                        + dz_dX(2,2)*dz_dX(0,1)*dz_dX(0,2)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,2)*dz_dX(0,1)*dz_dX(1,2)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,2,1,2)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,1,2)*d2N_dz2[2][1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,2,2)*d2N_dz2[2][1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,2,1)*d2N_dz2[2][1][a]
                                        + dz_dX(2,2)*dz_dX(1,1)*dz_dX(0,2)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,2)*dz_dX(1,1)*dz_dX(1,2)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,2,1,2)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,1,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,2,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,2,1)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*dz_dX(2,1)*dz_dX(0,2)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,2)*dz_dX(2,1)*dz_dX(1,2)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(0,2,2) = d3z_dX3(0,0,2,2)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,0,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,2)*dz_dX(0,2)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,0)*dz_dX(0,2)*dz_dX(1,2)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,0,2,2)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(1,2,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,0,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,0,2)*d2N_dz2[0][1][a]
                                        + dz_dX(0,0)*dz_dX(1,2)*dz_dX(0,2)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,0)*dz_dX(1,2)*dz_dX(1,2)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,0,2,2)*dN_dz[0][a]
                                        + dz_dX(0,0)*d2z_dX2(2,2,2)*d2N_dz2[0][2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,0,2)*d2N_dz2[0][2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,0,2)*d2N_dz2[0][2][a]
                                        + dz_dX(0,0)*dz_dX(2,2)*dz_dX(0,2)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,0)*dz_dX(2,2)*dz_dX(1,2)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,0,2,2)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(0,2,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,0,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,0,2)*d2N_dz2[1][0][a]
                                        + dz_dX(1,0)*dz_dX(0,2)*dz_dX(0,2)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,0)*dz_dX(0,2)*dz_dX(1,2)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,0,2,2)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(1,2,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,0,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,0,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,0)*dz_dX(1,2)*dz_dX(0,2)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,0)*dz_dX(1,2)*dz_dX(1,2)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,0,2,2)*dN_dz[1][a]
                                        + dz_dX(1,0)*d2z_dX2(2,2,2)*d2N_dz2[1][2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,0,2)*d2N_dz2[1][2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,0,2)*d2N_dz2[1][2][a]
                                        + dz_dX(1,0)*dz_dX(2,2)*dz_dX(0,2)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,0)*dz_dX(2,2)*dz_dX(1,2)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,0,2,2)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(0,2,2)*d2N_dz2[2][0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,0,2)*d2N_dz2[2][0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,0,2)*d2N_dz2[2][0][a]
                                        + dz_dX(2,0)*dz_dX(0,2)*dz_dX(0,2)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,0)*dz_dX(0,2)*dz_dX(1,2)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,0,2,2)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(1,2,2)*d2N_dz2[2][1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,0,2)*d2N_dz2[2][1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,0,2)*d2N_dz2[2][1][a]
                                        + dz_dX(2,0)*dz_dX(1,2)*dz_dX(0,2)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,0)*dz_dX(1,2)*dz_dX(1,2)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,0,2,2)*dN_dz[2][a]
                                        + dz_dX(2,0)*d2z_dX2(2,2,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,0,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,0,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,0)*dz_dX(2,2)*dz_dX(0,2)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,0)*dz_dX(2,2)*dz_dX(1,2)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(1,2,2) = d3z_dX3(0,1,2,2)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,1,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,2)*dz_dX(0,2)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,1)*dz_dX(0,2)*dz_dX(1,2)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,1,2,2)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(1,2,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,1,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,1,2)*d2N_dz2[0][1][a]
                                        + dz_dX(0,1)*dz_dX(1,2)*dz_dX(0,2)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,1)*dz_dX(1,2)*dz_dX(1,2)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,1,2,2)*dN_dz[0][a]
                                        + dz_dX(0,1)*d2z_dX2(2,2,2)*d2N_dz2[0][2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,1,2)*d2N_dz2[0][2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,1,2)*d2N_dz2[0][2][a]
                                        + dz_dX(0,1)*dz_dX(2,2)*dz_dX(0,2)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,1)*dz_dX(2,2)*dz_dX(1,2)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,1,2,2)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(0,2,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,1,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,1,2)*d2N_dz2[1][0][a]
                                        + dz_dX(1,1)*dz_dX(0,2)*dz_dX(0,2)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,1)*dz_dX(0,2)*dz_dX(1,2)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,1,2,2)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(1,2,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,1,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,1,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,1)*dz_dX(1,2)*dz_dX(0,2)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,1)*dz_dX(1,2)*dz_dX(1,2)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,1,2,2)*dN_dz[1][a]
                                        + dz_dX(1,1)*d2z_dX2(2,2,2)*d2N_dz2[1][2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,1,2)*d2N_dz2[1][2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,1,2)*d2N_dz2[1][2][a]
                                        + dz_dX(1,1)*dz_dX(2,2)*dz_dX(0,2)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,1)*dz_dX(2,2)*dz_dX(1,2)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,1,2,2)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(0,2,2)*d2N_dz2[2][0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,1,2)*d2N_dz2[2][0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,1,2)*d2N_dz2[2][0][a]
                                        + dz_dX(2,1)*dz_dX(0,2)*dz_dX(0,2)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,1)*dz_dX(0,2)*dz_dX(1,2)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,1,2,2)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(1,2,2)*d2N_dz2[2][1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,1,2)*d2N_dz2[2][1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,1,2)*d2N_dz2[2][1][a]
                                        + dz_dX(2,1)*dz_dX(1,2)*dz_dX(0,2)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,1)*dz_dX(1,2)*dz_dX(1,2)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,1,2,2)*dN_dz[2][a]
                                        + dz_dX(2,1)*d2z_dX2(2,2,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,1,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,1,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,1)*dz_dX(2,2)*dz_dX(0,2)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,1)*dz_dX(2,2)*dz_dX(1,2)*d3N_dz3[2][2][1][a];
            this->_d3phi[a].comp(2,2,2) = d3z_dX3(0,2,2,2)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*d2z_dX2(0,2,2)*d2N_dz2[0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,2)*dz_dX(0,2)*d3N_dz3[0][0][0][a]
                                        + dz_dX(0,2)*dz_dX(0,2)*dz_dX(1,2)*d3N_dz3[0][0][1][a]
                                        + d3z_dX3(0,2,2,2)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,2,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,2,2)*d2N_dz2[0][1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,2,2)*d2N_dz2[0][1][a]
                                        + dz_dX(0,2)*dz_dX(1,2)*dz_dX(0,2)*d3N_dz3[0][1][0][a]
                                        + dz_dX(0,2)*dz_dX(1,2)*dz_dX(1,2)*d3N_dz3[0][1][1][a]
                                        + d3z_dX3(0,2,2,2)*dN_dz[0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,2,2)*d2N_dz2[0][2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,2,2)*d2N_dz2[0][2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,2,2)*d2N_dz2[0][2][a]
                                        + dz_dX(0,2)*dz_dX(2,2)*dz_dX(0,2)*d3N_dz3[0][2][0][a]
                                        + dz_dX(0,2)*dz_dX(2,2)*dz_dX(1,2)*d3N_dz3[0][2][1][a]
                                        + d3z_dX3(1,2,2,2)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(0,2,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,2,2)*d2N_dz2[1][0][a]
                                        + dz_dX(0,2)*d2z_dX2(1,2,2)*d2N_dz2[1][0][a]
                                        + dz_dX(1,2)*dz_dX(0,2)*dz_dX(0,2)*d3N_dz3[1][0][0][a]
                                        + dz_dX(1,2)*dz_dX(0,2)*dz_dX(1,2)*d3N_dz3[1][0][1][a]
                                        + d3z_dX3(1,2,2,2)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,2,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,2,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*d2z_dX2(1,2,2)*d2N_dz2[1][1][a]
                                        + dz_dX(1,2)*dz_dX(1,2)*dz_dX(0,2)*d3N_dz3[1][1][0][a]
                                        + dz_dX(1,2)*dz_dX(1,2)*dz_dX(1,2)*d3N_dz3[1][1][1][a]
                                        + d3z_dX3(1,2,2,2)*dN_dz[1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,2,2)*d2N_dz2[1][2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,2,2)*d2N_dz2[1][2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,2,2)*d2N_dz2[1][2][a]
                                        + dz_dX(1,2)*dz_dX(2,2)*dz_dX(0,2)*d3N_dz3[1][2][0][a]
                                        + dz_dX(1,2)*dz_dX(2,2)*dz_dX(1,2)*d3N_dz3[1][2][1][a]
                                        + d3z_dX3(2,2,2,2)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(0,2,2)*d2N_dz2[2][0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,2,2)*d2N_dz2[2][0][a]
                                        + dz_dX(0,2)*d2z_dX2(2,2,2)*d2N_dz2[2][0][a]
                                        + dz_dX(2,2)*dz_dX(0,2)*dz_dX(0,2)*d3N_dz3[2][0][0][a]
                                        + dz_dX(2,2)*dz_dX(0,2)*dz_dX(1,2)*d3N_dz3[2][0][1][a]
                                        + d3z_dX3(2,2,2,2)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(1,2,2)*d2N_dz2[2][1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,2,2)*d2N_dz2[2][1][a]
                                        + dz_dX(1,2)*d2z_dX2(2,2,2)*d2N_dz2[2][1][a]
                                        + dz_dX(2,2)*dz_dX(1,2)*dz_dX(0,2)*d3N_dz3[2][1][0][a]
                                        + dz_dX(2,2)*dz_dX(1,2)*dz_dX(1,2)*d3N_dz3[2][1][1][a]
                                        + d3z_dX3(2,2,2,2)*dN_dz[2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,2,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,2,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*d2z_dX2(2,2,2)*d2N_dz2[2][2][a]
                                        + dz_dX(2,2)*dz_dX(2,2)*dz_dX(0,2)*d3N_dz3[2][2][0][a]
                                        + dz_dX(2,2)*dz_dX(2,2)*dz_dX(1,2)*d3N_dz3[2][2][1][a];
        }
        //
    } else libmesh_error();
}
//=================================================================================================
#undef dz_dX
#undef d2z_dX2
#undef d3z_dX3
#undef dN_dz
#undef d2N_dz2
#undef d3N_dz3
//=================================================================================================
void NURBS::calc_d3phidxidxidxi (const Point& crd) const {
    const int L = 0, M = 0, N = 0;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a);
            this->_d3phidz3[L][M][N][a] = this->_basef_3rd_deriv<0>(crd(0), i);
        }
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a);
            this->_d3phidz3[L][M][N][a] = this->_basef_3rd_deriv<0>(crd(0), i)
                                        * this->_basef_0th_deriv<1>(crd(1), j);
        }
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a),
                      k = this->index<2>(a);
            this->_d3phidz3[L][M][N][a] = this->_basef_3rd_deriv<0>(crd(0), i)
                                        * this->_basef_0th_deriv<1>(crd(1), j)
                                        * this->_basef_0th_deriv<2>(crd(2), k);
        }
    }
    //
    if ( this->_is_rational ) this->_rationalize_d3phi(L, M, N);
}
//=================================================================================================
void NURBS::calc_d3phidxidxideta (const Point& crd) const {
    const int L = 0, M = 0, N = 1;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a);
            this->_d3phidz3[L][M][N][a] = 0.0;
        }
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a);
            this->_d3phidz3[L][M][N][a] = this->_basef_2nd_deriv<0>(crd(0), i)
                                        * this->_basef_1st_deriv<1>(crd(1), j);
        }
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a),
                      k = this->index<2>(a);
            this->_d3phidz3[L][M][N][a] = this->_basef_2nd_deriv<0>(crd(0), i)
                                        * this->_basef_1st_deriv<1>(crd(1), j)
                                        * this->_basef_0th_deriv<2>(crd(2), k);
        }
    }
    //
    if ( this->_is_rational ) this->_rationalize_d3phi(L, M, N);
}
//=================================================================================================
void NURBS::calc_d3phidxidxidzeta (const Point& crd) const {
    const int L = 0, M = 0, N = 2;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a);
            this->_d3phidz3[L][M][N][a] = 0.0;
        }
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a),
            //          j = this->index<1>(a);
            this->_d3phidz3[L][M][N][a] = 0.0;
        }
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a),
                      k = this->index<2>(a);
            this->_d3phidz3[L][M][N][a] = this->_basef_2nd_deriv<0>(crd(0), i)
                                        * this->_basef_0th_deriv<1>(crd(1), j)
                                        * this->_basef_1st_deriv<2>(crd(2), k);
        }
    }
    //
    if ( this->_is_rational ) this->_rationalize_d3phi(L, M, N);
}
//=================================================================================================
void NURBS::calc_d3phidxidetadxi (const Point& crd) const {
    const int L = 0, M = 1, N = 0;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    this->_d3phidz3[L][M][N] = this->_d3phidz3[L][N][M];
    if ( this->_is_rational ) this->_sum__w_d3Ndz3[L][M][N] = this->_sum__w_d3Ndz3[L][N][M];
}
//=================================================================================================
void NURBS::calc_d3phidxidetadeta (const Point& crd) const {
    const int L = 0, M = 1, N = 1;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a);
            this->_d3phidz3[L][M][N][a] = 0.0;
        }
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a);
            this->_d3phidz3[L][M][N][a] = this->_basef_1st_deriv<0>(crd(0), i)
                                        * this->_basef_2nd_deriv<1>(crd(1), j);
        }
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a),
                      k = this->index<2>(a);
            this->_d3phidz3[L][M][N][a] = this->_basef_1st_deriv<0>(crd(0), i)
                                        * this->_basef_2nd_deriv<1>(crd(1), j)
                                        * this->_basef_0th_deriv<2>(crd(2), k);
        }
    }
    //
    if ( this->_is_rational ) this->_rationalize_d3phi(L, M, N);
}
//=================================================================================================
void NURBS::calc_d3phidxidetadzeta (const Point& crd) const {
    const int L = 0, M = 1, N = 2;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a);
            this->_d3phidz3[L][M][N][a] = 0.0;
        }
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a),
            //          j = this->index<1>(a);
            this->_d3phidz3[L][M][N][a] = 0.0;
        }
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a),
                      k = this->index<2>(a);
            this->_d3phidz3[L][M][N][a] = this->_basef_1st_deriv<0>(crd(0), i)
                                        * this->_basef_1st_deriv<1>(crd(1), j)
                                        * this->_basef_1st_deriv<2>(crd(2), k);
        }
    }
    //
    if ( this->_is_rational ) this->_rationalize_d3phi(L, M, N);
}
//=================================================================================================
void NURBS::calc_d3phidxidzetadxi (const Point& crd) const {
    const int L = 0, M = 2, N = 0;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    this->_d3phidz3[L][M][N] = this->_d3phidz3[L][N][M];
    if ( this->_is_rational ) this->_sum__w_d3Ndz3[L][M][N] = this->_sum__w_d3Ndz3[L][N][M];
}
//=================================================================================================
void NURBS::calc_d3phidxidzetadeta (const Point& crd) const {
    const int L = 0, M = 2, N = 1;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    this->_d3phidz3[L][M][N] = this->_d3phidz3[L][N][M];
    if ( this->_is_rational ) this->_sum__w_d3Ndz3[L][M][N] = this->_sum__w_d3Ndz3[L][N][M];
}
//=================================================================================================
void NURBS::calc_d3phidxidzetadzeta (const Point& crd) const {
    const int L = 0, M = 2, N = 2;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a);
            this->_d3phidz3[L][M][N][a] = 0.0;
        }
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a),
            //          j = this->index<1>(a);
            this->_d3phidz3[L][M][N][a] = 0.0;
        }
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a),
                      k = this->index<2>(a);
            this->_d3phidz3[L][M][N][a] = this->_basef_1st_deriv<0>(crd(0), i)
                                        * this->_basef_0th_deriv<1>(crd(1), j)
                                        * this->_basef_2nd_deriv<2>(crd(2), k);
        }
    }
    //
    if ( this->_is_rational ) this->_rationalize_d3phi(L, M, N);
}
//=================================================================================================
void NURBS::calc_d3phidetadxidxi (const Point& crd) const {
    const int L = 1, M = 0, N = 0;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    this->_d3phidz3[L][M][N] = this->_d3phidz3[N][L][M];
    if ( this->_is_rational ) this->_sum__w_d3Ndz3[L][M][N] = this->_sum__w_d3Ndz3[N][L][M];
}
//=================================================================================================
void NURBS::calc_d3phidetadxideta (const Point& crd) const {
    const int L = 1, M = 0, N = 1;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    this->_d3phidz3[L][M][N] = this->_d3phidz3[N][L][M];
    if ( this->_is_rational ) this->_sum__w_d3Ndz3[L][M][N] = this->_sum__w_d3Ndz3[N][L][M];
}
//=================================================================================================
void NURBS::calc_d3phidetadxidzeta (const Point& crd) const {
    const int L = 1, M = 0, N = 2;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    this->_d3phidz3[L][M][N] = this->_d3phidz3[M][L][N];
    if ( this->_is_rational ) this->_sum__w_d3Ndz3[L][M][N] = this->_sum__w_d3Ndz3[M][L][N];
}
//=================================================================================================
void NURBS::calc_d3phidetadetadxi (const Point& crd) const {
    const int L = 1, M = 1, N = 0;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    this->_d3phidz3[L][M][N] = this->_d3phidz3[N][L][M];
    if ( this->_is_rational ) this->_sum__w_d3Ndz3[L][M][N] = this->_sum__w_d3Ndz3[N][L][M];
}
//=================================================================================================
void NURBS::calc_d3phidetadetadeta (const Point& crd) const {
    const int L = 1, M = 1, N = 1;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a);
            this->_d3phidz3[L][M][N][a] = 0.0;
        }
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a);
            this->_d3phidz3[L][M][N][a] = this->_basef_0th_deriv<0>(crd(0), i)
                                        * this->_basef_3rd_deriv<1>(crd(1), j);
        }
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a),
                      k = this->index<2>(a);
            this->_d3phidz3[L][M][N][a] = this->_basef_0th_deriv<0>(crd(0), i)
                                        * this->_basef_3rd_deriv<1>(crd(1), j)
                                        * this->_basef_0th_deriv<2>(crd(2), k);
        }
    }
    //
    if ( this->_is_rational ) this->_rationalize_d3phi(L, M, N);
}
//=================================================================================================
void NURBS::calc_d3phidetadetadzeta (const Point& crd) const {
    const int L = 1, M = 1, N = 2;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a);
            this->_d3phidz3[L][M][N][a] = 0.0;
        }
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a),
            //          j = this->index<1>(a);
            this->_d3phidz3[L][M][N][a] = 0.0;
        }
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a),
                      k = this->index<2>(a);
            this->_d3phidz3[L][M][N][a] = this->_basef_0th_deriv<0>(crd(0), i)
                                        * this->_basef_2nd_deriv<1>(crd(1), j)
                                        * this->_basef_1st_deriv<2>(crd(2), k);
        }
    }
    //
    if ( this->_is_rational ) this->_rationalize_d3phi(L, M, N);
}
//=================================================================================================
void NURBS::calc_d3phidetadzetadxi (const Point& crd) const {
    const int L = 1, M = 2, N = 0;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    this->_d3phidz3[L][M][N] = this->_d3phidz3[N][L][M];
    if ( this->_is_rational ) this->_sum__w_d3Ndz3[L][M][N] = this->_sum__w_d3Ndz3[N][L][M];
}
//=================================================================================================
void NURBS::calc_d3phidetadzetadeta (const Point& crd) const {
    const int L = 1, M = 2, N = 1;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    this->_d3phidz3[L][M][N] = this->_d3phidz3[L][N][M];
    if ( this->_is_rational ) this->_sum__w_d3Ndz3[L][M][N] = this->_sum__w_d3Ndz3[L][N][M];
}
//=================================================================================================
void NURBS::calc_d3phidetadzetadzeta (const Point& crd) const {
    const int L = 1, M = 2, N = 2;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a);
            this->_d3phidz3[L][M][N][a] = 0.0;
        }
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a),
            //          j = this->index<1>(a);
            this->_d3phidz3[L][M][N][a] = 0.0;
        }
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a),
                      k = this->index<2>(a);
            this->_d3phidz3[L][M][N][a] = this->_basef_0th_deriv<0>(crd(0), i)
                                        * this->_basef_1st_deriv<1>(crd(1), j)
                                        * this->_basef_2nd_deriv<2>(crd(2), k);
        }
    }
    //
    if ( this->_is_rational ) this->_rationalize_d3phi(L, M, N);
}
//=================================================================================================
void NURBS::calc_d3phidzetadxidxi (const Point& crd) const {
    const int L = 2, M = 0, N = 0;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    this->_d3phidz3[L][M][N] = this->_d3phidz3[N][M][L];
    if ( this->_is_rational ) this->_sum__w_d3Ndz3[L][M][N] = this->_sum__w_d3Ndz3[N][M][L];
}
//=================================================================================================
void NURBS::calc_d3phidzetadxideta (const Point& crd) const {
    const int L = 2, M = 0, N = 1;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    this->_d3phidz3[L][M][N] = this->_d3phidz3[M][N][L];
    if ( this->_is_rational ) this->_sum__w_d3Ndz3[L][M][N] = this->_sum__w_d3Ndz3[M][N][L];
}
//=================================================================================================
void NURBS::calc_d3phidzetadxidzeta (const Point& crd) const {
    const int L = 2, M = 0, N = 2;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    this->_d3phidz3[L][M][N] = this->_d3phidz3[M][N][L];
    if ( this->_is_rational ) this->_sum__w_d3Ndz3[L][M][N] = this->_sum__w_d3Ndz3[M][N][L];
}
//=================================================================================================
void NURBS::calc_d3phidzetadetadxi (const Point& crd) const {
    const int L = 2, M = 1, N = 0;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    this->_d3phidz3[L][M][N] = this->_d3phidz3[N][M][L];
    if ( this->_is_rational ) this->_sum__w_d3Ndz3[L][M][N] = this->_sum__w_d3Ndz3[N][M][L];
}
//=================================================================================================
void NURBS::calc_d3phidzetadetadeta (const Point& crd) const {
    const int L = 2, M = 1, N = 1;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    this->_d3phidz3[L][M][N] = this->_d3phidz3[N][M][L];
    if ( this->_is_rational ) this->_sum__w_d3Ndz3[L][M][N] = this->_sum__w_d3Ndz3[N][M][L];
}
//=================================================================================================
void NURBS::calc_d3phidzetadetadzeta (const Point& crd) const {
    const int L = 2, M = 1, N = 2;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    this->_d3phidz3[L][M][N] = this->_d3phidz3[M][L][N];
    if ( this->_is_rational ) this->_sum__w_d3Ndz3[L][M][N] = this->_sum__w_d3Ndz3[M][L][N];
}
//=================================================================================================
void NURBS::calc_d3phidzetadzetadxi (const Point& crd) const {
    const int L = 2, M = 2, N = 0;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    this->_d3phidz3[L][M][N] = this->_d3phidz3[N][L][M];
    if ( this->_is_rational ) this->_sum__w_d3Ndz3[L][M][N] = this->_sum__w_d3Ndz3[N][L][M];
}
//=================================================================================================
void NURBS::calc_d3phidzetadzetadeta (const Point& crd) const {
    const int L = 2, M = 2, N = 1;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    this->_d3phidz3[L][M][N] = this->_d3phidz3[N][L][M];
    if ( this->_is_rational ) this->_sum__w_d3Ndz3[L][M][N] = this->_sum__w_d3Ndz3[N][L][M];
}
//=================================================================================================
void NURBS::calc_d3phidzetadzetadzeta (const Point& crd) const {
    const int L = 2, M = 2, N = 2;
    if ( ! this->_calculate_d3phidz3[L][M][N] ) return;
    //
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a);
            this->_d3phidz3[L][M][N][a] = 0.0;
        }
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a),
            //          j = this->index<1>(a);
            this->_d3phidz3[L][M][N][a] = 0.0;
        }
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a),
                      k = this->index<2>(a);
            this->_d3phidz3[L][M][N][a] = this->_basef_0th_deriv<0>(crd(0), i)
                                        * this->_basef_0th_deriv<1>(crd(1), j)
                                        * this->_basef_3rd_deriv<2>(crd(2), k);
        }
    }
    //
    if ( this->_is_rational ) this->_rationalize_d3phi(L, M, N);
}
//=================================================================================================
